/**
 * PMLib - PMLSequencer matrix example
 *
 * This example demonstrates matrix functionality of the PMLSequencer
 * object. Using the setInputMatrix() and setOutputMatrix() functions
 * input and output matrix objects may be assigned to the sequencer,
 * allowing automated sampling or parsing.
 *
 * Within the example, mouse location is routed through an input matrix, 
 * the results of which are sent to the sequencer. On playback, the
 * sequencer output is routed through the assigned output matrix and the 
 * results are rendered on screen.
 *
 * Usage: Press 'r' to initiate recording and then draw some shapes. Press
 * 'r' again to stop recording and save the sequence. Press 'p' to play
 * back the stored sequence.
 */

import pmlib.*;

PMLSequencer sequencer;
PMLMatrix m_input, m_output;

void setup() {
  size(400, 400);
  background(0);    

  // Create the input matrix
  m_input = new PMLMatrix();
  
  // Create two input channels using the screen dimensions as limits
  m_input.addInput("/mousex", 0, width);
  m_input.addInput("/mousey", 0, height);
  
  // Create two respective outputs with value scaling between 0 and 1
  m_input.addOutput("/mousex", "/varx", 0f, 1f);
  m_input.addOutput("/mousey", "/vary", 0f, 1f);

  // Create the output matrix
  m_output = new PMLMatrix();
  
  // Create two input channels matching the output channels of the previous matrix
  m_output.addInput("/varx", 0f, 1f);
  m_output.addInput("/vary", 0f, 1f);
  
  // Create two respective one-to-many mappings with value scaling to the screen dimensions
  // -- Changes to the output scale will be reflected in the playback rendering
  m_output.addOutput("/varx", "/px0", 0, width);
  m_output.addOutput("/vary", "/py0", 0, height);
  m_output.addOutput("/varx", "/px1", width, 0);
  m_output.addOutput("/vary", "/py1", height, 0);

  // Create the sequencer
  sequencer = new PMLSequencer();
  sequencer.setVerbose(true);
  
  // Set the input matrix and sampling interval
  // -- Changes to the sampling interval will be reflected in the playback rendering
  sequencer.setInputMatrix(m_input);
  sequencer.setMatrixInterval(1);
  
  // Set the output matrix
  sequencer.setOutputMatrix(m_output);  
}

void draw() {  
  // Route mouse location to the input matrix
  m_input.route("/mousex", mouseX);
  m_input.route("/mousey", mouseY);

  // Render the mouse location
  noStroke();
  fill(255, 0, 100, 150);
  ellipse(mouseX, mouseY, 10, 10);

  // Render the values of the output matrix
  fill(100, 0, 255, 150);
  ellipse(m_output.get("/px0").floatValue(), m_output.get("/py0").floatValue(), 10, 10);
  ellipse(m_output.get("/px1").floatValue(), m_output.get("/py1").floatValue(), 10, 10);
  
  // Iterate the sequencer by one frame
  sequencer.frame();
}

void keyReleased() {
  if(key == 'p' || key == 'P' && !sequencer.isRecording()) {
    if(!sequencer.isPlaying()) sequencer.read(sketchPath("example.seq"));   
    sequencer.play();
  }
  if(key == 'r' || key == 'R' && !sequencer.isPlaying()) {
    if(sequencer.isRecording()) sequencer.write(sketchPath("example.seq"));
    else {
      background(0);
      sequencer.record();
    }
  }
}

