package pmlib;
/**
 * PMLCurve Class
 * <p>The PMLCurve object allows the creation of value curves. The object iterates
 * between initial and target values over the specified time period.
 * <p>This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * <p>You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see http://www.gnu.org/licenses/.
 * @author Stephen Callear
 */
public class PMLCurve extends Object {
    private float value, pvalue, initial, target, exponent;
    private int time, time_ref;
    private long init_time;
    private boolean active;
    /**
     * Create an object instance with the specified initial value
     * @param v initial value
     */
    public PMLCurve(float v){
        value = v;
        pvalue = v;
        initial = v;
        target = v;
        exponent = 1f;
        time = 0;
        time_ref = 0;
        init_time = System.currentTimeMillis();
        active = false;
    }
    /**
     * Return the current object value
     * @return the current object value
     */
    public float floatValue(){
        pvalue = value;
        if (millis() < time_ref + time){
            double m = ((double) millis() - (double) time_ref) / (double) time;
            float c = (float) Math.pow(m, exponent);
            value = initial + (target - initial) * c;
        } else {
            if (value != target) value = target;
            active = false;
        }
        return value;
    }
    /**
     * Return the current object value as a rounded integer
     * @return the current object value as a rounded integer
     */
    public int intValue(){
        return Math.round(floatValue());
    }
    /**
     * Return the difference between the current and last requested object values    
     * @return the difference between the current and last requested object values
     */
    public float deltaFloatValue(){
        return value - pvalue;
    }
    /**
     * Return the difference between the current and last requested object values
     * @return the difference between the current and last requested object values
     */
    public int deltaIntValue(){
        return Math.round(value - pvalue);
        
    }
    /**
     * Set the object value to that specified
     * @param v value
     */
    public void set(float v){
        active = false;
        time = 0;
        initial = value;
        value =  v;
        target = v;
    }
    /**
     * Create a curve with the specified target value and iteration time
     * @param v target value
     * @param t iteration time
     */
    public void set(float v, int t){
        set(value, v, t, 1f);
    }
    /**
     * Create a curve with the specified target value, iteration time and curve
     * exponent
     * @param v target value
     * @param t iteration time
     * @param e curve exponent
     */
    public void set(float v, int t, float e){
        set(value, v, t, e);
    }
    /**
     * Create a curve with the specified initial and target values, and iteration
     * time
     * @param i initial value
     * @param v target value
     * @param t iteration time
     */
    public void set(float i, float v, int t){
        set(i, v, t, 1f);
    }
    /**
     * Create a curve with the specified initial and target values, iteration time
     * and curve exponent
     * @param i initial value
     * @param v target value
     * @param t iteration time
     * @param e curve exponent
     */
    public void set(float i, float v, int t, float e){
      active = true;
      time = t > 0 ? t : 0;
      time_ref = millis();
      target = v;
      initial = i;
      exponent = e;
    }
    /**
     * Return the activity state of the object
     * @return the activity state of the object
     */
    public boolean isActive(){
        return active;
    }
    /**
     * Return true if the value of this object has changed due to a call to set()
     * or internal ramping
     * <p>The value of this object is deemed to have changed if the difference
     * between current and previous values is less than 1*10^-10.
     * @return true if the value of this object has changed
     */
    public boolean hasChanged(){
        if(Math.abs(floatValue() - pvalue) > 1E-11f) return true;
        return false;
    }
    /**
     * Return the value of this object as a String
     * @return the value of this object as a String
     */
    @Override public String toString(){
        return String.valueOf(floatValue());
    }
    /**
     * Return the time in milliseconds since object creation
     * @return the time in milliseconds since object creation
     */
    private int millis(){
        long m = System.currentTimeMillis() - init_time;
        return (int) m;
    }
}
