package pmlib;
/**
 * PMLEvent Class
 * <p>The PMLEvent object allows an array of floating point values to be stored
 * alongside a String identifier and frame number.
 * <p>This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * <p>You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see http://www.gnu.org/licenses/.
 * @author Stephen Callear
 */
public class PMLEvent extends Object implements Comparable{
    private String tag;
    private float[] vars;
    private int frame;
    /**
     * Create a new object instance with the specified variables
     * @param t tag
     * @param f frame
     * @param v values
     */
    public PMLEvent(String t, int f, float[] v){
        this.tag = t;
        this.vars = v;
        this.frame = f;
    }
    /**
     * Return the event tag string
     * @return the event tag string
     */
    public String tag(){
        return tag;
    }
    /**
     * Return the stored values as an array of floats
     * @return the stored values as an array of floats
     */
    public float[] vars(){
        return vars;
    }
    /**
     * Set the frame number
     * @param f frame number
     */
    public void setFrame(int f){
        this.frame = f;
    }
    /**
     * Return the frame number
     * @return the frame number
     */
    public int frame(){
        return frame;
    }
    /**
     * Return the contents of the object as a string in the format
     * 'frame tag var_0 var_1 ... var_n'
     * @return the contents of the object as a string
     */
    @Override public String toString(){
        String str = Integer.toString(frame) + " " + tag + " ";
        for(int i = 0; i < vars.length; i ++){
            // Check float and format to string
            int ival = (int) vars[i];
            if(ival == vars[i])
                str = str.concat(Integer.toString(ival));
            else                
                str = str.concat(Float.toString(vars[i]));
            if(i < vars.length - 1)
                str = str.concat(" ");
        }
        return str;
    }
    /**
     * Create and return an SeqEvent object from a string in the format
     * 'frame tag var_0 var_1 ... var_n'
     * <p>If the frame or variable values cannot be parsed into integer and
     * float values respectively a null value will be returned.
     * @param s string
     * @return a PMLEvent object
     */
    public static PMLEvent fromString(String s){
        if(s == null || s.isEmpty()) return null;
        String[] str = s.split(" ");
        if(str.length < 2) return null;
        int f;
        float[] v = new float[str.length - 2];
        try {
            f = Integer.parseInt(str[0]);
            for(int i = 0; i < v.length; i ++)
                v[i] = Float.parseFloat(str[i + 2]);
        }
        catch (NumberFormatException e){
            return null;
        }
        return new PMLEvent(str[1], f, v);
    }
    /**
     * Implementation of the Comparable interface - objects are compared based on
     * their frame numbers
     * @param obj object to compare with
     * @return the result of the comparison
     */
    public int compareTo(Object obj){
        PMLEvent se = (PMLEvent) obj;
        if(se.frame() > frame) return -1;
        else if (se.frame() < frame) return 1;
        else return 0;
    }
    /**
     * Return true if the tag of this object matches that specified
     * @param t tag
     * @return true if the tag if this object matches that specified
     */
    public boolean hasTag(String t){
        if(tag.equals(t)) return true;
        return false;
    }
    /**
     * Return true if the variables of this object match those specified
     * @param v variables
     * @return true if the variables of this object match those specified
     */
    public boolean hasVars(float[] v){
        if(vars.length != v.length) return false;
        for(int i = 0; i < vars.length; i ++)
            if(vars[i] != v[i]) return false;
        return true;
    }
    /**
     * Return a copy of this object
     * @return a copy of this object
     */
    public PMLEvent copy(){
        PMLEvent e = new PMLEvent(tag, frame, vars);
        return e;
    }
}
