package pmlib;
import java.util.ArrayList;
import java.lang.reflect.Method;
/**
 * PMLEventHandler Class
 * <p>The PMLEventHandler object permits the routing of PMLEvent and PMLFloat 
 * objects to user specified methods using reflection.
 * <p>This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * <p>You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see http://www.gnu.org/licenses/.
 * @author Stephen Callear
 */
public class PMLEventHandler extends Object {
    private ArrayList<TaggedMethod> methods;
    /**
     * Create a new object instance
     */
    public PMLEventHandler(){
        methods = new ArrayList<TaggedMethod>();
    }
    /**
     * Add an object method to the event handler
     * <p>Object methods should be public with no return value. Currently only 
     * methods with <code>int</code>, <code>float</code> and <code>String</code>
     * arguments are supported. This limitation is based on the assertion that 
     * other data types are unnecessary within the context of the PMLib library.
     * <p>Method arguments should be specified by an array of characters with
     * <code>'i'</code>, <code>'f'</code> and <code>'s'</code> denoting
     * <code>int</code>, <code>float</code> and <code>String</code> respectively.
     * For example, a method with the arguments <code>(int, String, float, float)</code>
     * should be specified with <code>{ 'i', 's', 'f', 'f' }</code>.
     * @param tag tag
     * @param mobj method object
     * @param mstr method string
     * @param args method arguments
     * @return true if the method was successfully added
     */
    public boolean add(String tag, Object mobj, String mstr, char[] args){
        Method m = getMethod(mobj, mstr, args);
        if(m == null) return false;        
        return methods.add(new TaggedMethod(tag, mobj, m));        
    }
    /**
     * Remove all object methods with the specified tag
     * @param tag tag
     * @return the total methods removed
     */
    public int remove(String tag){
        int c = 0;
        for(TaggedMethod m : methods){
            if(m.hasTag(tag)){
                methods.remove(m);
                c ++;
            }
        }
        return c;
    }
    /**
     * Remove all object methods matching that specified
     * @param mobj method object
     * @param mstr method string
     * @param args method arguments
     * @return the total methods removed
     */
    public int remove(Object mobj, String mstr, char[] args){
        int c = 0;
        Method tm = getMethod(mobj, mstr, args);
        if( tm == null) return c;
        for(TaggedMethod m :  methods){
            if(m.hasMethod(tm)){
                methods.remove(m);
                c ++;
            }
        }
        return c;
    }
    /**
     * Clear all object methods
     */
    public void clear(){
        methods.clear();
    }
    /**
     * Print the contents of this object to the output
     */
    public void debug(){
        System.out.println(this.toString() + " - " + Integer.toString(size()) + " methods.");
        int c = 0;
        for(TaggedMethod m : methods)
            System.out.println(Integer.toString(c++) + ": " + m.toString());
    }
    /**
     * Route the value of the specified object to any attached methods with
     * matching tags
     * @param flt tagged float
     */
    public void route(PMLFloat flt){
        for(TaggedMethod m : methods){
            if(m.hasTag(flt.tag()))
                invokeMethod(m.object(), m.method(), flt);
        }
    }
    /**
     * Route the values of the specified objects to any attached methods with
     * matching tags
     * @param flt tagged float array
     */
    public void route(PMLFloat[] flt){
        for(PMLFloat v : flt)
            route(v);
    }
    /**
     * Route the event variables of the specified object to any attached methods
     * with matching tag
     * @param evnt tagged event
     */
    public void route(PMLEvent evnt){
        for(TaggedMethod m : methods){
            if(m.hasTag(evnt.tag()))
                invokeMethod(m.object(), m.method(), evnt);
        }
    }
    /**
     * Route the event variables of the specified objects to any methods with
     * matching tags
     * @param evnt event
     */
    public void route(PMLEvent[] evnt){
        for(PMLEvent e : evnt)
            route(e);
    }    
    /**
     * Return the number of attached methods
     * @return the number of attached methods
     */
    public int size(){
        return methods.size();        
    }
    private static boolean invokeMethod(Object mobj, Method meth, PMLFloat flt){
        try {
            meth.invoke(mobj, formatArgs(meth, flt));
        }
        catch (Exception e){
            System.err.println("Error: Invokation failed - " + e.toString());
            return false;
        }
        return true;
    }
    private static boolean invokeMethod(Object mobj, Method meth, PMLEvent evnt){
        try {
            meth.invoke(mobj, formatArgs(meth, evnt));
        }
        catch (Exception e){
            System.err.println("Error: Invokation failed - " + e.toString());
            return false;
        }
        return true;               
    }
    private static Object[] formatArgs(Method meth, PMLFloat flt){
        Class[] marg = meth.getParameterTypes();
        if(marg.length != 1) return null;
        Object arg;
        if(marg[0].isAssignableFrom(int.class)) arg = flt.intValue();
        else if(marg[0].isAssignableFrom(float.class)) arg = flt.floatValue();
        else if(marg[0].isAssignableFrom(String.class)) arg = Float.toString(flt.floatValue());
        else return null;
        return new Object[] { arg };  
    }
    private static Object[] formatArgs(Method meth, PMLEvent evnt){
        Class[] marg = meth.getParameterTypes();
        if(evnt.vars().length < marg.length) return null;
        Object[] args = new Object[marg.length];
        for(int i = 0; i < args.length; i ++){
            if(marg[i].isAssignableFrom(int.class)) args[i] = (int) evnt.vars()[i];
            else if(marg[i].isAssignableFrom(float.class)) args[i] = evnt.vars()[i];
            else if(marg[i].isAssignableFrom(String.class)) args[i] = Float.toString(evnt.vars()[i]);
            else return null;
        }
        return args;        
    }
    
    private static Method getMethod(Object mobj, String mstr, char[] args){
        Class c = mobj.getClass();
        Method[] m = c.getDeclaredMethods();
        for(int i = 0; i < m.length; i ++)
            if(m[i].getName().equals(mstr)){
                if(hasArguments(m[i], args)) return m[i];
            }        
        return null;
    }
    private static boolean hasArguments(Method meth, char[] args){
        Class[] carg = new Class[args.length];
        for(int i = 0; i < args.length; i ++){
            if(args[i] == 'i') carg[i] = int.class;
            else if(args[i] == 'f') carg[i] = float.class;
            else if(args[i] == 's') carg[i] = String.class;
            else return false;
        }
        return hasArguments(meth, carg);        
    }
    private static boolean hasArguments(Method meth, Class[] args){
        Class[] marg = meth.getParameterTypes();
        if(marg.length != args.length) return false;
        for(int i = 0; i < marg.length; i ++)
            if(!marg[i].isAssignableFrom(args[i])) return false;
        return true;            
    }               
    private static class TaggedMethod extends Object {
        private String tag;
        private Object object;
        private Method method;        
        public TaggedMethod(String t, Object o, Method m){
            tag = t;
            object = o;
            method = m;               
        }
        public boolean hasTag(String t){
            return tag.equals(t);
        }
        public boolean hasObject(Object o){
            return object.equals(o);
        }
        public boolean hasMethod(Method m){
            return method.equals(m);
        }        
        public String tag(){
            return tag;            
        }
        public Object object(){
            return object;
        }
        public Method method(){
            return method;
        }
        @Override public String toString(){
            return tag + " - " + method.toString();
        }
    }
}
