package pmlib;
/**
 * PMLFloat Class
 * <p>The PMLFloat object allows the storage of a float value alongside a
 * String identifier.
 * <p>This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * <p>You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see http://www.gnu.org/licenses/.
 * @author Stephen Callear
 */
public class PMLFloat extends Object {
    private String tag;
    private PMLCurve value;
    private PMLRange range;
    private int ramp;
    /**
     * Create a new object instance with the specified tag and value with no value
     * ramping
     * @param t tag
     * @param v value
     */
    public PMLFloat(String t, float v){
        this(t, v, 0);
    }
    /**
     * Create a new object instance with the specified tag, value and ramp duration
     * @param t tag
     * @param v value
     * @param r ramp duration (ms)
     */
    public PMLFloat(String t, float v, int r){
        tag = t;
        value = new PMLCurve(v);
        ramp = r;
    }
    /**
     * Return the stored value as a boolean
     * <p>If a value range is specified, true will be returned if the stored value
     * is greater than the midpoint of the range. If no value range is specified,
     * true will be returned if the stored value is greater than zero.
     * @return the stored value as a boolean
     */
    public boolean booleanValue(){
        if(range == null)
            return value.floatValue() > 0f ? true : false;
        else
            return range.normalise(value.floatValue()) > 0.5f ? true : false;
    }
    /**
     * Return the stored value as a rounded integer
     * @return the stored value as a rounded integer
     */
    public int intValue(){
        if(range == null)
            return value.intValue();    
        else
            return range.limit(value.intValue());            
    }
    /**
     * Return the stored value as a float
     * @return the stored value as a float
     */
    public float floatValue(){
        if(range == null)
            return value.floatValue();
        else           
            return range.limit(value.floatValue());
    }    
    /**
     * Return a PMLEvent object with the tag and value of this object instance
     * and the frame number specified
     * @param f frame
     * @return a PMLEvent object with the tag and value of this object instance
     */
    public PMLEvent toEvent(int f){
        return new PMLEvent(tag, f, new float[] { value.floatValue() });
    }
    /**
     * Set the value of this object
     * @param v value
     */
    public void set(int v){
        value.set((float) v, ramp);
    }
    /**
     * Set the value of this object
     * @param v value
     */
    public void set(float v){
        value.set(v, ramp);
    }
    /**
     * Set the ramp duration
     * @param r ramp duration (ms)
     */
    public void setRamp(int r){
        ramp = r;
    }
    /**
     * Define the output range of this object
     * <p>If the specified object is null, the output will not be limited.
     */
    public void setRange(PMLRange r){
        range = r;
    }
    /**
     * Return the output range for this object
     * <p>If this object has no specified range, null will be returned.
     * @return the output range for this object 
     */
    public PMLRange getRange(){
        return range;
    }
    /**
     * Return the stored tag
     * @return the stored tag
     */
    public String tag(){
        return tag;
    }
    /**
     * Return true if the object tag matches that specified
     * @param t tag
     * @return true if the object tag matches that specified
     */
    public boolean hasTag(String t){
        return tag.equals(t);
    }
    /**
     * Return true if the value of this object differs from the previous set value
     * <p>The value is deemed not to have changed if the difference between current
     * and previous values is less than 1*10^-10.
     * @return true if the value of this object differs from the previous set value
     */
    public boolean hasChanged(){
        return value.hasChanged();
    }  
    /**
     * Return the PMLFloat values as a String
     * @return the PMLFloat values as a String
     */
    @Override public String toString(){
        return tag + " " + value.toString();
    }
}
