package pmlib;
import java.util.ArrayList;
import java.util.Arrays;
/**
 * PMLMatrix Class
 * <p>The PMLMatrix object allows multiple PMLInput objects to be attached. Values
 * may be routed to each attached object using the relevant tag.
 * <p>This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * <p>You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see http://www.gnu.org/licenses/.
 * @author Stephen Callear
 */
public class PMLMatrix extends Object {
    private ArrayList<PMLInput> matrix;    
    private PMLRoute output;
    private PMLEventHandler eventHandler;
    /**
     * Create a new object instance with the default size
     */
    public PMLMatrix(){
        matrix = new ArrayList();
        output = new PMLRoute();
    }
    /**
     * Add an input to the matrix with the specified tag and default limits (0, 1)
     * @param t tag
     */
    public void addInput(String t){
        addInput(t, 0f, 1f);
    }
    /**
     * Add an input to the matrix with the specified tag and limits
     * @param t tag
     * @param min minimum
     * @param max maximum
     */
    public void addInput(String t, float min, float max){
        matrix.add( new PMLInput(t, min, max) );
    }
    /**
     * Add an output to the matrix with default limits (0, 1) and no value ramping
     * @param t_in input channel to attach the output to
     * @param t_out output channel tag
     * @return true if the channel was successfully created
     */
    public boolean addOutput(String t_in, String t_out){
        return addOutput(new String[] { t_in }, t_out, 0f, 1f, 0);
    }
    /**
     * Add an output to the matrix with the specified limits and no value ramping
     * @param t_in input channel to attach the output to
     * @param t_out output channel tag
     * @param min minimum
     * @param max maximum
     * @return true if the channel was successfully created
     */
    public boolean addOutput(String t_in, String t_out, float min, float max){
         return addOutput(new String[] { t_in }, t_out, min, max, 0);
    }
    /**
     * Add an an output to the matrix with the specified limits and ramp duration
     * @param t_in input channel to attach the output to
     * @param t_out output channel tag
     * @param min minimum
     * @param max maximum
     * @param rmp ramp duration (ms)
     * @return true if the channel was successfully created
     */
    public boolean addOutput(String t_in, String t_out, float min, float max, int rmp){
        return addOutput(new String[] { t_in }, t_out, min, max, rmp);
    }
    /**
     * Add an output to multiple input channels with default limits (0, 1) and
     * no value ramping
     * @param t_in input channels to attach the output to
     * @param t_out output channel tag
     * @return true if the channel was successfully created
     */
    public boolean addOutput(String[] t_in, String t_out){
        return addOutput(t_in, t_out, 0f, 1f, 0);
    }
    /**
     * Add an output to the multiple input channels with the specified limits
     * and no value ramping
     * @param t_in input channels to attach the output to
     * @param t_out output channel tag
     * @param min minimum
     * @param max maximum
     * @return true if the channel was successfully created
     */
    public boolean addOutput(String[] t_in, String t_out, float min, float max){
        return addOutput(t_in, t_out, min, max, 0);
    }
    /**
     * Add an an output to multiple input channels the specified limits and ramp
     * duration
     * @param t_in input channels to attach the output to
     * @param t_out output channel tag
     * @param min minimum
     * @param max maximum
     * @param rmp ramp duration (ms)
     * @return true if the channel was successfully created
     */
    public boolean addOutput(String t_in[], String t_out, float min, float max, int rmp){
        int c = 0;
        for(int i = 0; i < t_in.length; i ++){
            PMLInput[] ca = getInput(t_in[i]);
            for(int j = 0; j < ca.length; j ++){
                if(!ca[j].hasOutput(t_out)){
                    ca[j].attach(new PMLRange(t_out, min, max));
                    c ++;
                }
            }
        }
        if(c > 0){
            output.add(t_out, min, rmp);
            return true;
        }
        return false;
    }
    /**
     * Remove any inputs with the specified tag
     * @param t tag
     * @return total input channels removed
     */
    public int removeInput(String t){
        PMLInput[] ca = getInput(t);
        if(ca.length == 0) return 0;
        for(int i = 0; i < ca.length; i ++) matrix.remove(ca[i]);
        return ca.length;
    }
    /**
     * Remove any outputs with the specified tag
     * @param t tag
     * @return total output channels removed
     */
    public int removeOutput(String t){
        int count = 0;
        for(PMLInput ch : matrix)
            count += ch.removeOutput(t);
        output.remove(t);
        return count;
    }
    /**
     * Return a reference to the output PMLFloat object with the specified tag
     * @param t tag
     * @return a reference to the output PMLFloat object with the specified tag
     */
    public PMLFloat get(String t){
        return output.get(t);
    }
    /**
     * Return references to all output PMLFloat objects as an array
     * @return references to all output PMLFloat objects as an array
     */
    public PMLFloat[] getAll(){
        return output.toArray();
    }
    /**
     * Clear the matrix
     */
    public void clear(){        
        matrix.clear();
        output.clear();
    }
    /**
     * Return the total input and output channels
     * @return the total input and output channels
     */
    public int size(){
        return matrix.size() + output.size();
    }    
    /**
     * Return true if an input with the specified tag is attached
     * @param t tag
     * @return true if an input with the specified tag is attached
     */
    public boolean hasInput(String t){
        for(PMLInput ch : matrix)
            if(ch.hasTag(t)) return true;
        return false;
    }
    /**
     * Return true if an output with the specified tag is attached
     * @param t tag
     * @return true if an output with the specified tag is attached
     */
    public boolean hasOutput(String t){
        for(PMLInput ch : matrix)
            if(ch.hasOutput(t)) return true;
        return false;
    }
    /**
     * Route the specified PMLFloat to any applicable inputs
     * <p>If a valid input channel exists the mapped value will be routed
     * to the relevant output, a reference to which may be obtained using the
     * <code>get()</code> method. If an event handler is specified, the mapped
     * value will also be routed to any specified object methods.
     * @param f PMLFloat
     */
    public void route(PMLFloat f){
        route(f.tag(), f.floatValue());
    }
    /**
     * Route the specified value to any applicable inputs
     * <p>If a valid input channel exists the mapped value will be routed
     * to the relevant output, a reference to which may be obtained using the
     * <code>get()</code> method. If an event handler is specified, the mapped
     * value will also be routed to any specified object methods.
     * @param t tag
     * @param v value
     */
    public void route(String t, float v){
        ArrayList<PMLFloat> fl = new ArrayList();
        for(PMLInput ch : matrix){
            if(ch.hasTag(t))
                fl.addAll(Arrays.asList(ch.route(v)));
        }
        PMLFloat[] fa = new PMLFloat[fl.size()];
        fl.toArray(fa);
        if(output != null) output.route(fa);
        if(eventHandler != null) eventHandler.route(fa);
    }
    /**
     * Set the matrix event handler
     * <p>If the event handler is not null, matrix values will be routed to valid
     * object methods specified within the event handler upon calling <code>route()</code>.
     * @param h event handler
     */
    public void setEventHandler(PMLEventHandler h){
        eventHandler = h;
    }    
    private PMLInput[] getInput(String t){
        ArrayList<PMLInput> cl = new ArrayList();
        for(PMLInput ch : matrix)
            if(ch.hasTag(t)) cl.add(ch);
        PMLInput[] ca = new PMLInput[cl.size()];
        cl.toArray(ca);
        return ca;        
    }
    /**
     * PMLInput Class    
     */
    private static class PMLInput extends PMLRange {
        private ArrayList<PMLRange> channels;
        public PMLInput(String t){
            this(t, 0.0f, 1.0f, 16);
        }
        public PMLInput(String t, int s){
            this(t, 0.0f, 1.0f, s);
        }
        public PMLInput(String t, float a, float b){
            this(t, a, b, 16);
        }
        public PMLInput(String t, float a, float b, int s){
            super(t, a, b);
            channels = new ArrayList(s);
        }
        public void attach(PMLRange ch){
            channels.add(ch);
        }
        public void remove(PMLRange ch){
            channels.remove(ch);
        }
        public void clear(){
            channels.clear();
        }
        public boolean hasOutput(String t){
            for(PMLRange ch : channels)
                if(ch.hasTag(t)) return true;
            return false;
        }
        public int removeOutput(String t){
            PMLRange[] ca = getOutput(t);
            if(ca.length == 0) return 0;
            for(int i = 0; i < ca.length; i ++) channels.remove(ca[i]);
            return ca.length;
        }
        public PMLRange[] getOutput(String t){
            ArrayList<PMLRange> cl = new ArrayList();
            for(PMLRange ch : channels)
                if(ch.hasTag(t)) cl.add(ch);
            PMLRange[] ca = new PMLRange[cl.size()];
            cl.toArray(ca);
            return ca;
        }
        public int size(){
            return channels.size();
        }
        public PMLFloat[] route(float v){
            PMLFloat[] fa = new PMLFloat[channels.size()];
            for(int i = 0; i < channels.size(); i ++){
                PMLRange ch = channels.get(i);
                fa[i] = new PMLFloat(ch.tag(), ch.scale(normalise(v)));
            }
            return fa;
        }
    }      
}

