package pmlib;
/**
 * PMLRange Class
 * <p>The PMLRange object allows the storage of upper and lower range values, alongside
 * an identifier tag. The class provides limit, normalise and scale functionality.
 * <p>This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * <p>You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see http://www.gnu.org/licenses/.
 * @author Stephen Callear
 */
public class PMLRange extends Object {
    private String tag;
    private float min, max;
    /**
     * Create a new object instance with the specified minimum and maximum values
     * @param a minimum
     * @param b maximum
     */
    public PMLRange(int a, int b){
        this(null, (float) a, (float) b);
    }
    /**
     * Create a new object instance with the specified minimum and maximum values
     * @param a minimum
     * @param b maximum
     */
    public PMLRange(float a, float b){
        this(null, a, b);
    }
    /**
     * Create a new object instance with the specified tag, minimum and maximum
     * values
     * @param t tag
     * @param a minimum
     * @param b maximum
     */
    public PMLRange(String t, int a, int b){
        this(t, (float) a, (float) b);
    }
    /**
     * Create a new object instance with the specified tag, minimum and maximum
     * values
     * @param t tag
     * @param a minimum
     * @param b maximum
     */
    public PMLRange(String t, float a, float b){
        tag = t;
        min = a;
        max = b;
    }
    /**
     * Return the stored minimum value
     * @return the stored minimum value
     */
    public float min(){
        return min;
    }
    /**
     * Return the stored maximum value
     * @return the stored maximum value
     */
    public float max(){
        return max;
    }
    /**
     * Return the stored tag string
     * @return the stored tag string
     */
    public String tag(){
        return tag;
    }
    /**
     * Return true if the stored tag matches that specified
     * @param t tag
     * @return true if the stored tag matches that specified
     */
    public boolean hasTag(String t){
        return tag.equals(t);
    }
    /**
     * Limit the specified value to the range of this object
     * <p>The returned value will be rounded to the nearest integer value.
     * @param v value
     * @return the limited value
     */
    public int limit(int v){
        return Math.round(limit((float) v, min, max));
    }
    /**
     * Limit the specified value to the range of this object
     * @param v value
     * @return the limited value
     */
    public float limit(float v){
        return limit(v, min, max);
    }
    /**
     * Limit the specified value to the specified range
     * @param v value
     * @param a minimum
     * @param b maximum
     * @return the limited value
     */
    public static float limit(float v, float a, float b){
        if(v < a) return a;
        else if(v > b) return b;
        else return v;
    }
    /**
     * Normalise the specified value based on the range of this object
     * @param v value
     * @return the normalised value
     */
    public float normalise(int v){
        return normalise((float) v, min, max);
    }
    /**
     * Normalise the specified value based on the range of this object
     * @param v value
     * @return the normalised value
     */
    public float normalise(float v){
        return normalise(v, min, max);
    }
    /**
     * Normalise the specified value based on the specified range
     * @param v value
     * @param a minimum
     * @param b maximum
     * @return the normalised value
     */
    public static float normalise(float v, float a, float b){
        float vmin, vmax;
        boolean invert;
        if(a > b){
            vmin = b;
            vmax = a;
            invert = true;
        } else {
            vmin = a;
            vmax = b;
            invert = false;
        }
        float nv = (limit(v, vmin, vmax) - vmin) / (vmax - vmin);
        if(invert) return 1.0f - nv;
        else return nv;
    }
    /**
     * Scale the specified value to the range of this object
     * @param v value (0.0 - 1.0)
     * @return the scaled value
     */
    public float scale(float v){
        return scale(v, min, max);
    }
    /**
     * Scale the specified value to the specified range
     * @param v value (0.0 - 1.0)
     * @param a minimum
     * @param b maximum
     * @return the scaled value
     */
    public static float scale(float v, float a, float b){        
        float vlim, vmin, vmax;
        if(a > b) {
            vlim = 1.0f - limit(v, 0.0f, 1.0f);
            vmin = b;
            vmax = a;            
        } else {
            vlim = limit(v, 0.0f, 1.0f);
            vmin = a;
            vmax = b;
        }
        return vlim * (vmax - vmin) + vmin;
    }
    /**
     * Return the contents of this object as a String
     * @return the contents of this object as a String
     */
    @Override public String toString(){
        String ts = "";
        if(tag != null) ts = tag + " ";
        return ts + min + " " + max;
    }
}
