package pmlib;
import java.util.ArrayList;
/**
 * PMLRoute Class
 * <p>The PMLRoute object facilitates the dynamic creation and storage of multiple
 * PMLFloat objects. Values sent to the object are routed to any stored objects
 * with the correct tag preserving references. Only one instance of each tag is
 * permitted per object.
 * <p>This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * <p>You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see http://www.gnu.org/licenses/.
 * @author Stephen Callear
 */
public class PMLRoute extends Object{
    private ArrayList<PMLFloat> map;
    /**
     * Create a new object instance with the default size
     */
    public PMLRoute(){
        map = new ArrayList();
    }
    /**
     * Add a tag to the object
     * <p>The tag will only be added if no such tag is already stored
     * @param t tag
     * @return true if the tag was successfully added
     */
    public boolean add(String t){
        return add(t, 0f, 0);
    }
    /**
     * Add a tag to the object with the specified initial value
     * <p>The tag will only be added if no such tag is already stored
     * @param t tag
     * @param v initial value
     * @return true if the tag was successfully added
     */
    public boolean add(String t, float v){
        return add(t, v, 0);
    }
    /**
     * Add a tag to the object with the specified ramp time
     * <p>The tag will only be added if no such tag is already stored
     * @param t tag
     * @param r ramp time (ms)
     * @return true if the tag was successfully added
     */
    public boolean add(String t, int r){
        return add(t, 0f, r);
    }
    /**
     * Add a tag to the object with the specified initial value and ramp time
     * <p>The tag will only be added if no such tag is already stored
     * @param t tag
     * @param v initial value
     * @param r ramp time (ms)
     * @return true if the tag was successfully added
     */
    public boolean add(String t, float v, int r){
        if(get(t) == null){
            map.add(new PMLFloat(t, v, r));
            return true;
        }
        return false;
    }
    /**
     * Remove a tag from the object
     * @param t tag
     * @return true if the tag was successfully removed
     */
    public boolean remove(String t){
        PMLFloat ch = get(t);
        if(ch != null){
            map.remove(ch);
            return true;
        }
        return false;
    }
    /**
     * Remove all tags from the object
     */
    public void clear(){
        map.clear();
    }
    /**
     * Route an array of data to any stored PMLFloat objects with the specified tags
     * @param f data array
     * @return the number of modified objects
     */
    public int route(PMLFloat[] f){
        int c = 0;
        for(int i = 0; i < f.length; i ++)
            if(route(f[i])) c ++;
        return c;
    }
    /**
     * Route data to any stored PMLFloat objects with the specified tag
     * @param f data
     * @return true if the specified tag exists
     */
    public boolean route(PMLFloat f){
        return route(f.tag(), f.floatValue());
    }
    /**
     * Route data to any stored PMLFloat objects with the specified tag
     * @param t tag
     * @param v value
     * @return true if the specified tag exists
     */
    public boolean route(String t, float v){
        PMLFloat ch = get(t);
        if(ch != null){
            ch.set(v);
            return true;
        }
        return false;
    }
    /**
     * Return the stored PMLFloat object with the specified tag
     * <p>
     * Returns null if no such object was found
     * @param t tag
     * @return the stored PMLFloat object with the specified tag
     */
    public PMLFloat get(String t){
        for(PMLFloat ch : map)
            if(ch.hasTag(t)) return ch;
        return null;
    }
    /**
     * Return all stored PMLFloat objects as an array
     * @return all stored PMLFloat objects as an array
     */
    public PMLFloat[] toArray(){
        PMLFloat[] arr = new PMLFloat[map.size()];
        map.toArray(arr);
        return arr;
    }
    /**
     * Return the size of this object
     * @return the size of this object
     */
    public int size(){
        return map.size();
    }
}

