package pmlib;
import java.util.ArrayList;
import java.util.Arrays;
import java.io.File;
/**
 * PMLSequencer Class
 * <p>The PMLSequencer object facilitates the recording and playback of PMLEvent
 * objects. Sequences may be loaded from, or saved to text files and the class
 * includes static sequence load, reverse, offset and combine functions.
 * <p>This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * <p>You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see http://www.gnu.org/licenses/.
 * @author Stephen Callear
 */
public class PMLSequencer extends Object {
    private PMLEvent [] sequence_p;
    private PMLMatrix matrix_in, matrix_out;
    private PMLEventHandler eventHandler;
    private int frame_count, event_count, interval;
    private boolean playback, record, verbose;
    private String tag;
    private ArrayList<PMLEvent> sequence_r;
    /**
     * Create a new object instance
     */
    public PMLSequencer(){
        this("PMLSequencer");
    }
    /**
     * Create a new object instance with the specified tag
     */
    public PMLSequencer(String t){    
        frame_count = 0;
        event_count = 0;
        interval = 1;
        playback = false;
        record = false;
        verbose = false;
        sequence_r = new ArrayList(1024);
        tag = t;
    }
    /**
     * Iterate the sequencer by one frame
     * <p>If the sequencer is in playback, frame events will be parsed and output to
     * the output matrix if applicable. If the sequencer is recording and a matrix
     * is assigned, the input matrix will be sampled.
     * @return all parsed events as an array of PMLEvent objects
     */
    public PMLEvent[] frame(){
        return frame(1);
    }
    /**
     * Iterate the sequencer by the specified number of frames
     * <p>If the sequencer is in playback, frame events will be parsed and output to
     * the output matrix and event handler if applicable. If the sequencer is 
     * recording and a matrix is assigned, the input matrix will be sampled.
     * @param f frames
     * @return all parsed events as an array of PMLEvent objects
     */
    public PMLEvent[] frame(int f){
        PMLEvent[] events = new PMLEvent[0];
        if(!isActive()) return events;
        if(playback){
            if(frame_count > getSequenceLength()) stopPlayback();
            else events = parseEvents(f);
            matrixOutput(events);
            if(eventHandler != null) eventHandler.route(events);
        }
        if(record) matrixSample();
        frame_count += f;
        return events;
    }
    private PMLEvent[] parseEvents(int f){
        ArrayList lst = new ArrayList();
        boolean p = true;
        while(p && event_count < sequence_p.length){
            if(sequence_p[event_count].frame() < frame_count + f){
                lst.add(sequence_p[event_count]);
                event_count ++;
            }
            else p = false;
        }
        PMLEvent[] arr;
        arr = new PMLEvent[lst.size()];
        lst.toArray(arr);
        return arr;
    }
    private void matrixSample(){
        if(matrix_in == null || frame_count % interval != 0) return;
        PMLFloat[] pfa = matrix_in.getAll();
        for(PMLFloat pf : pfa)
            if(pf.hasChanged()) sequence_r.add(createEvent(pf));
    }
    private void matrixOutput(PMLEvent[] e){
        if(matrix_out == null ) return;
        for(PMLEvent ev : e)
            if(ev.vars().length > 0) matrix_out.route(ev.tag(), ev.vars()[0]);
    }
    /**
     * Return the current sequencer frame
     * @return the current sequencer frame;
     */
    public int getCurrentFrame(){
        return frame_count;
    }
    /**
     * Return the length in frames of the most recently played sequence
     * @return the length in frames of the most recently played sequence
     */
    public int getSequenceLength(){
        if(sequence_p == null || sequence_p.length < 1) return 0;
        return sequence_p[sequence_p.length - 1].frame();
    }
    /**
     * Return true if the loaded sequence contains the specified event
     * @param s event string
     * @return true if the loaded sequence contains the specified event
     */
    public boolean hasEvent(String s){
        if(sequence_p == null) return false;
        for(int i = 0; i < sequence_p.length; i ++)
            if(sequence_p[i].hasTag(s)) return true;
        return false;
    }
    /**
     * Return the amount of variables attached to the first instance of the
     * specified event.
     * <p>If the sequence is empty or the specified event cannot be found, -1 will
     * be returned.
     * @param s event string
     * @return the amount of variables attached to the specified event
     */
    public int getEventVars(String s){
        if(sequence_p == null) return -1;
        for(int i = 0; i < sequence_p.length; i ++){
            if(sequence_p[i].hasTag(s)) return sequence_p[i].vars().length;
        }
        return -1;
    }
    /**
     * Return the event range for the specified event and variable index
     * <p>The variable index specifies the event variable from which to derive
     * the event range, with a value of 0 specifying the first variable and so on.
     * <p>The return float array takes the format: { minimum, maximum, mean }
     * @param s event string
     * @param v variable index
     * @return the event range for the specified event and variable index
     */
    public float[] getEventRange(String s, int v){
        if(sequence_p == null) return null;
        Float min = null;
        Float max = null;
        int count = 0;
        float mean = 0;
        for(int i = 0; i < sequence_p.length; i ++){
            if(sequence_p[i].hasTag(s) && v < sequence_p[i].vars().length){
                if(min == null || sequence_p[i].vars()[v] < min.floatValue())
                    min = new Float(sequence_p[i].vars()[v]);
                if(max == null || sequence_p[i].vars()[v] > max.floatValue())
                    max = new Float(sequence_p[i].vars()[v]);
                mean += sequence_p[i].vars()[v];
                count ++;
            }
        }
        if(count > 0) mean /= (float) count;
        if(min == null || max == null) return null;
        return new float[] { min.floatValue(), max.floatValue(), mean };
    }
    /**
     * Return a new PMLEvent object for the current frame derived from the specified
     * PMLFloat object
     * @param f PMLFloat object
     * @return a new PMLEvent object for the current frame
     */
    public PMLEvent createEvent(PMLFloat f){
        return f.toEvent(frame_count);
    }
    /**
     * Return a new PMLEvent object for the current frame with the specified variables
     * @param s event tag
     * @return a new PMLEvent object for the current frame
     */
    public PMLEvent createEvent(String s){
        return createEvent(s, new float[] { });
    }
    /**
     * Return a new PMLEvent object for the current frame with the specified variables
     * @param s event tag
     * @param v variable
     * @return a new PMLEvent object for the current frame
     */
    public PMLEvent createEvent(String s, float v){
        return createEvent(s, new float[] { v });
    }
    /**
     * Return a new PMLEvent object for the current frame with the specified variables
     * @param s event tag
     * @param v variables
     * @return a new PMLEvent object for the current frame
     */
    public PMLEvent createEvent(String s, float[] v){
        return new PMLEvent(s, frame_count, v);
    }
    
    /**
     * Store a sequencer event if the sequencer is recording
     * @param e sequencer event
     * @return true if the event was stored
     */
    public boolean event(PMLEvent e){
        if(record){
            sequence_r.add(e);
            return true;
        }
        return false;
    }
    /**
     * Store a sequencer event if the sequencer is recording
     * @param s event string
     * @return true if the event was stored
     */
    public boolean event(String s){
        return event(s, new float[] { });
    }
    /**
     * Store a sequencer event if the sequencer is recording
     * @param s event string
     * @param v variable
     * @return true if the event was stored
     */
    public boolean event(String s, float v){
        return event(s, new float[] { v });
    }
    /**
     * Store a sequencer event if the sequencer is recording
     * @param s event string
     * @param v variable array
     * @return true if the event was stored
     */
    public boolean event(String s, float[] v){
        return event(new PMLEvent(s, frame_count, v));
    }
    /**
     * Load the contents of the record buffer into the playback buffer
     * @return false if the record buffer is empty
     */
    public boolean readBuffer(){
        if(playback) stopPlayback();
        if(sequence_r.isEmpty()){
            printerr("record buffer is empty.");
            return false;
        }
        sequence_p = new PMLEvent[sequence_r.size()];
        sequence_r.toArray(sequence_p);
        Arrays.sort(sequence_p);
        println("sequence loaded with " + Integer.toString(sequence_p.length) + " events.");
        return true;
    }
    /**
     * Load a sequence from the file specified
     * @param f sequence file
     * @return true if the sequence was successfully loaded
     */
    public boolean read(String f){
        try {
            return read(new File(f));
        }
        catch (NullPointerException e){
            printerr("sequence read failed - NullPointerException.");
            return false;
        }
    }
    /**
     * Load a sequence from the file specified
     * @param f sequence file
     * @return true if the sequence was successfully loaded
     */
    public boolean read(File f){
        if(playback) stopPlayback();
        sequence_p = readSequence(f);
        if(sequence_p == null){
            printerr("sequence " + f.getName() + " read failed.");
            return false;
        }
        else if(sequence_p.length < 1){
            printerr("sequence " + f.getName() + " is empty.");
            return false;
        }
        println("sequence " + f + " read with " + Integer.toString(sequence_p.length) + " events.");
        return true;
    }
    /**
     * Return true if a sequence has been loaded
     * @return true if a sequence has been loaded
     */
    public boolean sequenceLoaded(){
        if(sequence_p != null) return true;
        return false;
    }
    /**
     * Toggle sequencer playback
     */
    public void play(){
        if(playback)
            stopPlayback();
        else if(sequenceLoaded()){
            playback = true;
            println("sequence playback started.");
        }
        else
            printerr("playback buffer is empty.");
    }
    /**
     * Toggle sequencer recording
     */
    public void record(){
        if(record)
            stopRecording();
        else {
            record = true;
            sequence_r = new ArrayList(1024);
            println("sequence recording started.");
        }
    }
    /**
     * Stop sequencer recording/playback
     */
    public void stop(){
        stopPlayback();
        stopRecording();
    }
    /**
     * Stop sequencer playback
     */
    private void stopPlayback(){
        if(!playback) return;
        println("sequence playback stopped - " + event_count + " events processed.");
        frame_count = 0;
        event_count = 0;
        playback = false;
    }
    /**
     * Stop sequence recording
     */
    private void stopRecording(){
        if(!record) return;
        println("sequence recording stopped - " + sequence_r.size() + " events recorded.");
        frame_count = 0;
        record = false;
    }
    /**
     * Return true if the sequencer is active
     * @return true if the sequencer is active;
     */
    public boolean isActive(){
        if(playback || record) return true;
        return false;
    }
    /**
     * Return true if the sequencer is playing
     * @return true if the sequencer is playing
     */
    public boolean isPlaying(){
        return playback;
    }
    /**
     * Return true if the sequencer is recording
     * @return true if the sequencer is recording
     */
    public boolean isRecording(){
        return record;
    }
    /**
     * Write the recorded sequence to the specified file
     * @param f file
     * @return true if the file was successfully written
     */
    public boolean write(String f){
        try {
            return write(new File(f));
        }
        catch (NullPointerException e){
            printerr("sequence write failed - NullPointerException");
            return false;
        }
    }
    /**
     * Write the recorded sequence to the specified file
     * @param f file
     * @return true if the file was successfully written
     */
    public boolean write(File f){
        if(sequence_r.size() <= 0){
            printerr("sequence write failed - sequence is empty.");
            return false;
        }
        if(record) stopRecording();
        PMLEvent[] se_list = new PMLEvent[sequence_r.size()];
        sequence_r.toArray(se_list);
        Arrays.sort(se_list);
        String[] se_strs = new String[se_list.length];
        for(int i = 0; i < se_list.length; i ++)
            se_strs[i] = se_list[i].toString();
        if(PMLTools.writeStrings(se_strs, f)){
            println("sequence " + f.getName() + " written.");
            return true;
        }
        return false;       
    }
    /**
     * Set the input matrix
     * <p>If the sequencer is recording and the matrix is not null then it will
     * be sampled on calling frame(). If output values within the matrix have changed
     * then they will be stored in the current sequence.
     * <p>If ramping is enabled then the ramped value will be stored, resulting
     * in more constant value changes. As such, it is advisable to disable matrix
     * value ramping within this context.
     * @param m matrix
     */
    public void setInputMatrix(PMLMatrix m){
        matrix_in = m;
    }
    /**
     * Set the matrix sampling interval
     * <p>The matrix sampling interval specifies how regularly the input matrix
     * should be sampled. A value of 1 would result in sampling every frame,
     * whilst a value of 10 would result in sampling every 10 frames.
     * <p>Default: 1
     * @param i interval
     */
    public void setMatrixInterval(int i){
        interval = i > 0 ? i : 1;
    }
    /**
     * Set the output matrix
     * <p>If the sequencer is playing and the matrix is not null, then sequencer
     * events will be routed to the appropriate matrix input upon calling frame().
     * @param m matrix
     */
    public void setOutputMatrix(PMLMatrix m){
        matrix_out = m;
    }
    /**
     * Set the sequencer event handler
     * <p>If the sequencer is playing and the event handler is not null, then
     * sequencer events will be routed to the object methods specified within
     * the event handler upon calling frame().
     * @param h event handler
     */
    public void setEventHandler(PMLEventHandler h){
        eventHandler = h;
    }
    /**
     * Set whether this sequencer is verbose
     * <p>Default: false
     * @param b boolean
     */
    public void setVerbose(boolean b){
        verbose = b;
    }
    /**
     * Return the loaded sequence as an array of PMLEvent objects
     * <p>If there is no loaded sequence an empty array will be returned
     * @return PMLEvent array
     */
    public PMLEvent[] toArray(){
        if(sequence_p == null) return new PMLEvent[] {};
        PMLEvent[] seq_a = new PMLEvent[sequence_p.length];
        for(int i = 0; i < sequence_p.length; i ++)
            seq_a[i] = sequence_p[i].copy();
        return seq_a;
    }
    /**
     * Returns all instances of the specified event tag as an array of PMLEvent
     * objects
     * <p>If there is no loaded sequence or there are no instances of the specified tag
     * an empty array will be returned
     * @param t tag
     * @return PMLEvent array
     */
    public PMLEvent[] toArray(String t){
        if(t == null || t.isEmpty()) return toArray();
        if(sequence_p == null) return new PMLEvent[] {};
        ArrayList seq_l = new ArrayList(sequence_p.length);
        for(int i = 0; i < sequence_p.length; i ++)
            if(sequence_p[i].hasTag(t)) seq_l.add(sequence_p[i].copy());
        PMLEvent[] seq_a = new PMLEvent[seq_l.size()];
        seq_l.toArray(seq_a);
        Arrays.sort(seq_a);
        return seq_a;
    }
    private void println(String s){
        if(verbose) System.out.println(tag + ": " + s);
    }
    private void printerr(String s){    
        System.err.println(tag + ": " + s);
    }
    
    /***************    Static Functions ***************/
    
    /**
     * Load the specified sequence and return an array of PMLEvent objects
     * <p>Sequence entries that cannot be parsed will be ignored and the returned
     * array will be sorted
     * @param f sequence file
     * @return an array of PMLEvent objects
     */
    public static PMLEvent[] readSequence(String f){
        return readSequence(new File(f));
    }
    /**
     * Load the specified sequence and return an array of PMLEvent objects
     * <p>Sequence entries that cannot be parsed will be ignored and the returned
     * array will be sorted. If the specified file cannot be read or is empty
     * a null value will be returned.
     * @param f sequence file
     * @return an array of PMLEvent objects
     */
    public static PMLEvent[] readSequence(File f){
        String[] str = PMLTools.readStrings(f);
        if(str == null) return null;
        ArrayList lst = new ArrayList(str.length);
        PMLEvent e;
        for(String s : str){
            e = PMLEvent.fromString(s);
            if(e != null) lst.add(e);
        }
        if(lst.isEmpty()) return null;
        PMLEvent[] arr = new PMLEvent[lst.size()];
        lst.toArray(arr);
        Arrays.sort(arr);
        return arr;
    }
    /**
     * Reverse the specified sequence
     * @param s input sequence
     * @return true if the sequence was successfully modified
     */
    public static boolean reverseSequence(String s){
        return reverseSequence(s, s);
    }
    /**
     * Reverse the specified sequence and store the result in the chosen file
     * @param s0 input sequence
     * @param s1 output sequence
     * @return true if the sequence was successfully written
     */
    public static boolean reverseSequence(String s0, String s1){
        PMLEvent[] evi = readSequence(s0);
        if(evi == null || evi.length < 2) return false;
        // Get the start and end frame values - these should be reliable as
        // loadSequence() checks and sorts all loaded events.
        int sf = evi[0].frame();
        int ef = evi[evi.length - 1].frame();
        // Reverse the sequence preserving start and end frames
        PMLEvent[] evo = new PMLEvent[evi.length];
        for(int i = 0; i < evo.length; i ++)
            evo[i] = new PMLEvent(evi[i].tag(), ef - evi[i].frame() + sf, evi[i].vars());
        Arrays.sort(evo);
        // Write the sequence to the output file
        String[] str = new String[evo.length];
        for(int i = 0; i < evo.length; i ++)
            str[i] = evo[i].toString();
        return PMLTools.writeStrings(str, s1);
    }
    /**
     * Offset each event frame in the specified sequence by the denoted amount
     * @param s input sequence
     * @param v frame offset
     * @return true if the sequence was successfully modified
     */
    public static boolean offsetSequence(String s, int v){
        return offsetSequence(s, s, v);
    }
    /**
     * Offset each event frame in the specified sequence by the denoted amount and
     * store the result in the chosen output file
     * @param s0 input sequence
     * @param s1 output sequence
     * @param v frame offset
     * @return true if the sequence was successfully written
     */
    public static boolean offsetSequence(String s0, String s1, int v){
        // Load the sequence using readSequence() to ensure it is valid
        PMLEvent[] evi = readSequence(s0);
        if(evi == null || evi.length < 1) return false;
        PMLEvent[] evo = new PMLEvent[evi.length];
        // Apply the frame offset
        int f;
        for(int i = 0; i < evi.length; i ++){
            f = evi[i].frame() + v > 0 ? evi[i].frame() + v : 0;
            evo[i] = new PMLEvent(evi[i].tag(), f, evi[i].vars());
        }
        Arrays.sort(evo);
        // Write the modified sequence
        String[] str = new String[evo.length];
        for(int i = 0; i < evo.length; i ++)
            str[i] = evo[i].toString();
        return PMLTools.writeStrings(str, s1);
    }
    /**
     * Apply a random frame offset to each event within the specified sequence
     * <p>The offset value will be within the range specified by +/- r.
     * @param s input sequence
     * @param r offset range
     * @return true if the sequence was successfully modified
     */
    public static boolean randomOffsetSequence(String s, int r){
        return randomOffsetSequence(s, s, r);
    }
    /**
     * Apply a random frame offset to each event within the specified input
     * sequence and store the result to the specified output
     * <p>The offset value will be within the range specified by +/- r.
     * @param s0 input sequence
     * @param s1 output sequence
     * @param r offset range
     * @return true if the sequence was successfully written
     */
    public static boolean randomOffsetSequence(String s0, String s1, int r){
        // Load the sequence using readSequence() to ensure it is valid
        PMLEvent[] evi = readSequence(s0);
        if(evi == null || evi.length < 1) return false;
        PMLEvent[] evo = new PMLEvent[evi.length];
        // Apply the frame offset        
        int f;
        for(int i = 0; i < evi.length; i ++){
            f = evi[i].frame() + (int) (Math.random() * 2d * (double) r) - r;
            if(f < 0) f = 0;
            evo[i] = new PMLEvent(evi[i].tag(), f, evi[i].vars());
        }
        Arrays.sort(evo);
        // Write the modified sequence
        String[] str = new String[evo.length];
        for(int i = 0; i < evo.length; i ++)
            str[i] = evo[i].toString();
        return PMLTools.writeStrings(str, s1);
    }
    /**
     * Combine the specified sequences and store the result in the first
     * @param s0 input sequence 1
     * @param s1 input sequence 2
     * @return true if the sequence was successfully modified
     */
    public static boolean combineSequences(String s0, String s1){
        return combineSequences(s0, s1, s0);
    }
    /**
     * Combine the specified sequences and store the result in the specified
     * output file
     * @param s0 input sequence 1
     * @param s1 input sequence 2
     * @param s2 output sequence
     * @return true if the sequence was successfully written
     */
    public static boolean combineSequences(String s0, String s1, String s2){
        // Load the sequences using loadSequence() to ensure they are valid
        PMLEvent[] ev1 = readSequence(s0);
        PMLEvent[] ev2 = readSequence(s1);
        if(ev1 == null || ev2 == null) return false;
        // Combine and sort the sequences
        PMLEvent[] evo = new PMLEvent[ev1.length + ev2.length];
        System.arraycopy(ev1, 0, evo, 0, ev1.length);
        System.arraycopy(ev2, 0, evo, ev1.length, ev2.length);
        Arrays.sort(evo);
        // Write the output sequence
        String[] str = new String[evo.length];
        for(int i = 0; i < evo.length; i ++)
            str[i] = evo[i].toString();
        return PMLTools.writeStrings(str, s2);
    }
    /**
     * Read the specified sequence and store the data in an HTML table
     * @param s sequence
     * @param f HTML file
     * @return true if the file was successfully written
     */
    public static boolean sequenceToHtml(String s, String f){
        // Load the sequence using readSequence() to ensure it is valid
        PMLEvent[] evi = readSequence(s);
        if(evi == null || evi.length < 1) return false;
        File sf = new File(s);
        // Get maximum variables
        int mv = 0;
        for(PMLEvent e : evi) if(e.vars().length > mv) mv = e.vars().length;
        // Generate html
        StringBuilder sb = new StringBuilder();
        sb.append("<html>\n<body>\n");
        sb.append("<h2>");
        sb.append(sf.getName());
        sb.append("</h2>\n");
        sb.append("<table cellpadding=\"3\">\n");
        // Headers
        sb.append("<tr>\n<th>Frame</th>\n<th>Tag</th>\n");
        for(int i = 0; i < mv; i ++){
            sb.append("<th>Var ");
            sb.append(i);
            sb.append("</th>\n");
        }
        sb.append("</tr>\n");
        // Data
        for(PMLEvent e : evi){
            sb.append("<tr>\n");
            sb.append("<td>");
            sb.append(e.frame());
            sb.append("</td>\n");
            sb.append("<td>");
            sb.append(e.tag());         
            sb.append("</td>\n");
            for(int i = 0; i < mv; i ++){
                sb.append("<td>");
                if(i < e.vars().length)
                    sb.append(e.vars()[i]);
                sb.append("</td>\n");                
            }
            sb.append("</tr>\n");            
        }
        sb.append("</table>\n</body>\n</html>");
        // Write html
        return PMLTools.writeStrings(new String[] { sb.toString() }, f);
    }
}
