package pslib;
/**
 * AttractorForce Class
 * <p> The AttractorForce object extends the PSForce class allowing an attractive 
 * force to be applied to a DynamicParticle object.
 * <p>If no radius is specified the force magnitude increases, the further a
 * particle gets from the force position. If a radius is specified, then force
 * magnitude is calculated based on the position of the particle within the force
 * radius, with the centre creating the largest magnitude.
 * <p>This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * <p>You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see http://www.gnu.org/licenses/.
 * @author Stephen Callear
 */
public class AttractiveForce extends PSForce {
    private PSVector position;
    private float damping, curve;
    private Float radius;
    /**
     * Create a static object instance with the specified values
     * @param p position
     * @param m magnitude
     */
    public AttractiveForce(PSVector p, float m){
        this(p, m, m, STATIC, 1f);
    }
    /**
     * Create a static object instance with the specified values
     * @param p position
     * @param r radius
     * @param m magnitude
     */
    public AttractiveForce(PSVector p, float r, float m){
        this(p, m, m, STATIC, 1f);
        radius = new Float(r);
    }
    /**
     * Create a dynamic object instance with the specified values
     * @param p position
     * @param r radius
     * @param mi minimum magnitude
     * @param ma maximum magnitude
     * @param m mode
     * @param c decay curve exponent
     */
    public AttractiveForce(PSVector p, float r, float mi, float ma, int m, float c){
        this(p, mi, ma, m, c);
        radius = new Float(r);
    }
    /**
     * Create a dynamic object instance with the specified values
     * @param p position
     * @param mi minimum magnitude
     * @param ma maximum magnitude
     * @param m mode
     * @param c decay curve exponent

     */
    public AttractiveForce(PSVector p, float mi, float ma, int m, float c){
        super(mi, ma, m, c);
        position = p;
        damping = 0.01f;
        curve = 1.0f;
        radius = null;
    }
    /**
     * Set the force position
     * @param p position
     */
    public void setPosition(PSVector p){
        position = p;
    }
    /**
     * Return the force position
     * @return the force position
     */
    public PSVector getPosition(){
        return position;
    }
    /**
     * Set force damping for non-radial behaviour
     * <p>
     * Default: 0.01
     * @param d damping
     */
    public void setDamping(float d){
        damping = d;
    }
    /**
     * Set the distance curve exponent
     * @param c curve exponent
     */
    public void setDistanceCurve(float c){
        curve = c;
    }
    /**
     * Update the specified particle
     * @param p particle
     */
    public void update(DynamicParticle p){
        PSVector vel = p.getVelocity(); // Get the particle velocity vector assuming a copy
        PSVector vec = getForceVector(p.getPosition());
        vec.scale(getMagnitude(p));
        vel.add(vec);
        p.setVelocity(vel); // Re-apply velocity vector assuming a copy
    }
    private PSVector getForceVector(PSVector p){
        PSVector fv = new PSVector();
        float d = position.distance(p);
        if(d == 0) return fv;
        double mag;
        if(radius != null){ // Radial attractors get stronger as things get closer
            if(d > radius) return fv;
            mag = Math.pow(1.0 - (double) (d / radius), (double) curve);
        } else { // Non-radial attractors get weaker as things get closer
            mag = Math.pow((double) (d * damping), (double) curve);
        }
        fv = PSVector.sub(position, p);
        fv.normalise();
        fv.scale((float) mag);
        return fv;
    }
}

