package pslib;
/**
 * AudioParticle Class
 * <p>The AudioParticle object extends the functionality of the DynamicParticle class
 * by incorporating spatial audio calculation.
 * <p>This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * <p>You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see http://www.gnu.org/licenses/.
 * @author Stephen Callear
 */
public class AudioParticle extends DynamicParticle {
    private float volume, radius_min, radius_max;
    /**
     * Create a static object instance
     */
    public AudioParticle(){
        this(0);
    }
    /**
     * Create a dynamic object instance with the specified decay time
     * @param d decay (ms)
     */
    public AudioParticle(int d){
        super(d);
        volume = 1.0f;
        radius_min = 0.0f;
        radius_max = 100.0f;
    }
    /**
     * Set the minimum and maximum audio radii
     * <p>If the listener position is within the minimum radius, the particle volume
     * will be 1, while if the listener position is outside the maxium radius, the
     * particle volume will be 0.
     * @param min minimum radius
     * @param max maximum radius
     */
    public void setLimits(float min, float max){
        radius_min = min;
        radius_max = max;
    }
    /**
     * Set the particle volume scalar
     * @param v volume
     */
    public void setVolume(float v){
        volume = v;
    }
    /**
     * Return the volume value relative to the specified listener position
     * @param lpv listener position vector
     * @return the volume value relative to the specified listener position
     */
    public final float getRelativeVolume(PSVector lpv){
        float dl, dn, v;
        dl = PSTools.constrainFloat(lpv.distance(getPosition()), radius_min, radius_max);
        dn = 1.0f - (dl - radius_min) / (radius_max - radius_min);
        v = (float) Math.pow((double) dn, 2.0) * volume;
        return v;
    }
    /**
     * Return the left/right pan value relative to the specified listener position
     * @param lpv listener position vector
     * @param llv listener look vector
     * @param luv listener up vector
     * @return the left/right pan value relative to the specified listener position
     */
    public final float getRelativeLRPan(PSVector lpv, PSVector llv, PSVector luv){
        llv.normalise();
        PSVector lrv = luv.cross(llv); // Right vector
        PSVector pdv = PSVector.sub(getPosition(), lpv); // Difference vector
        lrv.normalise();
        pdv.normalise();
        float p = lrv.dot(pdv);
        return p;
    }
    /**
     * Return the front/back pan value relative to the specified listener position
     * @param lpv listener position vector
     * @param llv listener look vector
     * @return the front/back pan value relative to the specified listener position
     */
    public final float getRelativeFBPan(PSVector lpv, PSVector llv){
        PSVector pdv = PSVector.sub(getPosition(), lpv);
        llv.normalise();
        pdv.normalise();
        float p = llv.dot(pdv);
        return p;
     
    }
    /**
     * Return the up/down pan value relative to the specified listener position
     * @param lpv listener position vector
     * @param luv listener up vector
     * @return the up/down pan value relative to the specified listener position
     */
    public final float getRelativeUDPan(PSVector lpv, PSVector luv){
        PSVector pdv = PSVector.sub(getPosition(), lpv);
        luv.normalise();
        pdv.normalise();
        float p = luv.dot(pdv);
        return p;
    }
    /**
     * Return an array containing audio data relative to the specified listener
     * position
     * <p>Format: { volume, l/r pan, f/b pan, u/d pan }
     * @param lpv listener position vector
     * @param llv listener look vector
     * @param luv listener up vector
     * @return an array containing audio data relative to the specified listener
     * position
     */
    public final float[] getRelativeAudioData(PSVector lpv, PSVector llv, PSVector luv){
        float vol = getRelativeVolume(lpv);
        float lrp = getRelativeLRPan(lpv, llv, luv);
        float fbp = getRelativeFBPan(lpv, llv);
        float udp = getRelativeUDPan(lpv, luv);
        float[] ad = { vol, lrp, fbp, udp };
        return ad;
    }
    /**
     * Return a copy of this particle
     * <p>The returned copy will use the volume and limit values of the source
     * particle and will adhere to the rules defined in the DynamicParticle map()
     * function.
     * @see DynamicParticle
     * @return a copy of this particle
     */
    @Override public AudioParticle copy(){
        AudioParticle p = new AudioParticle();
        p.setVolume(volume);
        p.setLimits(radius_min, radius_max);
        super.map(p);
        return p;

    }

}
