package pslib;
/**
 * BoidParticle Class
 * <p>The BoidParticle object extends the functionality of the ParentParticle object by
 * incorporating flocking simulation. Boid position is calculated based on the
 * rules of separation, cohesion and alignment of flock members as defined by
 * Reynolds.
 * <p>This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * <p>You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see http://www.gnu.org/licenses/.
 * @author Stephen Callear
 */
public class BoidParticle extends ParentParticle {
    private float sep_m, ali_m, coh_m, btp_m, sep_d, chk_d, btp_r;
    private float max_velocity, max_force, damping;
    private PSVector btp_c;
    /**
     * Create a new static object instance with the default buffer size of 1
     */
    public BoidParticle(){
        this(1, 0);
    }
    /**
     * Create a new static object instance with the specified buffer size
     * @param b buffer size
     */
    public BoidParticle(int b){
        this(b, 0);
    }
    /**
     * Create a new dynamic object instance with the specified buffer size and
     * decay time
     * @param b buffer size
     * @param d decay (ms)
     */
    public BoidParticle(int b, int d){
        super(b, d);
        sep_m = 1.5f;
        ali_m = 1f;
        coh_m = 1f;
        btp_m = 2f;
        sep_d = 60.0f;
        chk_d = 60.0f;
        max_velocity = 5.0f;
        max_force = 0.1f;
        damping = 0.01f;
        btp_r = 0.0f;
    }
    /**
     * Set the boid force
     * @param s separation
     * @param a alignment
     * @param c cohesion
     * @param b bound-to-place
     */
    public void setForceScalars(float s, float a, float c, float b){
        sep_m = s;
        ali_m = a;
        coh_m = c;
        btp_m = b;
    }
    /**
     * Set the boid separation distance
     * <p> The separation distance is the distance which the boid will attempt
     * to maintain between itself and its neighbours.
     * <p>Default: 60.0
     * @param d separation distance
     */
    public void setSeparationDistance(float d){
        sep_d = d;
    }
    /**
     * Set the boid check distance
     * <p>Boid velocity is calculated based on all surrounding boids within the
     * specified radius.
     * <p>Default: 60.0
     * @param d check distance
     */
    public void setCheckDistance(float d){
        chk_d = d;
    }
    /**
     * Set the bound-to-place location and radius
     * <p>Boids will attempt to stay within the spherical area specifed. If the
     * central point is specified as null, the boid will not adhere to the
     * bound-to-place rule.
     * <p>Default: null, 0.0
     * @param c central point
     * @param r radius
     */
    public void setBoundToPlace(PSVector c, float r){
        btp_c = c;
        btp_r = r;
    }
    /**
     * Set the boid velocity damping amount
     * <p>Default: 0.01
     * @param d damping
     */
    public void setDamping(float d){
        damping = d;
    }
    /**
     * Set the maximum velocity for the boid
     * <p>Default: 5.0
     * @param v maximum velocity
     */
    public void setMaxVelocity(float v){
        max_velocity = v;
    }
    /**
     * Set the maximum acceleration force for the boid
     * <p>Default: 0.1
     * @param f maximum force
     */
    public void setMaxForce(float f){
        max_force = f;
    }
    /**
     * Update the boid
     * <p>Boid velocity and position is calculated based on the position and
     * velocity data for each GenericParticle within the specified array
     * @param pa GenericParticle array
     */
    public void update(GenericParticle[] pa){
        // Update behaviours
        super.update();
        // Accumulate boid behaviour forces
        PSVector acc = new PSVector();
        PSVector sep = PSVector.scale(separation(pa), sep_m);
        PSVector ali = PSVector.scale(alignment(pa), ali_m);
        PSVector coh = PSVector.scale(cohesion(pa), coh_m);
        PSVector btp = PSVector.scale(boundToPlace(), btp_m);

        acc.add(sep);
        acc.add(ali);
        acc.add(coh);
        acc.add(btp);
        // Apply accumulated vector
        PSVector vel = getVelocity();
        PSVector pos = getPosition();
        vel.add(acc);       
        vel.limit(max_velocity);
        pos.add(vel);
        // Reapply vectors to allow for any attached particles that rely on it
        setVelocity(vel);
        setPosition(pos);
    }
    /**
     * Return the separation force
     * @param pa GenericParticle array
     * @return the separation force
     */
    private PSVector separation(GenericParticle[] pa){
        PSVector v = new PSVector();
        int count = 0;
        for(int i = 0; i < pa.length; i ++){
            float d = getPosition().distance(pa[i].getPosition());
            if(d > 0 && d < sep_d){
                PSVector dif = PSVector.sub(getPosition(), pa[i].getPosition());
                dif.scale(1.0f / d);
                v.add(dif);
                count ++;
            }
        }
        if(count > 0) v.scale(1.0f / (float) count);
        if(v.magnitude() > 0){
            v.normalise();
            v.scale(max_velocity);
            v.sub(getVelocity());
            v.limit(max_force);
        }
        return v;
    }
    /**
     * Return the alignment force
     * @param pa GenericParticle array
     * @return the alignment force
     */
    private PSVector alignment(GenericParticle[] pa){
        PSVector v = new PSVector();
        int count = 0;
        for(int i = 0; i < pa.length; i ++){
            float d = getPosition().distance(pa[i].getPosition());
            if(d > 0 && d < chk_d){
                v.add(pa[i].getVelocity());
                count ++;
            }
        }
        if(count > 0) v.scale(1.0f / (float) count);
        if(v.magnitude() > 0){
            v.normalise();
            v.scale(max_velocity);
            v.sub(getVelocity());
            v.limit(max_force);
        }
        return v;
    }
    /**
     * Return the cohesion force
     * @param pa GenericParticle array
     * @return the cohesion force
     */
    private PSVector cohesion(GenericParticle[] boids){
        PSVector v = new PSVector();
        int count = 0;
        for(int i = 0; i < boids.length; i ++){
            float d = getPosition().distance(boids[i].getPosition());
            if(d > 0 && d < chk_d){
                v.add(boids[i].getPosition());
                count ++;
            }
        }
        if(count > 0){
            v.scale(1.0f / (float) count);
            return tendToPlace(v, false);
        }
        return v;
    }
    /**
     * Return the tend-to-place force for the specified target
     * @param target target location
     * @param slowdown slowdown toggle
     * @return the tend-to-place force for the specified target
     */
    private PSVector tendToPlace(PSVector target, boolean slowdown){
        PSVector v;
        PSVector dif = PSVector.sub(target, getPosition());
        float d = dif.magnitude();
        if(d > 0){
            dif.normalise();
            if(slowdown && d < chk_d)
                dif.scale(max_velocity * d * damping);
            else
                dif.scale(max_velocity);
            v = PSVector.sub(dif, getVelocity());
            v.limit(max_force);
        } else {
            v = new PSVector();
        }
        return v;  
    }
    /**
     * Return the bound-to-place force
     * @return the bound-to-place force
     */
    private PSVector boundToPlace(){
        PSVector v = new PSVector();
        if(btp_c == null) return v;
            float d = getPosition().distance(btp_c);
        if(d >= btp_r)
            return tendToPlace(btp_c, false);
        return v;
    }
    /**
     * Return a copy of this particle
     * <p>The returned particle will copy of all of this objects variables and will
     * adhere to the rules defined in <code>DynamicParticle.map()</code>.
     * @see DynamicParticle
     * @return a copy of this particle
     */    
    @Override public BoidParticle copy(){        
        BoidParticle p = new BoidParticle(getBufferSize(), getInitialDecayTime());
        map(p);
        return p;    
    }
    /**
     * Map the properties of this particle to that specified preserving references
     * <p>The returned object takes a copy of all orbiter particle variables.
     * DynamicParticle variables are mapped based on the rules defined in <code>
     * DynamicParticle.map()</code>.
     * @see DynamicParticle
     * @param p particle
     */
    public final void map(BoidParticle p){
        super.map(p);
        p.setBoundToPlace(btp_c, btp_r);
        p.setCheckDistance(chk_d);
        p.setDamping(damping);
        p.setForceScalars(sep_m, ali_m, coh_m, btp_m);
        p.setMaxForce(max_force);
        p.setMaxVelocity(max_velocity);
        p.setSeparationDistance(sep_d);
    }    
}
