package pslib;
/**
 * BufferedObject Class
 * <p>The BufferedObject object allows input objects to be buffered. The buffer will
 * dynamically restructure itself to preserve ordering.
 * <p>This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * <p>You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see http://www.gnu.org/licenses/.
 * @author Stephen Callear
 */
public class BufferedObject extends Object {
    private Object[] arr;
    private boolean initialised;
    /**
     * Create a new object instance with the specified buffer size
     * @param s buffer size
     */
    public BufferedObject(int s){
        arr = new Object[s];
        initialised = false;
    }
    /**
     * Store the specified object reference
     * @param obj object
     */
    public final void set(Object obj){
        if(initialised) addBufferElement(obj);
        else initBuffer(obj);
    }
    /**
     * Return the most recently stored object reference
     * @return the most recently stored object reference
     */
    public final Object get(){
        return get(0);
    }
    /**
     * Return the object reference stored at the specified index
     * @param i index
     * @return the object reference stored at the specified index
     */
    public final Object get(int i){
        if(initialised && i >= 0 && i < arr.length) return arr[i];
        return null;
    }
    /**
     * Store the contents of the buffer in the specified array
     * <p>
     * If the input array size differs from the buffer size, values will be removed
     * or duplicated as required.
     * @param arr_n object array
     * @return the modified object array
     */
    public final Object[] toArray(Object[] arr_n){
        if(!initialised) return arr_n;
        int i = 0;
        while(i < arr_n.length){
            if(i < arr.length)  arr_n[i] = arr[i];
            else arr_n[i] = arr[arr.length - 1];
            i ++;
        }
        return arr_n;
    }
    /**
     * Return the current buffer size
     * @return the current buffer size
     */
    public final int size(){
        return arr.length;
    }
    /**
     * Resize the buffer
     * @param s size
     */
    public final void resize(int s){
        Object[] arr_n = new Object[s];
        toArray(arr_n);
        arr = arr_n;
    }
    /**
     * Clear the buffer contents
     */
    public final void clear(){
        for(int i = 0; i < arr.length; i ++)
            arr[i] = null;
        initialised = false;
    }
    /**
     * Return true if the buffer has been initialised
     * @return true if the buffer has been initialised
     */
    public final boolean isInitialised(){
        return initialised;
    }
    private void addBufferElement(Object obj){
        Object[] arr_n = new Object[arr.length];
        arr_n[0] = obj;
        for(int i = 1; i < arr.length; i ++)
            arr_n[i] = arr[i - 1];
        arr = arr_n;
    }
    private void initBuffer(Object obj){
        for(int i = 0; i < arr.length; i ++)
            arr[i] = obj;
        if(!initialised) initialised = true;
    }

}
