package pslib;
/**
 * BufferedParticle Class
 * <p>The BufferedParticle object extends the functionality of the DynamicParticle class
 * by incorporating position, colour, decay and size buffers.
 * <p>This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * <p>You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see http://www.gnu.org/licenses/.
 * @author Stephen Callear
 */
public class BufferedParticle extends DynamicParticle{
    private StaticParticle[] buffer;
    private float spacing;
    private boolean initialised;
    /**
     * Create a new object instance with the specified buffer size
     * @param b buffer size
     */
    public BufferedParticle(int b){
        this(b, 0);
    }
    /**
     * Create a new object instance with the specified buffer size and decay
     * @param b buffer size
     * @param d decay (ms)
     */
    public BufferedParticle(int b, int d){
        super(d);
        buffer = new StaticParticle[b];
        initialised = false;
    }
    /**
     * Set the current particle position and update the buffer
     * @param p position
     */
    @Override public void setPosition(PSVector p){
        super.setPosition(p);
        if(initialised){
            if(p.distance(buffer[0].position) > spacing) addBufferElement(getSnapshot());
        } else initialiseBuffer(getSnapshot());
    }
    /**
     * Set the minimum buffered element spacing
     * <p>
     * Default: 0
     * @param s spacing
     */
    public void setSpacing(float s){
        spacing = s;
    }
    /**
     * Return true if the buffer is initialised
     * @return true if the buffer is initialised
     */
    public final boolean isInitialised(){
        return initialised;
    }
    /**
     * Initialise the buffer with the specified StaticParticle object
     * @param p StaticParticle
     */
    public final void initialiseBuffer(StaticParticle p){
        for(int i = 0; i < buffer.length; i ++)
            buffer[i] = p;
        if(!initialised) initialised = true;
    }
    /**
     * Return an array containing all buffered particles
     * @return a StaticParticle array containing all buffered particles
     */
    protected final StaticParticle[] getBuffer(){
        return buffer;
    }
    /**
     * Return the current buffer size
     * @return the current buffer size
     */
    public final int getBufferSize(){
        return buffer.length;
    }
    /**
     * Return an array containing the positions of each buffered element relative
     * to the current particle position
     * @return an array containing the positions of each buffered element relative
     * to the current particle position
     */
    protected final PSVector[] getRelativePositions(){
        PSVector[] rp = new PSVector[buffer.length];
        for(int i = 0; i < rp.length; i ++){
            rp[i] = PSVector.sub(buffer[i].position, getPosition());
        }
        return rp;
    }
    /**
     * Return an array containing the positions of each buffered element relative
     * to the preceding element
     * @return an array containing the positions of each buffered element relative
     * to the preceding element
     */
    protected final PSVector[] getCumulativePositions(){     
        PSVector[] cp = new PSVector[buffer.length];
        cp[0] = PSVector.sub(buffer[0].position, getPosition());
        for(int i = 1; i < cp.length; i ++)
            cp[i] = PSVector.sub(buffer[i].position, buffer[i - 1].position);
        return cp;
    }
    /**
     * Return the cumulative length of the buffer trail
     * <p>The cumulative length is calculated by summing the distance between
     * each buffer element and that preceding it.
     * @return the cumulative length of the buffer trail
     */
    protected final float getCumulativeLength(){
        float len = 0f;
        for(int i = 1; i < buffer.length; i ++)
            len += buffer[i].position.distance(buffer[i - 1].position);
        return len;
    }
    /**
     * Return an array of buffer taper values
     * <p>The returned array contains a taper value for each element within the
     * buffer. Taper values are based on the cumulative length of the buffer
     * trail and the relative position of each element within the trail. Values
     * range between 1 for the first buffer element and 0 for the last.
     * @return an array of buffer taper values
     */
    protected final float[] getTaperValues(){
        float len = 0f;
        float[] tv = new float[buffer.length];
        tv[0] = 0f;
        for(int i = 1; i < buffer.length; i ++){
            len += buffer[i].position.distance(buffer[i - 1].position);
            tv[i] = len;
        }
        for(int i = 0; i < tv.length; i ++)
            tv[i] = 1f - tv[i] / len;
        return tv;
    }        
    /**
     * Resize the buffer
     * @param s buffer size
     */
    public final void resizeBuffer(int s){
        StaticParticle[] bn = new StaticParticle[s];
        for(int i = 0; i < bn.length; i ++){
            if(i < buffer.length)
                bn[i] = buffer[i];
            else
                bn[i] = buffer[buffer.length - 1];
        }
        buffer = bn;
        bn = null;
    }
    /**
     * Return a copy of this particle
     * <p>The returned contain an empty buffer the same size as that of the source
     * particle and will adhere to the rules defined in <code>DynamicParticle.map()
     * </code>.     
     * @see DynamicParticle
     * @return a copy of this particle
     */
    @Override public BufferedParticle copy(){
        BufferedParticle p = new BufferedParticle(buffer.length);
        map(p);
        return p;
    }
    /**
     * Map the properties of this particle to that specified preserving references
     * <p>The returned object takes a copy of all BufferedParticle variables.
     * DynamicParticle variables are mapped based on the rules defined in
     * <code>DynamicParticle.map()</code>.
     * @see DynamicParticle
     * @param p particle
     */
    public final void map(BufferedParticle p){
        if(p.getBufferSize() != getBufferSize())
            p.resizeBuffer(getBufferSize());
        p.setSpacing(spacing);
        super.map(p);
    }
    /**
     * Add an element to the buffer
     * @param p StaticParticle
     */
    private void addBufferElement(StaticParticle p){
        StaticParticle[] bn = new StaticParticle[buffer.length];
        bn[0] = p;
        for(int i = 1; i < bn.length; i ++)
            bn[i] = buffer[i - 1];
        buffer = bn;
        bn = null;
    }
}
