package pslib;
import processing.opengl.PGraphicsOpenGL;
/**
 * DynamicParticle Class
 * <p>The DynamicParticle object extends the functionality of the GenericParticle
 * class by permitting dynamic behaviour over a particle life-span defined within
 * the constructor.
 * <p>Envelopes may be assigned to both particle alpha and size parameters
 * allowing the modulation of variables dependent on the current particle decay.
 * Similarly PSBehaviours may be attached to the particle, permitting per-frame
 * modulation of parameters.
 * <p>The results of alpha and size modulation envelopes may be obtained through
 * the getRenderAlpha() and getRenderSize() functions. Calls to setAlpha() and
 * setSize() will allow modification of the base value for the modulation envelope,
 * while calls to getAlpha() and getSize() will return this base value.
 * <p>This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * <p>You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see http://www.gnu.org/licenses/.
 * @author Stephen Callear
 */
public class DynamicParticle extends GenericParticle implements PSConstants {
    private PSCurve decay;
    private PSRange alpha_env, size_env;
    private PSBehaviour[] behaviour;
    private float mass;
    private int init_decay;    
    private boolean dynamic;
    /**
     * Create a new static object instance
     */
    public DynamicParticle(){
        this(0);
    }    
    /**
     * Create a new dynamic object instance with the specified decay time
     * @param d decay time (ms)
     */
    public DynamicParticle(int d){
        super();
        decay = new PSCurve(1f);
        init_decay = d;
        dynamic = false;
        if(d > 0){
            decay.set(0f, d);
            dynamic = true;
        }
        mass = 1f;
        alpha_env = new PSRange(1f, 1f);
        size_env = new PSRange(1f, 1f);
        behaviour = null;
    }
    /**
     * Set the particle decay time
     * <p>Calls to the function will reset the current particle decay curve and
     * begin a new curve of the duration specified.
     * @param t decay time (ms)
     */
    public void setDecayTime(int t){
        init_decay = t;
        decay.set(1f);
        dynamic = false;
        if(t > 0){
            decay.set(0f, t);
            dynamic = true;
        }
    }
    /**
     * Return the current decay value between 1 and 0
     * @return the current decay value
     */
    public float getDecay(){
        return decay.get();
    }
    /**
     * Return the remaining decay time in milliseconds
     */
    public int getDecayTime(){
        return Math.round(decay.get() * init_decay);
    }
    /**
     * Return the decay time specified on particle creation in milliseconds
     * @return the decay time specified on particle creation
     */
    public int getInitialDecayTime(){
        return init_decay;
    }
    /**
     * Return true if the particle is dynamic and is therefore decaying
     * @return true if the particle is dynamic
     */
    public boolean isDynamic(){
        return dynamic;
    }
    /**
     * Return true if the particle is dead
     * @return true if the particle is dead
     */
    public boolean isDead(){
        return decay.get() > 0f ? false : true;        
    }
    /**
     * Set the particle mass
     * <p>Default: 1.0
     * @param m mass
     */
    public void setMass(float m){
        mass = m;
    }
    /**
     * Return the particle mass
     * @return the particle mass
     */
    public float getMass(){
        return mass;
    }
    /**
     * Set the particle behaviour
     * @param b behaviour
     */
    public void setBehaviour(PSBehaviour b){
        behaviour = new PSBehaviour[] { b };
    }
    /**
     * Set the particle behaviour
     * <p>The specified behaviours will be applied in the order in which the exist
     * within the array
     * @param b behaviour
     */
    public void setBehaviour(PSBehaviour[] b){
        behaviour = b;
    }
    /**
     * Return the attached particle behaviour array
     * @return the attached particle behaviour array
     */
    protected PSBehaviour[] getBehaviour(){
        return behaviour;
    }
    /**
     * Clear the particle behaviours
     */
    public void clearBehaviour(){
        behaviour = null;
    }
    /**
     * Set the particle alpha modulation envelope for dynamic particles
     * <p>The alpha modulation value is a scalar applied to the particle alpha as
     * specified with setAlpha(). For example, a minimum value of 0.5 and a maximum
     * value of 2.0 with a LINEAR mode and a curve exponent of 1.0 will result in
     * a modulation of particle alpha between half and double its specified value
     * throughout the duration of the particle.
     * @param e envelope
     */
    public void setAlphaEnvelope(PSRange e){
        alpha_env = e;
    }
    /**
     * Return the modulated particle alpha value
     * @return the modulated particle alpha value
     */
    public float getRenderAlpha(){
        return getAlpha() * alpha_env.get(1f - decay.get());
    }
    /**
     * Set the particle size modulation envelope for dynamic particles
     * <p>The alpha modulation value is a scalar applied to the particle alpha as
     * specified with setSize(). For example, a minimum value of 0.5 and a maximum
     * value of 2.0 with a LINEAR mode and a curve exponent of 1.0 will result in
     * a modulation of particle size between half and double its specified value
     * throughout the duration of the particle.
     * @param e
     */
    public void setSizeEnvelope(PSRange e){
        size_env = e;
    }
    /**
     * Return the modulated particle size value
     * @return the modulated particle size value
     */
    public float getRenderSize(){
        return getSize() * size_env.get(1f - decay.get());
    }
    /**
     * Return a StaticParticle containing the current particle position, colour,
     * modulated alpha, modulated size and remaining decay time.
     * @return a StaticParticle containing the current particle parameters
     */
    public StaticParticle getSnapshot(){
        return new StaticParticle(getPosition().copy(), getColour().copy(), getRenderAlpha(),
                                  getRenderSize(), getDecayTime());
    }
    /**
     * Update the particle
     */
    public void update(){
        if(behaviour == null || behaviour.length < 1) return;
        for(int i = 0; i < behaviour.length; i ++)
            behaviour[i].update(this);
    }
    /**
     * Render the particle to the specified OpenGL renderer
     * @param pgl OpenGL renderer
     */
    @Override public void render(PGraphicsOpenGL pgl){
        render(pgl, null, null, null);
    }
    /**
     * Render the particle to the specified OpenGL renderer
     * @param pgl OpenGL renderer
     * @param cpv camera position vector
     * @param ctv camera target vector
     * @param cuv camera up vector
     */
    @Override public void render(PGraphicsOpenGL pgl, PSVector cpv, PSVector ctv, PSVector cuv){
        PSColour c = getColour();
        int alpha = Math.round((float) c.getAlpha() * getRenderAlpha());
        pgl.noStroke();
        pgl.fill(c.getRed(), c.getGreen(), c.getBlue(), alpha);
        pgl.pushMatrix();
        pgl.translate(getPosition().x, getPosition().y, getPosition().z);
        PSShapes.circle(pgl, 16, getRenderSize() * 0.5f, cpv, ctv, cuv);
        pgl.popMatrix();
    }
    /**
     * Render the particle location to the specified OpenGL renderer
     * <p>The debug method renders the particle as a GenericParticle object
     * discarding any dynamic modifications to particle properties.
     * @param pgl OpenGL renderer
     */
    public void debug(PGraphicsOpenGL pgl){
        debug(pgl, null, null, null);
    }
    /**
     * Render the particle location to the specified OpenGL renderer
     * <p>The debug method renders the particle as a GenericParticle object
     * discarding any dynamic modifications to particle properties.
     * @param pgl OpenGL renderer
     * @param cpv camera position vector
     * @param ctv camera target vector
     * @param cuv camera up vector
     */
    public void debug(PGraphicsOpenGL pgl, PSVector cpv, PSVector ctv, PSVector cuv){
        super.render(pgl, cpv, ctv, cuv);
    }
    /**
     * Return a copy of this DynamicParticle
     * <p>The returned object takes a copy of position, velocity, colour, alpha,
     * size and mass variables, while references to particle alpha envelope, size
     * envelope and attached behaviours are used. This decision was made to
     * facilitate particle templates, which would rarely require independent
     * envelopes or behaviours.
     * @return a copy of this DynamicParticle
     */
    public DynamicParticle copy(){
        DynamicParticle p = new DynamicParticle();
        this.map(p);
        return p;
    }
    /**
     * Map the properties of this particle to that specified preserving references
     * <p>The returned object takes a copy of position, velocity, colour, alpha,
     * size and mass variables, while references to particle alpha envelope, size
     * envelope and attached behaviours are used. This decision was made to
     * facilitate particle templates, which would rarely require independent
     * envelopes or behaviours.
     * @param p particle
     */
    public final void map(DynamicParticle p){
        // Copied components
        p.setDecayTime(init_decay);
        p.getPosition().set(getPosition());
        p.getVelocity().set(getVelocity());
        p.getColour().set(getColour().copy());
        p.setAlpha(getAlpha());
        p.setSize(getSize());
        p.setMass(getMass());
        // Referential components
        p.setAlphaEnvelope(alpha_env);
        p.setSizeEnvelope(size_env);
        p.setBehaviour(behaviour);
    }
}
