package pslib;
import java.util.Random;
/**
 * FlickerBehaviour Class
 * <p>The FlickerBehaviour object performs brownian deviation upon the size and
 * alpha components of the particle to which it is applied. Size and alpha deviation
 * may be linked or operate independently.
 * <p>This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * <p>You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see http://www.gnu.org/licenses/.
 * @author Stephen Callear
 */
public class FlickerBehaviour extends Object implements PSBehaviour {
    private Random random;
    private float alpha_min, alpha_max, alpha_step, size_min, size_max, size_step;
    private boolean link;
    /**
     * Create a new object instance
     * @param ami alpha minimum
     * @param ama alpha maximum
     * @param ast alpha step
     * @param smi size minimum
     * @param sma size maximum
     * @param sst size step
     */
    public FlickerBehaviour(float ami, float ama, float ast, float smi, float sma, float sst){
        alpha_min = ami;
        alpha_max = ama;
        alpha_step = ast;
        size_min = smi;
        size_max = sma;
        size_step = sst;
        random = new Random();
        link = true;
    }
    /**
     * Update the specified particle
     * @param p particle
     */
    public void update(DynamicParticle p){
        float rs;
        float r = random.nextFloat();
        // Alpha update
        rs = r * alpha_step * 2f - alpha_step;
        p.setAlpha(PSTools.constrainFloat(p.getAlpha() + rs, alpha_min, alpha_max));
        // Size update
        if(!link) r = random.nextFloat();
        rs = r * size_step * 2f - size_step;
        p.setSize(PSTools.constrainFloat(p.getSize() + rs, size_min, size_max));

    }
    /**
     * Set the alpha modulation variables
     * @param mi minimum
     * @param ma maximum
     * @param s step
     */
    public final void setAlpha(float mi, float ma, float s){
        alpha_min = mi;
        alpha_max = ma;
        alpha_step = s;
    }
    /**
     * Set the size modulation variables
     * @param mi minimum
     * @param ma maximum
     * @param s step
     */
    public final void setSize(float mi, float ma, float s){
        size_min = mi;
        size_max = ma;
        size_step = s;
    }
    /**
     * Set alpha and size deviation link
     * <p>Default: true
     * @param b link
     */
    public final void link(boolean b){
        link = b;
    }
}
