package pslib;
import processing.opengl.PGraphicsOpenGL;
import java.util.Random;
import java.util.Arrays;
import java.util.ArrayList;
/**
 * GenericEmitter Class
 * <p>The GenericEmitter class extends the functionality of the DynamicParticle
 * class by allowing further DynamicParticles to be attached. Attached particles
 * may then be subjected to a random initial velocity and any attached PSForce
 * objects or PSBehaviour objects.
 * <p>This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * <p>You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see http://www.gnu.org/licenses/.
 * @author Stephen Callear
 */
public class GenericEmitter extends DynamicParticle implements PSConstants {
    private DynamicParticle[] buffer;
    private PSForce[] forces;
    private PSVector direction;
    private PSVector dimensions;
    private float velocity_max, init_velocity_min, init_velocity_max,
            radius_min, radius_max, spread, speed;
    private int direction_mode;
    private boolean stealing;
    private Random random;
    /**
     * Create a new object instance with the specified buffer size
     * @param b buffer size
     */
    public GenericEmitter(int b){
        this(b, 0);
    }
    /**
     * Create a new object instance with the specified buffer size and decay time
     * @param b buffer size
     * @param d decay (ms)
     */
    public GenericEmitter(int b, int d){
        super(d);
        random = new Random();
        // Create lists
        buffer = new DynamicParticle[b];        
        // Set particle creation defaults
        velocity_max = 5f;
        init_velocity_min = 1f;
        init_velocity_max = 2f;
        radius_min = 0f;
        radius_max = 0f;
        spread = 1f;
        // Set direction defaults
        direction_mode = STATIC;
        direction = new PSVector(0f, -1f, 0f);
        // Set misc defaults
        speed = 1f;
        dimensions = new PSVector(1f, 1f, 1f);
        stealing = false;
    }
    /**
     * Set the colour of all attached particles
     * @param c colour
     */
    @Override public void setColour(PSColour c){
        super.setColour(c);
        for(DynamicParticle p : buffer)
            if(p != null && !p.isDead()) p.setColour(c);
    }
    /**
     * Set the colour of all attached particles with the specified fade time
     * @param c colour
     * @param t fade time (ms)
     */
    @Override public void setColour(PSColour c, int t){
        super.setColour(c, t);
        for(DynamicParticle p : buffer)
            if(p != null && !p.isDead()) p.setColour(c, t);
    }
    /**
     * Set the colour of all attached particles
     * <p>Particle colour will be assigned at random from the specified WeightedList
     * @param c colour list
     */
    public void setColour(WeightedList<PSColour> c){
        super.setColour(c.get());
        for(DynamicParticle p : buffer)
            if(p != null && !p.isDead()) p.setColour(c.get());
    }
    /**
     * Set the colour of all attached particles with the specified fade time
     * <p>Particle colour will be assigned at random from the specified WeightedList
     * @param c colour list
     * @param t fade time (ms)
     */
    public void setColour(WeightedList<PSColour> c, int t){
        super.setColour(c.get(), t);
        for(DynamicParticle p : buffer)
            if(p != null && !p.isDead()) p.setColour(c.get(), t);
    }
    /**
     * Set the alpha value of all attached particles
     * @param a alpha
     */
    @Override public void setAlpha(float a){
        super.setAlpha(a);
        for(DynamicParticle p : buffer)
            if(p != null && !p.isDead()) p.setAlpha(a);
    }
    /**
     * Set the particle speed multiplier
     * @param s speed multiplier
     */
    public final void setSpeed(float s){
        speed = s;
    }
    /**
     * Set the initial velocity scalar for attached particles
     * <p>Initial particle velocity is calculated as the inverse of the parent
     * velocity multiplied by the specified values.
     * <p>Default: 1.0
     * @param v scalar
     */
    public final void setInitialVelocity(float v){
        setInitialVelocity(v, v);
    }
    /**
     * Set the initial velocity scalar range for attached particles
     * <p>Initial particle velocity is calculated as the inverse of the parent
     * velocity multiplied by a random value between the specified limits.
     * <p>Default: 1.0, 2.0
     * @param min
     * @param max
     */
    public final void setInitialVelocity(float min, float max){
        init_velocity_min = min;
        init_velocity_max = max;
    }
    /**
     * Set the particle creation radius
     * <p>
     * The particle creation radius specifies the distance from the emitter at
     * which particles will be created.
     * <p>
     * Default: 0.0
     * @param r radius
     */
    public final void setRadius(float r){
        setRadius(r, r);
    }
    /**
     * Set the particle creation radius range
     * <p>
     * The particle creation radius range specifies the minimum and maximum distances
     * from the emitter at which particles will be created.
     * <p>
     * Default: 0.0, 0.0
     * @param min minimum
     * @param max maximum
     */
    public final void setRadius(float min, float max){
        radius_min = min;
        radius_max = max;
    }
    /**
     * Set the particle emission direction
     * <p>
     * Default: (0.0, -1.0, 0.0)
     * @param d direction vector
     */
    public final void setDirection(PSVector d){
        direction = d;
    }
    /**
     * Set the particle emission spread
     * <p>
     * The particle emission spread specifies the emission cone for the emitter.
     * A value of 1.0 will result in spherical emission, while a value of 0.0
     * will result in emission along the direction vector.
     * <p>
     * Default: 1.0
     * @param s spread
     */
    public final void setSpread(float s){
        spread = PSTools.constrainFloat(s, 0f, 1f);
    }
    /**
     * Set the maximum velocity for all attached particles
     * <p>
     * Default: 5.0
     * @param v maximum velocity
     */
    public final void setMaxVelocity(float v){
        velocity_max = v;
    }
    /**
     * Set the direction mode - STATIC, LEADING, TRAILING
     * <p>
     * The direction mode allows the particle emission direction to be modified
     * based on the emitter velocity. The STATIC mode will result in particle
     * emission along the direction vector, while LEADING and TRAILING will result
     * in particle emission along the velocity vector and inverse velocity vector
     * respectively.
     * <p>
     * Default: STATIC
     * @param m mode
     */
    public final void setDirectionMode(int m){
        direction_mode = m;
    }
    /**
     * Set the emitter dimension scalars
     * <p>The dimension scalar vector defines the dimensionality of the emitter.
     * For example, a dimension scalar vector of (1.0, 1.0, 0.0) result in all
     * attached particles being emitted only on the two dimensional xy plane.
     * <p>All attached forces and targets will also be assigned the dimension scalar
     * vector of their parent emitter.
     * @param v
     */
    public final void setDimensionScalars(PSVector v){
        dimensions = v;
    }
    /**
     * Return a reference to the emitter dimension scalar vector
     * @return a reference to the emitter dimension scalar vector
     */
    public final PSVector getDimensionScalars(){
        return dimensions;
    }
    /**
     * Enable/disable particle stealing
     * <p>If particle stealing is enabled and the buffer is full then older
     * particles will be removed to allow for newer ones to be attached.
     */
    public final void setParticleStealing(boolean b){
        stealing = b;
    }
    /**
     * Return true if this and all attached particles are dead
     * @return true if this and all attached particles are dead
     */
    @Override public boolean isDead(){
        if(super.isDead() && attachedParticles() == 0) return true;
        else return false;
    }
    /**
     * Return all attached particles as an array
     * <p>The function finds all non-null and alive particles within the buffer
     * and copies references to them into the returned array. As such it is not
     * particularly efficient and should not be thought of as a direct reference
     * to the particle buffer.
     * @return all attached particles as an array
     */
    public final DynamicParticle[] toArray(){
        ArrayList lst = new ArrayList(buffer.length);
        for(DynamicParticle p : buffer)
            if(p != null && !p.isDead()) lst.add(p);
        DynamicParticle[] arr = new DynamicParticle[lst.size()];
        lst.toArray(arr);
        return arr;
    }
    /**
     * Set the emitter buffer size
     * <p>The emitter buffer size dictates the maximum total particles that may
     * be attached to the emitter. If the specified size is less than the current
     * buffer size, attached particles will be culled, eldest first.
     * @param s size
     */
    public final void setBufferSize(int s){
        if(s == buffer.length) return;
        DynamicParticle[] arr = toArray();
        buffer = new DynamicParticle[s];
        int bs = arr.length;
        if(s < arr.length){
            Arrays.sort(arr);
            bs = s;
        }
        System.arraycopy(arr, 0, buffer, 0, bs);
    }
    /**
     * Return the emitter buffer size
     * @return the emitter buffer size
     */
    public final int getBufferSize(){
        return buffer.length;
    }
    /**
     * Return the current particle count
     * @return the current particle count
     */
    public final int attachedParticles(){
        int s = 0;
        for(int i = 0; i < buffer.length; i ++)
            if(buffer[i] != null && !buffer[i].isDead()) s ++;
        return s;
    }
    /**
     * Render all attached particles to the specified OpenGL renderer
     * @param pgl OpenGL renderer
     */
    @Override public void render(PGraphicsOpenGL pgl){
        render(pgl, null, null, null);
    }
    /**
     * Render all attached particles to the specified OpenGL renderer
     * @param pgl OpenGL renderer
     * @param cpv camera position vector
     * @param ctv camera target vector
     * @param cuv camera up vector
     */
    @Override public void render(PGraphicsOpenGL pgl, PSVector cpv, PSVector ctv, PSVector cuv){
        for(DynamicParticle p : buffer)
            if(p != null && !p.isDead()) p.render(pgl, cpv, ctv, cuv);
    }
    /**
     * Render the emitter location to the specified OpenGL renderer
     * @param pgl OpenGL renderer
     */
    @Override public void debug(PGraphicsOpenGL pgl){
        debug(pgl, null, null, null);
    }
    /**
     * Render the emitter location to the specified OpenGL renderer
     * @param pgl OpenGL renderer
     * @param cpv camera position vector
     * @param ctv camera target vector
     * @param cuv camera up vector
     */
    @Override public void debug(PGraphicsOpenGL pgl, PSVector cpv, PSVector ctv, PSVector cuv){
        super.render(pgl, cpv, ctv, cuv);
        for(DynamicParticle p : buffer)
            if(p != null && !p.isDead()) p.debug(pgl, cpv, ctv, cuv);
    }
    /**
     * Update this all attached particles
     */
    @Override public void update(){        
        // Update particles
        for(DynamicParticle p : buffer)
            if(p != null && !p.isDead()) updateParticle(p);
        // Update behaviours
        PSBehaviour[] b = super.getBehaviour();
        if(b == null || b.length < 1) return;
        for(int i = 0; i < b.length; i ++){
            if(b[i] instanceof PSEmission){ // Emission behaviours
                PSEmission e = (PSEmission) b[i];
                e.update(this);
            } else // Standard behaviours
                b[i].update(this);
        }
    }
    /**
     * Update the specified particle
     * @param p particle
     */
    protected void updateParticle(DynamicParticle p){
        // Particle Updates
        p.update();
        // Forces
        if(forces != null && forces.length > 0){
            for(PSForce f : forces)
                f.update(p);
        }
        // Apply velocities assuming copies not references
        PSVector pv = p.getPosition();
        PSVector vv = p.getVelocity();
        vv.limit(velocity_max);
        vv.scale(speed);
        pv.add(vv);
        p.setVelocity(vv);
        p.setPosition(pv);
    }
    /**
     * Attach a particle to the emitter
     * <p>When a particle is attached to the emitter, the particle position and
     * velocity vectors will be modified to reflect the behaviour of the emitter.
     * References to both vectors will, however, be preserved.
     * @param p particle
     * @return true if the particle was successfully attached
     */
    public final boolean attach(DynamicParticle p){
        int i = getEmptyIndex();
        if(i < 0 || super.isDead()) return false;
        PSVector pv, vv, dv;
        // Calculate velocity vector
        if(direction_mode == LEADING || direction_mode == TRAILING){
            dv = getVelocity().copy();
            dv.normalise();
            if(direction_mode == TRAILING) dv.scale(-1.0f);
        } else
            dv = direction; // Radial dispersion
        vv = getEmissionVector(dv, spread);
        // Calculate position offset vector
        pv = vv.copy();
        pv.normalise();
        // Randomly scale the vector between the specified range to the power of 1/3 to ensure equal distribution within the sphere
        pv.scale((float) Math.pow(random.nextDouble(), 1d / 3d) * (radius_max - radius_min) + radius_min);
        pv.add(getPosition());
        // Calculate velocity vector
        vv.scale(random.nextFloat() * (init_velocity_max - init_velocity_min) + init_velocity_min);
        // Scale vectors to the emitter dimensions
        pv.scale(dimensions);
        vv.scale(dimensions);
        // Apply vectors preserving references and store
        p.getPosition().set(pv);
        p.getVelocity().set(vv);
        p.setPosition(p.getPosition());
        p.setVelocity(p.getVelocity());
        buffer[i] = p;
        return true;
    }
    /**
     * Attach an array of particles to the emitter
     * @param pa particle array
     * @return the amount of particles attached
     */
    public final int attach(DynamicParticle[] pa){
        int c = 0;
        for(int i = 0; i < pa.length; i ++)
            if(attach(pa[i])) c++;
        return c;
    }
    private int getEmptyIndex(){
        if(buffer[0] == null || buffer[0].isDead())
            return 0;
        int sind = 0;        
        float dval = buffer[0].getDecay();        
        for(int i = 1; i < buffer.length; i ++){
            if(buffer[i] == null || buffer[i].isDead())
                return i;
            else if(buffer[i].getDecay() < dval){
                dval = buffer[i].getDecay();
                sind = i;
            }
        }
        if(stealing)
            return sind;
        else
            return -1;    
    }
    /**
     * Remove the specified particle from the emitter
     * <p>If multiple references to the particle exist, they will all be removed
     * @param p particle
     * @return true if the specified particle was found and removed
     */
    public final boolean remove(DynamicParticle p){
        boolean b = false;
        for(int i = 0; i < buffer.length; i ++){
            if(buffer[i] == p){
                buffer[i] = null;
                if(!b) b = true;
            }
        }
        return b;
    }
    /**
     * Clear all attached particles
     */
    public final void clear(){
        for(int i = 0; i < buffer.length; i ++)
            buffer[i] = null;
    }
    /**
     * Attach a force to the emitter
     * @param f force
     */
    public final void setForces(PSForce f){
        forces = new PSForce[] { f };
    }
    /**
     * Attach an array of forces to the emitter
     * @param f force array
     */
    public final void setForces(PSForce[] f){
        forces = f;
    }
    /**
     * Remove all attached forces
     */
    public final void clearForces(){
        forces = null;
    }   
    /**
     * Return a copy of this emitter
     * <p>The returned copy will contain an empty buffer the same size as that of
     * the source particle and will adhere to the rules defined in the map() function
     * of this object.
     * @return a copy of this emitter
     */
    @Override public GenericEmitter copy(){
        GenericEmitter e = new GenericEmitter(buffer.length);
        this.map(e);
        return e;
    }
    /**
     * Map the properties of this e to that specified preserving references
     * <p>The returned object takes a copy of all emitter variables, with the
     * exception for the dimension scalar vector for which a reference is used.
     * DynamicParticle variables are mapped based on the rules defined in the
     * DynamicParticle map() function.
     * @see DynamicParticle
     * @param e emitter
     */
    public final void map(GenericEmitter e){
        if(e.getBufferSize() != buffer.length) e.setBufferSize(buffer.length);
        e.setForces(forces);
        e.setDimensionScalars(dimensions);
        e.setDirection(direction);
        e.setDirectionMode(direction_mode);
        e.setInitialVelocity(init_velocity_min, init_velocity_max);
        e.setMaxVelocity(velocity_max);
        e.setRadius(radius_min, radius_max);
        e.setSpeed(speed);
        e.setSpread(spread);
        e.setParticleStealing(stealing);
        super.map(e);
    }
    protected PSVector getRadialVector(){
        float[] xyz = new float[3];
        for(int i = 0; i < xyz.length; i ++)
            xyz[i] = random.nextFloat() * 2f - 1f;
        PSVector v = new PSVector(xyz);
        v.normalise();

        return v;
    }
    protected PSVector getEmissionVector(PSVector d, float s){
        PSVector dv = d.copy();
        PSVector cv = getRadialVector();
        dv.normalise();
        dv.scale(1.0f - s);
        cv.scale(s);
        cv.add(dv);
        cv.normalise();
        return cv;
    }
}
