package pslib;
import processing.opengl.PGraphicsOpenGL;
/**
 * GenericParticle Class
 * <p>The GenericParticle class allows the creation of a particle with position,
 * velocity, colour, size and alpha variables. The class is primarily designed
 * to be extended with custom rendering routines.
 * <p>This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * <p>You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see http://www.gnu.org/licenses/.
 * @author Stephen Callear
 */
public class GenericParticle extends Object{
    private PSVector position, velocity;
    private PSColourRamp colour;
    private float alpha, size;
    /**
     * Create a new object instance with default values
     */
    public GenericParticle(){
        position = new PSVector();
        velocity = new PSVector();
        colour = new PSColourRamp(new PSColour(255));
        alpha = 1f;
        size = 10f;
    }
    /**
     * Return a reference to the particle position vector
     * @return a reference to the particle position vector
     */
    public PSVector getPosition(){
        return position;
    }
    /**
     * Set the particle position
     * @param position position
     */
    public void setPosition(PSVector position){
        this.position = position;
    }
    /**
     * Return a reference to the particle velocity vector
     * @return a reference to the particle velocity vector
     */
    public PSVector getVelocity(){
        return velocity;
    }
    /**
     * Set the particle velocity
     * @param velocity velocity
     */
    public void setVelocity(PSVector velocity){
        this.velocity = velocity;
    }
    /**
     * Return the current particle size
     * @return the current particle size
     */
    public float getSize(){
        return size;
    }
    /**
     * Set the particle size
     * @param size size
     */
    public void setSize(float size){
        this.size = size;
    }
    /**
     * Return the current particle colour
     * @return the current particle colour
     */
    public PSColour getColour(){
        return colour.get();
    }
    /**
     * Set the particle colour
     * @param colour colour
     */
    public void setColour(PSColour colour){
        this.colour.set(colour);
    }
    /**
     * Set the particle colour
     * @param colour colour
     * @param time transform duration
     */
    public void setColour(PSColour colour, int time){
        this.colour.set(colour, time);
    }
    /**
     * Set the particle alpha
     * @param alpha alpha
     */
    public void setAlpha(float alpha){
        this.alpha = alpha;
    }
    /**
     * Return the particle alpha multiplier
     * @return the particle alpha multiplier
     */
    public float getAlpha(){
        return alpha;
    }
    /**
     * Render the particle within two-dimensional space to the specified
     * rendering context
     * @param pgl OpenGL renderer
     */
    public void render(PGraphicsOpenGL pgl){
        render(pgl, null, null, null);
    }
    /**
     * Render the particle within three-dimensional space to the specified
     * rendering context
     * @param pgl OpenGL renderer
     * @param cpv camera position vector
     * @param ctv camera target vector
     * @param cuv camera up vector
     */
    public void render(PGraphicsOpenGL pgl, PSVector cpv, PSVector ctv, PSVector cuv){
        PSColour c = colour.get();
        int av = Math.round((float) c.getAlpha() * alpha);
        pgl.noStroke();
        pgl.fill(c.getRed(), c.getGreen(), c.getBlue(), av);
        pgl.pushMatrix();
        pgl.translate(position.x, position.y, position.z);
        PSShapes.circle(pgl, 16, size * 0.5f, cpv, ctv, cuv);
        pgl.popMatrix();
    }
}
