package pslib;
import processing.core.PImage;
import processing.opengl.PGraphicsOpenGL;
/**
 * ImageParticle Class
 * <p>The ImageParticle object allows an image to be rendered as a billboard
 * within 3D space. By default the object exhibits the usual billboard behaviour;
 * orienting itself towards the camera. Alternatively the particle may exhibit
 * directional behaviour; orienting itself to its current velocity vector.
 * Additionally, rotation and rotation increment may be specified.
 * <p>This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * <p>You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see http://www.gnu.org/licenses/.
 * @author Stephen Callear
 */
public class ImageParticle extends DynamicParticle {
    private PImage texture;
    private PSVector aspect;
    private float rotation, increment;
    private boolean directional;
    /**
     * Create a new object instance with the specified texture
     * @param t texture
     */
    public ImageParticle(PImage t){
        this(t, 0);
    }
    /**
     * Create a new object instance with the specified texture and decay time
     * @param t texture
     * @param d decay time (ms)
     */
    public ImageParticle(PImage t, int d){
        super(d);
        texture = t;
        aspect = getImageAspect(t);
        rotation = 0;
        increment = 0;
        directional = false;
    }
    /**
     * Set the particle image
     * @param t texture
     */
    public final void setTexture(PImage t){
        texture = t;
        aspect = getImageAspect(t);
    }
    /**
     * Return a reference to the particle image
     * @return a reference to the particle image
     */
    public final PImage getTexture(){
        return texture;
    }
    /**
     * Set directional behaviour for the billboard
     * @param b true = directional, false = static
     */
    public final void setDirectional(boolean b){
        directional = b;
    }
    /**
     * Set the image rotation in radians
     * <p>Rotation occurs around the axis defined by difference between particle
     * and camera locations or around the z-axis if the camera location is
     * unspecified.
     * @param r rotation
     */
    public final void setRotation(float r){
        rotation = r % (2.0f * (float) Math.PI);
    }
    /**
     * Set the rotation increment in radians
     * <p>The rotation increment is applied to the image rotation on each
     * call to render() and offers a crude mechanism for automatic rotation.
     * @param i increment
     */
    public final void setRotationIncrement(float i){
        increment = i;
    }
    /**
     * Render the particle to the specified renderer
     * @param glc OpenGL renderer
     */
    @Override public void render(PGraphicsOpenGL glc){
        render(glc, null, null, null);
    }
    /**
     * Render the particle to the specified renderer
     * @param glc OpenGL renderer
     * @param cpv camera position vector
     * @param ctv camera target vector
     * @param cuv camera up vector
     */
    @Override public void render(PGraphicsOpenGL glc, PSVector cpv, PSVector ctv, PSVector cuv){
        // Increment rotation
        setRotation(rotation + increment);
        // Gather vertices
        PSColour c = getColour();
        int alpha = Math.round((float) c.getAlpha() * getRenderAlpha());
        PSVertex[] va;
        if(directional)
            va = getDirectionalVertices(cpv);
        else
            va = getVertices(cpv, cuv);
        // Render billboard
        glc.pushMatrix();
        glc.noStroke();
        glc.tint(c.getRed(), c.getGreen(), c.getBlue(), alpha);
        glc.translate(getPosition().x, getPosition().y, getPosition().z);
        glc.beginShape();
        glc.texture(texture);
        for(int i = 0; i < va.length; i ++){
            glc.vertex(va[i].position.x, va[i].position.y, va[i].position.z,
                       va[i].uvcoord.x, va[i].uvcoord.y);
        }
        glc.endShape();
        glc.popMatrix();
    }
    /**
     * Return a copy of this particle
     * <p>The returned copy will use a reference to the texture image of the source
     * particle and take copies of the rotation, rotation increment and directional
     * variables. The returned object will adhere to the rules defined in the
     * DynamicParticle map() function.
     * @see DynamicParticle
     * @return a copy of this particle
     */
    @Override public DynamicParticle copy(){
        ImageParticle p = new ImageParticle(texture);
        p.setRotation(rotation);
        p.setRotationIncrement(increment);
        p.setDirectional(directional);
        super.map(p);
        return p;
    }
    private PSVertex[] getVertices(PSVector cpv, PSVector cuv){
        // Calculate billboard orientation
        PSVector blv, brv, buv;
        if(cpv != null && cuv != null){
            blv = PSVector.sub(getPosition(), cpv);
            blv.normalise();
            brv = blv.cross(cuv);
            brv.normalise();
            buv = brv.cross(blv);
        } else {
            blv = PSVector.scale(CAM2D_LOOK, -1f);
            brv = PSVector.scale(CAM2D_RIGHT, -1f);
            buv = PSVector.scale(CAM2D_UP, -1f);
        }
        // Calculate vertex offsets
        float xs = aspect.x * getRenderSize() * 0.5f;
        float ys = aspect.y * getRenderSize() * 0.5f;
        brv.scale(xs);
        buv.scale(ys);
        // Calculate vertex positions
        PSVector v1, v2, v3, v4;
        v1 = new PSVector();
        v1.sub(brv);
        v1.sub(buv);
        v2 = new PSVector();
        v2.add(brv);
        v2.sub(buv);
        v3 = new PSVector();
        v3.add(brv);
        v3.add(buv);
        v4 = new PSVector();
        v4.sub(brv);
        v4.add(buv);
        // Rotate vertices around look vector
        v1.rotateAboutOrigin(rotation, blv);
        v2.rotateAboutOrigin(rotation, blv);
        v3.rotateAboutOrigin(rotation, blv);
        v4.rotateAboutOrigin(rotation, blv);
        // Generate vertex array
        float w = texture.width;
        float h = texture.height;
        PSVertex[] va = new PSVertex[4];
        va[0] = new PSVertex(0, v1, new PSVector(0.0f, 0.0f), null);
        va[1] = new PSVertex(1, v2, new PSVector(w   , 0.0f), null);
        va[2] = new PSVertex(2, v3, new PSVector(w   , h   ), null);
        va[3] = new PSVertex(3, v4, new PSVector(0.0f, h   ), null);
        return va;
    }
    private PSVertex[] getDirectionalVertices(PSVector cpv){
        // Calculate direction
        float xs = aspect.x * getRenderSize() * 0.5f;
        float ys = aspect.y * getRenderSize() * 0.5f;
        PSVector dir = getVelocity().copy();
        dir.normalise();
        dir.scale(ys);
        // Calculate limits
        PSVector p1, p2, p1_n, p2_n;
        p1 = dir.copy();
        p2 = PSVector.scale(dir, -1.0f);
        p1_n = getBillBoardNormal(p1, cpv);
        p2_n = getBillBoardNormal(p2, cpv);
        p1_n.scale(xs);
        p2_n.scale(xs);
        // Calculate vertices
        PSVector v1, v2, v3, v4;
        v1 = PSVector.sub(p1, p1_n);
        v2 = PSVector.add(p1, p1_n);
        v3 = PSVector.sub(p2, p2_n);
        v4 = PSVector.add(p2, p2_n);
        // Rotate vertices
        PSVector buv, brv, blv;
        buv = PSVector.sub(v3, v1);
        brv = PSVector.sub(v3, v4);
        buv.normalise();
        brv.normalise();
        blv = buv.cross(brv);
        blv.normalise();
        v1.rotateAboutOrigin(rotation, blv);
        v2.rotateAboutOrigin(rotation, blv);
        v3.rotateAboutOrigin(rotation, blv);
        v4.rotateAboutOrigin(rotation, blv);
        // Generate vertex array
        PSVertex[] va = new PSVertex[4];
        float w = texture.width;
        float h = texture.height;
        PSColour c1, c2;
        c1 = new PSColour(255, 0, 0);
        c2 = new PSColour(0, 255, 0);
        va[0] = new PSVertex(0, v1, new PSVector(0.0f, 0.0f), c1);
        va[1] = new PSVertex(1, v2, new PSVector(w   , 0.0f), c1);
        va[2] = new PSVertex(2, v3, new PSVector(w   , h   ), c2);
        va[3] = new PSVertex(3, v4, new PSVector(0.0f, h   ), c2);
        return va;
    }
    private PSVector getBillBoardNormal(PSVector p, PSVector cpv){
        PSVector blv, n1, n2, nf;
        if(p != null && cpv != null){
            blv = PSVector.sub(getPosition(), cpv);
            blv.normalise();
        }
        else blv = CAM2D_LOOK;
        n1 = p.cross(blv);
        n1.normalise();
        n2 = p.cross(n1);
        n2.normalise();
        nf = p.cross(n2);
        nf.normalise();
        return nf;
    }
    private PSVector getImageAspect(PImage t){
        PSVector ar;
        if(t.width > t.height)
            ar = new PSVector(1.0f, (float) t.height / (float) t.width, 0.0f);
        else
            ar = new PSVector((float) t.width / (float) t.height, 1.0f, 0.0f);
        return ar;
    }

}
