package pslib;
/**
 * Orbiter Class
 * <p>The Orbiter object extends the ParentParticle functionality by incorporating
 * orbit simulation. Object position is calculated from the accumulated gravities
 * of surrounding Orbiter objects.
 * <p>This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * <p>You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see http://www.gnu.org/licenses/.
 * @author Stephen Callear
 */
public class OrbiterParticle extends ParentParticle {
    private float max_velocity, max_force;
    private double damping;
    /**
     * Create a new object instance with the specified mass, default buffer size (1)
     * and no decay
     * @param m mass
     */
    public OrbiterParticle(float m){
        this(m, 1, 0);
    }
    /**
     * Create a new object instance with the specified mass and buffer size, and
     * no decay
     * @param m mass
     * @param b buffer size
     */
    public OrbiterParticle(float m, int b){
        this(m, b, 0);
    }
    /**
     * Create a new object instance with the specified mass, buffer size and decay
     * @param m mass
     * @param b buffer size
     * @param d decay (ms)
     */
    public OrbiterParticle(float m, int b, int d){
        super(b, d);
        setMass(m);
        damping = 1E-3d;
        max_velocity = 0f;
        max_force = 0f;
    }  
    /**
     * Set the gravitational constant
     * <p>The gravitational constant is applied in gravitational force calculations
     * as G in <code>F = G * (m1 * m2) / r</code>.
     * <p>Default: 0.001
     * @param d force damping
     */
    public final void setDamping(double d){
        damping = d;
    }
    /**
     * Set the maximum velocity magnitude
     * <p>If the specified value is greater than zero then the particle velocity
     * vector will be limited in magnitude to prior to application. It should be
     * noted that limiting the particle velocity vector may result in unpredictable
     * simulation results.
     * <p>Default 0.0
     * @param v maximum velocity
     */
    public final void setMaxVelocity(float v){
        max_velocity = v;
    }
    /**
     * Set the maximum force magnitude
     * <p>If the specified value is greater than zero then the accumulated 
     * gravitational force vector will be limited in magnitude prior to application.
     * It should be noted that limiting the maximum force magnitude may result in
     * unpredictable simulation results.
     * <p>Default 0.0
     * @param f maximum force
     */
    public final void setMaxForce(float f){
        max_force = f;
    }
    /**
     * Update the particle
     * <p> Orbiter velocity is calculated by accumulating the gravitational forces
     * of each element within the orbiter array.
     * @param pa orbiter array
     */
    public void update(DynamicParticle[] pa){
        // Apply behaviours
        super.update();
        // Apply orbit behaviour
        PSVector pos = getPosition();
        PSVector vel = getVelocity();                
        PSVector dst, vec = new PSVector();
        double rad, frc;
        for(DynamicParticle p : pa){
            if(p != this){                
                // F = G * (m1 * m2) / r
                rad = pos.distance(p.getPosition());
                frc = (damping / pa.length) * (p.getMass() * getMass()) / rad ;                
                // a = f / m
                dst = PSVector.sub(p.getPosition(), pos);
                dst.normaliseTo((float) (frc / (double) getMass()));
                if(max_force > 0) dst.limit(max_force);
                vec.add(dst);                
            }
        }
        vel.add(vec);
        if(max_velocity > 0) vel.limit(max_velocity);
        pos.add(vel);
        // Reapply vectors for any attached particles that rely on it
        setVelocity(vel);
        setPosition(pos);    
    }
    /**
     * Return a copy of this particle
     * <p>The returned particle will copy all of this objects variables and will
     * adhere to the rules defined in <code>DynamicParticle.map()</code>.
     * @see DynamicParticle
     * @return a copy of this particle
     */
    @Override public OrbiterParticle copy(){
        OrbiterParticle p = new OrbiterParticle(getMass());
        map(p);
        return p;
    }
    /**
     * Map the properties of this particle to that specified preserving references
     * <p>The returned object takes a copy of all orbiter particle variables.
     * DynamicParticle variables are mapped based on the rules defined in <code>
     * DynamicParticle.map()</code>.
     * @see DynamicParticle
     * @param p particle
     */
    public final void map(OrbiterParticle p){
        super.map(p);
        p.setDamping(damping);
        p.setMaxForce(max_force);
        p.setMaxVelocity(max_velocity);
    }
}