package pslib;
import java.awt.Color;
/**
 * PSColour Class
 * <p>The PSColour object allows the storage of RGBA component colour values.
 * Stored values may be modified as required whilst preserving the object reference.
 * <p>This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * <p>You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see http://www.gnu.org/licenses/.
 * @author Stephen Callear
 */
public class PSColour extends Object{
    private float red, green, blue, alpha;
    /**
     * Create a new object instance with default values
     */
    public PSColour(){
        this(0, 0, 0, 255);
    }
    /**
     * Create a new object instance with the specified value (0-255)
     * @param g greyscale
     */
    public PSColour(int g){
        this(g, g, g, 255);
    }
    /**
     * Create a new object instance with the specified values (0-255)
     * @param g greyscale
     * @param a alpha
     */
    public PSColour(int g, int a){
        this(g, g, g, a);
    }
    /**
     * Create a new object instance with the specified values (0-255)
     * @param r red
     * @param g green
     * @param b blue
     */
    public PSColour(int r, int g, int b){
        this(r, g, b, 255);
    }
    /**
     * Create a new object instance with the specified values (0-255)
     * @param r red
     * @param g green
     * @param b blue
     * @param a alpha
     */
    public PSColour(int r, int g, int b, int a){
        red  = toFloat(r);
        green = toFloat(g);
        blue  = toFloat(b);
        alpha = toFloat(a);
    }
    /**
     * Create a new object instance with the specified value (0f-1f)
     * @param g greyscale
     */
    public PSColour(float g){
        this(g, g, g, 1f);
    }
    /**
     * Create a new object instance with the specified values (0f-1f)
     * @param g greyscale
     * @param a alpha
     */
    public PSColour(float g, float a){
        this(g, g, g, a);
    }
    /**
     * Create a new object instance with the specified values (0f-1f)
     * @param r red
     * @param g green
     * @param b blue
     */
    public PSColour(float r, float g, float b){
        this(r, g, b, 1f);
    }
    /**
     * Create a new object instance with the specified values (0f-1f)
     * @param r red
     * @param g green
     * @param b blue
     * @param a alpha
     */
    PSColour(float r, float g, float b, float a){
        red   = constrain(r);
        green = constrain(g);
        blue  = constrain(b);
        alpha = constrain(a);
    }
    /**
     * Set the red component value (0-255)
     * @param r red
     */
    public final void setRed(int r){
        red = toFloat(r);
    }
    /**
     * Set the red component value (0f-1f)
     * @param r red
     */
    public final void setRed(float r){
        red = constrain(r);
    }
    /**
     * Set the green component value (0-255)
     * @param g green
     */
    public final void setGreen(int g){
        green = toFloat(g);
    }
    /**
     * Set the green component value (0f-1f)
     * @param g green
     */
    public final void setGreen(float g){
        green = constrain(g);
    }
    /**
     * Set the blue component value (0-255)
     * @param b blue
     */
    public final void setBlue(int b){
        blue = toFloat(b);
    }
    /**
     * Set the blue component value (0f-1f)
     * @param b blue
     */
    public final void setBlue(float b){
        blue = constrain(b);
    }
    /**
     * Set the alpha component value (0-255)
     * @param a alpha
     */
    public final void setAlpha(int a){
        alpha = toFloat(a);
    }
    /**
     * Set the alpha component value (0f-1f)
     * @param a alpha
     */
    public final void setAlpha(float a){
        alpha = constrain(a);
    }
    /**
     * Set the component values of this object to those of the specified
     * java.awt.Color object
     * @param c Color object
     */
    public final void set(Color c){
        red   = toFloat(c.getRed());
        green = toFloat(c.getGreen());
        blue  = toFloat(c.getBlue());
        alpha = toFloat(c.getAlpha());
    }
    /**
     * Set the component values of this object to those of the specified
     * PSColour object
     * @param c PSColour object
     */
    public final void set(PSColour c){
        float[] v = c.getComponents(new float[] {});
        red   = v[0];
        green = v[1];
        blue  = v[2];
        alpha = v[3];
    }
    /**
     * Set the component values of this object to those specified by the 32 bit
     * ARGB value
     * @param h ARGB value
     */
    public final void set(int h){
        PSColour c = integerToColour(h);
        this.set(c);
    }
    /**
     * Set the component values of this object (0-255)
     * @param r red
     * @param g green
     * @param b blue
     */
    public final void set(int r, int g, int b){
        red   = toFloat(r);
        green = toFloat(g);
        blue  = toFloat(b);
    }
    /**
     * Set the component values of this object (0-255)
     * @param r red
     * @param g green
     * @param b blue
     * @param a alpha
     */
    public final void set(int r, int g, int b, int a){
        red   = toFloat(r);
        green = toFloat(g);
        blue  = toFloat(b);
        alpha = toFloat(a);
    }
    /**
     * Set the component values of this object (0f-1f)
     * @param r red
     * @param g green
     * @param b blue
     */
    public final void set(float r, float g, float b){
        red   = constrain(r);
        green = constrain(g);
        blue  = constrain(b);
    }
    /**
     * Set the component values of this object (0f-1f)
     * @param r red
     * @param g green
     * @param b blue
     * @param a alpha
     */
    public final void set(float r, float g, float b, float a){
        red   = constrain(r);
        green = constrain(g);
        blue  = constrain(b);
        alpha = constrain(a);
    }
    /**
     * Return the component red value (0-255)
     * @return the component red value
     */
    public final int getRed(){
        return toInteger(red);
    }
    /**
     * Return the component green value (0-255)
     * @return the component green value
     */
    public final int getGreen(){
        return toInteger(green);
    }
    /**
     * Return the component blue value (0-255)
     * @return the component blue value
     */
    public final int getBlue(){
        return toInteger(blue);
    }
    /**
     * Return the component alpha value (0-255)
     * @return the component alpha value
     */
    public final int getAlpha(){
        return toInteger(alpha);
    }
    /**
     * Return the brightness value of this object (0-255)
     * <p>The brightness value is calculated as the average of the component colour
     * values.
     * @return the brightness value of this object
     */
    public final int getBrightness(){
        return toInteger((red + green + blue) / 3f);
    }
    /**
     * Return a copy of this object
     * @return a copy of this object
     */
    public final PSColour copy(){
        return new PSColour(red, green, blue, alpha);
    }
    /**
     * Return an integer array containing the component colour values of this object
     * (0-255)
     * @return an integer array containing the component colour values of this object
     */
    public final int[] getComponents(){
        return new int[] { toInteger(red), toInteger(green), toInteger(blue), toInteger(alpha) };
    }
    /**
     * Fill the specified integer array with the component colour values of this object
     * (0-255) and return the result
     * @param c integer array
     * @return the specified array containing the component colour values of this object
     */
    public final int[] getComponents(int[] c){
        c = new int[] { toInteger(red), toInteger(green), toInteger(blue), toInteger(alpha) };
        return c;
    }
    /**
     * Fill the specified float array with the component colour values of this object
     * (0f-1f) and return the result
     * @param c float array
     * @return the specified array containing the component colour values of this object
     */
    public final float[] getComponents(float[] c){
        c = new float[] { red, green, blue, alpha };
        return c;
    }
    /**
     * Return a java.awt.Color object containing the component colour values of this
     * object
     * @return a Color object containing the component colour values of this object
     */
    public final Color getColor(){
        return new Color(red, green, blue, alpha);
    }
    /**
     * Return a hexadecimal ARGB integer containing the component colour values of
     * this object
     * @return an integer containing the component colour values of this object
     */
    public final int getARGB(){
        return colourToInteger(this);        
    }

    /**
     * Return a 32bit ARGB integer representing the component colour values of
     * the specified PSColour object
     * @param c colour
     * @return an integer containing the component colour values of the specified
     * PSColour object
     */
    public static int colourToInteger(PSColour c){
        int[] rgba = c.getComponents();
        return (rgba[3] << 24) | (rgba[0] << 16) | (rgba[1] << 8) | rgba[2];
    }
    /**
     * Return a PSColour object with component values derived from the specified
     * 32bit ARGB integer
     * @param h ARGB integer
     * @return a PSColour object with component values derived from the specified
     * integer
     */
    public static PSColour integerToColour(int h){
        int r, g, b, a;
        r = (h >> 16) & 0xff;
        g = (h >> 8) & 0xff;
        b = h & 0xff;
        a = (h >> 24) & 0xff;
        return new PSColour(r, g, b, a);
    }
    
    private static int constrain(int i){
        if(i < 0) return 0;
        else if(i > 255) return 255;
        else return i;
    }
    private static float constrain(float f){
        if(f < 0f) return 0f;
        else if(f > 1f) return 1f;
        else return f;
    }
    private static int toInteger(float f){
        return (int) (f * 255f);
    }
    private static float toFloat(int i){
        return (float) constrain(i) / 255f;
    }
}
