package pslib;
import java.util.ArrayList;
import java.util.Random;
import java.io.File;
import java.io.IOException;
import java.awt.image.BufferedImage;
import javax.imageio.ImageIO;
/**
 * PSColourMap Class
 * <p>The PSColourMap object permits the generation of colour maps derived from
 * either image scripts or stored values. Once generated, the resultant map
 * may be output as a two-dimensional PSColour array, a WeightedList array.
 * Individual layers may also be output as either a PSColour array or WeightedList.
 * <p>This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * <p>You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see http://www.gnu.org/licenses/.
 * @author Stephen Callear
 */
public class PSColourMap extends Object {
    private static String LAYER_TAG = "layer";
    private static String WEIGHT_TAG = "weight";
    private static String MS_REGEX = "=";
    private static String IS_REGEX = "#";
    private static String VS_REGEX = ",";
    private PSColour[][] map;
    private int[][] weights;
    /**
     * Create a new object instance
     */
    public PSColourMap(){
        map = null;
        weights = null;
    }
    /**
     * Generate the colour map from the specified script.
     * <p>Script lines should be in the format: layer#[index]=[file],[values]
     * @param f script file
     * @return true if the map was successfully generated
     */
    public final boolean generateFromScript(String f){
        String[] sf = PSTools.readStrings(f);
        if(sf == null) return false;
        String sp = PSTools.removeFileName(f);
        // Read lines
        CSLine[] sl = new CSLine[sf.length];
        for(int i = 0; i < sf.length; i ++)
            sl[i] = CSLine.stringToLine(sf[i], MS_REGEX, IS_REGEX, VS_REGEX);
        // Get size
        int lc = 0;
        for(int i = 0; i < sl.length; i ++){
            if(sl[i] != null && sl[i].hasTag(LAYER_TAG) && sl[i].id > lc)
                lc = sl[i].id;
        }
        // Clean path names
        for(int i = 0; i < sl.length; i ++){
            if(sl[i] != null && sl[i].file.indexOf(File.separatorChar) == -1)
                sl[i].file = sp + sl[i].file;
        }
        // Generate map
        map = new PSColour[lc + 1][0];
        for(int i = 0; i < sl.length; i ++){
            if(sl[i] != null && sl[i].hasTag(LAYER_TAG))
                map[sl[i].id] = generateLayer(sl[i].file, sl[i].value);
        }
        if(hasNullValues(map))
            System.err.println("Warning: PSColourMap generated data contains null values.");
        return true;
    }
    /**
     * Generate the colour map from the specified file
     * <p>File lines should be in the format: layer#[index]=[red],[green],[blue],[alpha]
     * or weight#[index]=[v1],[v2], ... [vn]
     * @param f colour map file
     * @return true if the map was successfully generated
     */
    public final boolean read(String f){        
        String[] sf = PSTools.readStrings(f);
        if(sf == null) return false;
        // Read lines
        CMLine[] sl = new CMLine[sf.length];
        for(int i = 0; i < sf.length; i ++)
            sl[i] = CMLine.stringToLine(sf[i], MS_REGEX, IS_REGEX, VS_REGEX);
        // Get size
        int lc = 0;
        for(int i = 0; i < sl.length; i ++){
            if(sl[i] != null && sl[i].hasTag(LAYER_TAG) && sl[i].id > lc)
                lc = sl[i].id;
        }
        // Gather layers
        ArrayList<PSColour>[] mt = new ArrayList[lc + 1];
        for(int i = 0; i < mt.length; i ++)
            mt[i] = new ArrayList();
        for(int i = 0; i < sl.length; i ++){
            if(sl[i] != null && sl[i].hasTag(LAYER_TAG)){
                if(sl[i].vars.length > 3)
                    mt[sl[i].id].add(new PSColour(sl[i].vars[0], sl[i].vars[1], sl[i].vars[2], sl[i].vars[3]));
                else if(sl[i].vars.length > 2)
                    mt[sl[i].id].add(new PSColour(sl[i].vars[0], sl[i].vars[1], sl[i].vars[2]));
            }
        }
        // Convert to array
        map = new PSColour[mt.length][0];
        for(int i = 0; i < mt.length; i ++){
            map[i] = new PSColour[mt[i].size()];
            mt[i].toArray(map[i]);
        }
        // Gather weights
        weights = new int[lc + 1][0];
        for(int i = 0; i < weights.length; i ++)
            weights[i] = defaultWeights(map[i].length);
        for(int i = 0; i < sl.length; i ++){
            if(sl[i] != null && sl[i].hasTag(WEIGHT_TAG))
                weights[sl[i].id] = parseWeights(sl[i].vars, map[i].length);
        }
        if(hasNullValues(map))
            System.err.println("Warning: PSColourMap generated data contains null values.");
        return true;
    }
    /**
     * Write the map stored in this object to file
     * <p>The output format of the file matches that of the input format specified
     * by read().
     * @param f file
     * @return true if the map was successfully written
     */   
    public final  boolean write(String f){
        if(map == null) return false;
        // Generate a cleaned weight array
        int[][] dw = new int[map.length][0];
        for(int i = 0; i < dw.length; i ++)
            dw[i] = defaultWeights(map[i].length);
        if(weights != null){
            for(int i = 0; i < weights.length && i < dw.length; i ++)
                for(int j = 0; j < weights[i].length && j < dw[i].length; j ++)
                    dw[i][j] = weights[i][j];                   
        }        
        int fs = dw.length;
        for(int i = 0; i < map.length; i ++)
            fs += map[i].length;        
        CMLine[] sl = new CMLine[fs];
        // Create weight lines
        int si = 0;
        for(int i = 0; i < dw.length; i++)
            sl[si++] = new CMLine(WEIGHT_TAG, i, dw[i]);
        // Create layer lines
        for(int i = 0; i < map.length; i++){
            for(int j = 0; j < map[i].length; j++)
                sl[si++] = new CMLine(LAYER_TAG, i, map[i][j].getComponents());
        }
        // Create string array and write
        String[] sa = new String[sl.length];
        for(int i = 0; i < sl.length; i ++)
            sa[i] = CMLine.lineToString(sl[i], MS_REGEX, IS_REGEX, VS_REGEX);
        if(PSTools.writeStrings(sa, f)) return true;
        return false;        
    }
    /**
     * Copy the values of the specified map into this object
     * <p>The map stored within this object will be resized to match the input.
     * @param m colour map
     */
    public final void setMap(PSColour[][] m){
        map = new PSColour[m.length][0];
        for(int i = 0; i < m.length; i ++){
            map[i] = new PSColour[m[i].length];
            for(int j = 0; j < m[i].length; j ++)
                map[i][j] = m[i][j].copy();
        }
        if(hasNullValues(map))
            System.err.println("Warning: PSColourMap data contains null values.");
    }
    /**
     * Set the weights for the stored colour map
     * <p>The weight values are stored within the WeightedList output and dictate
     * the likelihood of value selection.
     * <p>If the specified array does not match the dimensions of the stored map,
     * this will be reflected in the output file. Such invariances will however
     * be corrected prior to WeightedList output.
     * @param w weights
     */
    public final void setWeights(int[][] w){
        weights = new int[w.length][0];
        for(int i = 0; i < w.length; i ++){
            weights[i] = new int[w[i].length];
            System.arraycopy(w[i], 0, weights[i], 0, w[i].length);
        }
    }    
    /**
     * Return the map stored in this object as a two-dimensional PSColour array
     * <p>If the map stored in this object is empty then a null value will be returned.
     * If any layers within the map are empty or null then the returned array will
     * contain null entries.
     * @return the map stored in this object as a two-dimensional PSColour array
     */
    public final PSColour[][] toArray(){
        if(map == null)
            return null;
        PSColour[][] arr = new PSColour[map.length][0];
        for(int i = 0; i < map.length; i ++){
            arr[i] = new PSColour[map[i].length];
            for(int j = 0; j < map[i].length; j ++)
                arr[i][j] = map[i][j].copy();
        }
        return arr;
    }
    /**
     * Return the specified layer of the map stored in this object as a PSColour array
     * <p>If the map stored in this object is empty or the specified index is invalid
     * then a null value will be returned.
     * @param i layer index
     * @return the specified layer of the map stored in this object as a PSColour array
     */
    public final PSColour[] toArray(int i){
        if(map == null || i > map.length - 1 || map[i] == null)
            return null;
        PSColour[] arr = new PSColour[map[i].length];
        for(int j = 0; j < map[i].length; j ++)
            arr[j] = map[i][j].copy();
        return arr;
    }
    /**
     * Return the map stored in this object as a WeightedList array
     * <p>The output WeightedList array stores all PSColour values alongside
     * their respective weightings.
     * <p>If the map stored in this object is empty then a null value will be returned.
     * If any layers within the map are empty or null then the returned array will
     * contain null entries.
     * @return the map stored in this object as a WeightedList array
     */
    public final WeightedList<PSColour>[] toWeightedList(){
        if(map == null)
            return null;
        WeightedList<PSColour>[] lst = new WeightedList[map.length];
        int[] w;
        for(int i = 0; i < map.length; i ++){
            if(weights == null || weights[i] == null)
                w = defaultWeights(map[i].length);
            else
                w = parseWeights(weights[i], map[i].length);
            if(map[i] == null)
                lst[i] = null;
            else
                lst[i] = new WeightedList<PSColour>(map[i], w);
        }
        return lst;        
    }
    /**
     * Return the specified layer of the map stored in this object as a WeightedList
     * object
     * <p>The output WeightedList object stores all of the layer PSColour values
     * alongside their respective weightings.
     * <p>If the map stored in this object is empty or the specified index is invalid
     * then a null value will be returned.
     * @param i layer index
     * @return the specified layer of the map stored in this object as a WeightedList
     * object
     */
    public final WeightedList<PSColour> toWeightedList(int i){
        if(map == null || i < 0 || i > map.length - 1 || map[i] == null)
            return null;
        int[] w;
        if(weights == null || i > weights.length || weights[i] == null)
            w = defaultWeights(map[i].length);
        else
            w = parseWeights(weights[i], map[i].length);
        return new WeightedList<PSColour>(map[i], w);
    }
    /**
     * Return the amount of layers within the stored map
     * <p>A value of -1 will be returned if the stored map is empty
     * @return the amount of layers within the stored map
     */
    public final int layers(){
        if(map == null) return -1;
        return map.length;
    }
    /**
     * Return the size of the specified map layer
     * <p>A value of -1 will be returned if the stored map is empty, the specified
     * index is invalid, or the specified layer is empty.
     * @param i layer index
     * @return the size of the specified map layer
     */
    public final int size(int i){
        if(map == null || i > map.length - 1 || map[i] == null) return -1;
        return map[i].length;
    }

    private static int[] parseWeights(int[] w, int s){
        int[] arr = new int[s];
        for(int i = 0; i < arr.length; i ++)
            arr[i] = 1;
        for(int i = 0; i < arr.length && i < w.length; i ++)
            arr[i] = w[i] < 1 ? 1 : w[i];
        return arr;
    }
    private static int[] defaultWeights(int s){
        int[] arr = new int[s];
        for(int i = 0; i < arr.length; i ++)
            arr[i] = 1;
        return arr;
    }
    private static PSColour[] generateLayer(String s, int v){
        File f = new File(s);
        BufferedImage img;
        try {
            img = ImageIO.read(f);
        }
        catch (IOException e){
            return null;
        }
        int w = img.getWidth(), h = img.getHeight();
        PSColour[] arr = new PSColour[v];
        Random r = new Random();
        for(int i = 0; i < v; i ++)
            arr[i] = PSColour.integerToColour(img.getRGB(r.nextInt(w), r.nextInt(h)));
        return arr;
    }
    private static boolean hasNullValues(PSColour[][] m){
        if(m == null) return true;
        for(int i = 0; i < m.length; i ++){
            if(m[i] == null) return true;
            for(int j = 0; j < m[i].length; j ++)
                if(m[i][j] == null) return true;
        }
        return false;
    }


    
    
    

    private static class CMLine extends Object {
        public String tag;
        public int id;
        public int[] vars;
        public CMLine(String t, int i, int[] v){
            tag = t;
            id = i;
            vars = v;
        }
        public boolean hasId(int i){
            if(id == i) return true;
            return false;
        }
        public boolean hasTag(String t){
            if(tag.equalsIgnoreCase(t)) return true;
            return false;
        }
        public static CMLine stringToLine(String s, String msr, String isr, String vsr){
            String[] sl, si, sv;
            sl = s.split(msr);
            if(sl.length < 2) return null;
            si = sl[0].split(isr);
            if(si.length < 2) return null;
            sv = sl[1].split(vsr);
            // si[0] = tag, si[1] = id, sv = vars
            int iv;
            int[] va = new int[sv.length];
            try {
                iv = Integer.parseInt(si[1]);
                for(int i = 0; i < sv.length; i ++)
                    va[i] = Integer.parseInt(sv[i]);
            }
            catch (NumberFormatException e){
                return null;
            }
            return new CMLine(si[0].toLowerCase(), iv, va);
        }
        public static String lineToString(CMLine s, String msr, String isr, String vsr){
            StringBuilder sb = new StringBuilder();
            sb.append(s.tag);
            sb.append(isr);
            sb.append(Integer.toString(s.id));
            sb.append(msr);
            for(int i = 0; i < s.vars.length; i ++){
                sb.append(Integer.toString(s.vars[i]));
                if(i < s.vars.length - 1)
                    sb.append(vsr);
            }
            return sb.toString();
        }
    }
    private static class CSLine extends Object {
        public String tag, file;
        public int id;
        public int value;
        public CSLine(String t, int i, String f, int v){
            tag = t;
            id = i;
            file = f;
            value = v;
        }
        public boolean hasId(int i){
            if(id == i) return true;
            return false;
        }
        public boolean hasTag(String t){
            if(tag.equalsIgnoreCase(t)) return true;
            return false;
        }
        public static CSLine stringToLine(String s, String msr, String isr, String vsr){
            String[] sl, si, sv;
            sl = s.split(msr);
            if(sl.length < 2) return null;
            si = sl[0].split(isr);
            if(si.length < 2) return null;
            sv = sl[1].split(vsr);
            // si[0] = tag, si[1] = id, sv[0] = file, sv[1] = value
            int iv, vv;
            try {
                iv = Integer.parseInt(si[1]);
                vv = Integer.parseInt(sv[1]);
            }
            catch (NumberFormatException e){
                return null;
            }
            return new CSLine(si[0].toLowerCase(), iv, sv[0], vv);
        }
        public static String lineToString(CSLine s, String msr, String isr, String vsr){
            StringBuilder sb = new StringBuilder();
            sb.append(s.tag);
            sb.append(isr);
            sb.append(Integer.toString(s.id));
            sb.append(msr);
            sb.append(s.file);
            sb.append(vsr);
            sb.append(Integer.toString(s.value));
            return sb.toString();
        }
    }
}
