package pslib;
/**
 * PSColourRamp Class
 * <p>The PSColourRamp object allows ramping between two colour values over
 * a specified time period.
 * <p>An initial colour value is set on creation. Calls to set() will set a target
 * colour which will be iterated to over the specified time period. Calls to get()
 * will update the object and return a reference to the current colour.
 * <p>This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * <p>You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see http://www.gnu.org/licenses/.
 * @author Stephen Callear
 * @see PSColour
 */
public class PSColourRamp extends Object {
    private PSColour rgb, rgb_p, rgb_t;
    private int time, time_r;
    long time_i;
    private boolean active;
    /**
     * Create a new object instance with the specified colour
     * @param c colour
     */
    public PSColourRamp(PSColour c){
        rgb = c;
        rgb_p = rgb.copy();
        rgb_t = rgb.copy();
        active = false;
        time = millis();
        time_r = millis();
        time_i = System.currentTimeMillis();
    }
    /**
     * Set the object colour
     * @param c colour
     */
    public final void set(PSColour c){
        rgb.set(c);
        rgb_p.set(c);
        rgb_t.set(c);
    }
    /**
     * Set the target colour and transform duration
     * @param c target colour
     * @param t duration (ms)
     */
    public final void set(PSColour c, int t){
        if(!active) active = true;
        time = t;
        time_r = millis();
        rgb_p = rgb.copy();
        rgb_t = c.copy();
    }
    /**
     * Update and return a reference to the object colour
     * @return a reference to the object colour
     */
    public final PSColour get(){
        if(active){
            int ct = millis() - time_r;
            float tp = (float) ct / (float) time;         
            if(tp < 1.0){
                float[] tc = rgb_t.getComponents(new float[] {});
                float[] pc = rgb_p.getComponents(new float[] {});
                float[] nc = new float[4];
                for(int i = 0; i < nc.length; i ++)
                    nc[i] = (tc[i] - pc[i]) * tp + pc[i];
                rgb.set(nc[0], nc[1], nc[2], nc[3]);
            }
            else {
                rgb.set(rgb_t);
                active = false;
            }
        }
        return rgb;
    }
    /**
     * Returns the time in milliseconds since this objects creation
     * @return the time in milliseconds since this objects creation
     */
    private int millis(){
        long m = System.currentTimeMillis() - time_i;
        return (int) m;
    }
}
