package pslib;
/**
 * PSCurve Class
 * <p>The PSCurve object allows the creation of value curves. The object iterates
 * between initial and target values over the specified time period and with
 * the desired curvature.
 * <p>This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * <p>You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see http://www.gnu.org/licenses/.
 * @author Stephen Callear
 */
public class PSCurve extends Object {
    private float value, initial, target, exponent;
    private int time, time_ref;
    private long init_time;
    private boolean active;
    /**
     * Create an object instance with the specified initial value
     * @param v initial value
     */
    public PSCurve(float v){
        value = v;
        initial = v;
        target = v;
        exponent = 1.0f;
        time = 0;
        time_ref = 0;
        init_time = System.currentTimeMillis();
        active = false;
    }
    /**
     * Return the current object value
     * @return the current object value
     */
    public final float get(){
        if (millis() < time_ref + time){
            double m = ((double) millis() - (double) time_ref) / (double) time;
            float c = (float) Math.pow(m, exponent);
            value = initial + (target - initial) * c;
        } else {
            if (value != target) value = target;
            active = false;
        }
        return value;
    }
    /**
     * Set the object value to that specified
     * @param v value
     */
    public final void set(float v){
        active = false;
        time = 0;
        initial = value;
        value = v;
        target = v;
    }
    /**
     * Create a curve with the specified target value and iteration time
     * @param v target value
     * @param t iteration time
     */
    public final void set(float v, int t){
        set(value, v, t, 1.0f);
    }
    /**
     * Create a curve with the specified target value, iteration time and curve
     * exponent
     * @param v target value
     * @param t iteration time
     * @param e curve exponent
     */
    public final void set(float v, int t, float e){
        set(value, v, t, e);
    }
    /**
     * Create a curve with the specified initial and target values, and iteration
     * time
     * @param i initial value
     * @param v target value
     * @param t iteration time
     */
    public final void set(float i, float v, int t){
        set(i, v, t, 1.0f);
    }
    /**
     * Create a curve with the specified initial and target values, iteration time
     * and curve exponent
     * @param i initial value
     * @param v target value
     * @param t iteration time
     * @param e curve exponent
     */
    public final void set(float i, float v, int t, float e){
      active = true;
      time = t;
      time_ref = millis();
      target = v;
      initial = i;
      exponent = e;
    }
    /**
     * Return the activity state of the object
     * @return the activity state of the object
     */
    public final boolean isActive(){
        return active;
    }
    /**
     * Return the time in milliseconds since object creation
     * @return the time in milliseconds since object creation
     */
    private int millis(){
        long m = System.currentTimeMillis() - init_time;
        return (int) m;
    }
}
