package pslib;
/**
 * PSForce Class
 * <p>The PSForce class is an abstract class for the creation of particle forces.
 * Objects that extend this class can be used with the RadialEmitter object to
 * modify the DynamicParticle trajectories.
 * <p>The PSForce class implements the PSBehaviour interface, allowing sub-classes
 * of PSForce to be used within the behavioural update loop of DynamicParticle
 * objects.
 * <p>This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * <p>You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see http://www.gnu.org/licenses/.
 * @author Stephen Callear
 */
public abstract class PSForce extends Object implements PSConstants, PSBehaviour {
    private PSRange range;
    private PSVector dimensions;
    private float scale;
    /**
     * Create a new object instance with the specified magnitude
     * @param m magnitude
     */
    public PSForce(float m){
        this(m, m, STATIC, 1f);
    }
    /**
     * Create a new object instance with the specified values
     * @param mi minimum magnitude
     * @param ma maximum magnitude
     * @param m envelope mode
     * @param c curve exponent
     */
    public PSForce(float mi, float ma, int m, float c){
        range = new PSRange(mi, ma, m, c);
        scale = 1f;
    }
    /**
     * Set the force scalar
     * @param s scalar
     */
    public final void setScale(float s){
        scale = s;
    }
    /**
     * Return the force scalar
     * @return the force scalar
     */
    public final float getScale(){
        return scale;
    }
    /**
     * Set the curve exponent
     * @param e exponent
     */
    public final void setCurve(float e){
        range.setCurve(e);
    }
    /**
     * Set the dimension scalar vector
     * @param d dimension scalar vector
     */
    public final void setDimensionScalars(PSVector d){
        dimensions = d;
    }
    /**
     * Return the dimension scalar vector
     * @return the dimension scalar vector
     */
    public final PSVector getDimensionScalars(){
        return dimensions;
    }
    /**
     * Return the force magnitude vector for the specified particle
     * @param p particle
     * @return the force magnitude vector for the specified particle
     */
    protected final PSVector getMagnitude(DynamicParticle p){
        PSVector v = new PSVector(1f, 1f, 1f);
        if(dimensions != null) v.scale(dimensions);
        v.scale(range.get(1f - p.getDecay()) * scale);
        return v;
    }
    /**
     * Update the specified particle
     * @param p particle
     */
    public abstract void update(DynamicParticle p);

}
