package pslib;
/**
 * PSRange Class
 * <p>The PSRange object allows a minimum and maximum value to be stored. Calls
 * to <code>get()</code> map a value between 0 and 1 to the range and apply the
 * specified envelope and curvature.
 * @author Stephen Callear
 */
public class PSRange extends Object implements PSConstants {
    private float min, max, curve;
    private int mode;
    /**
     * Create a new object instance with the specified values
     * @param mi minimum
     * @param ma maximum
     */
    public PSRange(float mi, float ma){
        this(mi, ma, LINEAR, 1f);
    }
    /**
     * Create a new object instance with the specified values
     * @param mi minimum
     * @param ma maximum
     * @param m envelope mode
     */
    public PSRange(float mi, float ma, int m){
        this(mi, ma, m, 1f);
    }
    /**
     * Create a new object instance with the specified values
     * @param mi minimum
     * @param ma maximum
     * @param m envelope mode
     * @param c curve exponent
     */
    public PSRange(float mi, float ma, int m, float c){
        min = mi <= ma ? mi : ma;
        max = ma >= mi ? ma : mi;
        mode = m;
        curve = c;
    }
    /**
     * Set the minimum and maximum values
     * @param mi minimum
     * @param ma maximum
     */
    public final void setRange(float mi, float ma){
        min = mi;
        max = ma;
    }
    public final float[] getRange(){
        return new float[] { min, max };
    }
    /**
     * Set the envelope mode
     * <p>Available modes are: LINEAR, INVERSE, TRIANGULAR and INVERSE_TRIANGULAR
     * @param m envelope mode
     */
    public final void setMode(int m){
        mode = m;
    }
    /**
     * Set the curve exponent
     * @param c curve exponent
     */
    public final void setCurve(float c){
        curve = c;
    }
    /**
     * Map the input value to the range of this object instance
     * <p>Input values will be constrained between 0 and 1
     * @param v input value
     * @return the mapped result
     */
    public final float get(float v){
        v = PSTools.constrainFloat(v, 0f, 1f);
        return applyEnvelope(v);       
    }
    private float applyEnvelope(float v){
        double env;
        if(mode == INVERSE_TRIANGULAR)       
            env = Math.pow(Math.abs(2d * v - 1d), curve) * (max - min) + min;
        else if(mode == TRIANGULAR)
            env = (1d - Math.pow(Math.abs(2d * v -1d), curve)) * (max - min) + min;
        else if(mode == INVERSE)
            env = (1d - Math.pow(v, curve)) * (max - min) + min;
        else if(mode == LINEAR)
            env = Math.pow(v, curve) * (max - min) + min;
        else // Static or otherwise
            env = max;
        return (float) env;        
    }
}
