package pslib;
import processing.opengl.PGraphicsOpenGL;
/**
 * PSShapes Class
 * <p>The PSShapes object allows a range of shapes to be rendered. All methods
 * are static and require an OpenGL rendering context.
 * <p>This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * <p>You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see http://www.gnu.org/licenses/.
 * @author Stephen Callear
 */
public class PSShapes {
    /**
     * Render a circle as a 2D object to the specified rendering context
     * @param glc OpenGL rendering context
     * @param seg segments
     * @param rad radius
     */
    public static void circle(PGraphicsOpenGL glc, int seg, float rad){
        fan(glc, seg, rad, 1.0f);
    }
    /**
     * Render a circle as a 3D billboard to the specified rendering context
     * @param glc OpenGL rendering context
     * @param seg segments
     * @param rad radius
     * @param cpv camera position vector
     * @param ctv camera target vector
     * @param cuv camera up vector
     */
    public static void circle(PGraphicsOpenGL glc, int seg, float rad, PSVector cpv, PSVector ctv, PSVector cuv){
        fan(glc, seg, rad, 1.0f, cpv, ctv, cuv);
    }
    /**
     * Render a fan as a 2D object to the specified rendering context
     * @param glc OpenGL rendering context
     * @param seg segments
     * @param rad radius
     * @param spd spread
     */
    public static void fan(PGraphicsOpenGL glc, int seg, float rad, float spd){
        fan(glc, seg, rad, spd, null, null, null);
    }
    /**
     * Render a fan as a 3D billboard to the specified rendering context
     * @param glc OpenGL rendering context
     * @param seg segments
     * @param rad radius
     * @param spd spread
     * @param cpv camera position vector
     * @param ctv camera target vector
     * @param cuv camera up vector
     */
    public static void fan(PGraphicsOpenGL glc, int seg, float rad, float spd, PSVector cpv, PSVector ctv, PSVector cuv){
        glc.pushMatrix();
        if(cpv != null && ctv != null && cuv != null){
            PSVector look = PSVector.sub(ctv, cpv);
            look.normalise();
            PSVector right = cuv.cross(look);
            PSVector up = look.cross(right);
            glc.applyMatrix(
              right.x, up.x, look.x, 0.0f,
              right.y, up.y, look.y, 0.0f,
              right.z, up.z, look.z, 0.0f,
              0.0f, 0.0f, 0.0f, 1.0f
            );
        }
        glc.rotateZ(PGraphicsOpenGL.PI * spd);
        glc.beginShape(PGraphicsOpenGL.TRIANGLE_STRIP);
        double segment_size = (2.0 * Math.PI * spd) / seg;
        for(int i = 0; i < seg + 1; i ++){
            int vc = i;
            float vx = (float) Math.sin(vc * segment_size) * rad;
            float vy = (float) Math.cos(vc * segment_size) * rad;
            glc.vertex(0.0f, 0.0f);
            glc.vertex(vx, vy);
        }
       glc.endShape();
       glc.popMatrix();
    }
    /**
     * Render a doughnut as a 2D object to the specified rendering context
     * @param glc OpenGL rendering context
     * @param seg segments
     * @param rad_i inner radius
     * @param rad_o outer radius
     */
    public static void doughnut(PGraphicsOpenGL glc, int seg, float rad_i, float rad_o){
        doughnut(glc, seg, rad_i, rad_o, null, null, null);
    }
    /**
     * Render a doughnut as a 3D billboard to the specified rendering context
     * @param glc OpenGL rendering context
     * @param seg segments
     * @param rad_i inner radius
     * @param rad_o outer radius
     * @param cpv camera position vector
     * @param ctv camera target vector
     * @param cuv camera up vector
     */
    public static void doughnut(PGraphicsOpenGL glc, int seg, float rad_i, float rad_o, PSVector cpv, PSVector ctv, PSVector cuv){
        glc.pushMatrix();
        glc.beginShape(PGraphicsOpenGL.TRIANGLE_STRIP);
        if(cpv != null && ctv != null && cuv != null){
            PSVector look = PSVector.sub(ctv, cpv);
            look.normalise();
            PSVector right = cuv.cross(look);
            PSVector up = look.cross(right);
            glc.applyMatrix(
              right.x, up.x, look.x, 0.0f,
              right.y, up.y, look.y, 0.0f,
              right.z, up.z, look.z, 0.0f,
              0.0f, 0.0f, 0.0f, 1.0f
            );
        }
        double segment_size = (2.0 * Math.PI) / seg;
        for(int i = 0; i < seg + 1; i ++){
            int vc;
            if(i < seg) vc = i;
            else vc = 0;
            float vx_inner = (float) Math.sin(vc * segment_size) * rad_i;
            float vy_inner = (float) Math.cos(vc * segment_size) * rad_i;
            float vx_outer = (float) Math.sin(vc * segment_size) * rad_o;
            float vy_outer = (float) Math.cos(vc * segment_size) * rad_o;
            glc.vertex(vx_inner, vy_inner);
            glc.vertex(vx_outer, vy_outer);
        }
        glc.endShape();
        glc.popMatrix();
    }
    /**
     * Render a star as a 2D object to the specified rendering context
     * @param glc OpenGL rendering context
     * @param seg segments
     * @param rad_i inner radius
     * @param rad_o outer radius
     */
    public static void star(PGraphicsOpenGL glc, int seg, float rad_i, float rad_o){
        star(glc, seg, rad_i, rad_o, null, null, null);
    }
    /**
     * Render a star as a 3D billboard to the specified rendering context
     * @param glc OpenGL rendering context
     * @param seg segments
     * @param rad_i inner radius
     * @param rad_o outer radius
     * @param cpv camera position vector
     * @param ctv camera target vector
     * @param cuv camera up vector
     */
    public static void star(PGraphicsOpenGL glc, int seg, float rad_i, float rad_o, PSVector cpv, PSVector ctv, PSVector cuv){
        glc.pushMatrix();
        glc.beginShape(PGraphicsOpenGL.TRIANGLE_STRIP);
        if(cpv != null && ctv != null && cuv != null){
            PSVector look = PSVector.sub(ctv, cpv);
            look.normalise();
            PSVector right = cuv.cross(look);
            PSVector up = look.cross(right);
            glc.applyMatrix(
              right.x, up.x, look.x, 0.0f,
              right.y, up.y, look.y, 0.0f,
              right.z, up.z, look.z, 0.0f,
              0.0f, 0.0f, 0.0f, 1.0f
            );
        }
        double segment_size = (2.0 * Math.PI) / (seg * 2.0);
        for(int i = 0; i < (seg * 2) + 1; i ++){
            int vc;
            if(i < seg * 2) vc = i;
            else vc = 0;
            float vx, vy;
            if((float) vc % 2 == 0.0){
                vx = (float) Math.sin(vc * segment_size) * rad_i;
                vy = (float) Math.cos(vc * segment_size) * rad_i;
            } else {
                vx = (float) Math.sin(vc * segment_size) * rad_o;
                vy = (float) Math.cos(vc * segment_size) * rad_o;
            }
            glc.vertex(0, 0);
            glc.vertex(vx, vy);
        }
        glc.endShape();
        glc.popMatrix();
    }
    /**
     * Render a grid as a 2D object to the specified OpenGL renderer
     * @param glc OpenGL renderer
     * @param w width
     * @param h height
     * @param xs horizontal spacing
     * @param ys vertical spacing
     */
    public static void grid(PGraphicsOpenGL glc, float w, float h, float xs, float ys){
        grid(glc, w, h, xs, ys, null, null, null);
    }
    /**
     * Render a grid as 3D billboard to the specified OpenGL renderer
     * @param glc OpenGL renderer
     * @param w width
     * @param h height
     * @param xs horizontal spacing
     * @param ys vertical spacing
     * @param cpv camera position vector
     * @param ctv camera target vector
     * @param cuv camera up vector
     */
    public static void grid(PGraphicsOpenGL glc, float w, float h, float xs, float ys, PSVector cpv, PSVector ctv, PSVector cuv){
        float hw = w * 0.5f;
        float hh = h * 0.5f;
        glc.pushMatrix();
        glc.beginShape(PGraphicsOpenGL.LINES);
        if(cpv != null && ctv != null && cuv != null){
            PSVector look = PSVector.sub(ctv, cpv);
            look.normalise();
            PSVector right = cuv.cross(look);
            PSVector up = look.cross(right);
            glc.applyMatrix(
              right.x, up.x, look.x, 0.0f,
              right.y, up.y, look.y, 0.0f,
              right.z, up.z, look.z, 0.0f,
              0.0f, 0.0f, 0.0f, 1.0f
            );
        }
        for(float i = -hw; i < hw; i += xs){
            if(i > -hw + 2 && i < hw - 2){
                glc.vertex(i, -hh, 0.0f);
                glc.vertex(i, hh, 0.0f);
            }
        }
        for(float i = -hh; i < hh; i += ys){
            if(i > -hh + 2 && i < hh - 2){
                glc.vertex(-hw, i, 0.0f);
                glc.vertex(hw, i, 0.0f);
            }
        }
        glc.endShape(PGraphicsOpenGL.LINES);
        glc.popMatrix();
    }
    /**
     * Render 3D axes to the specified OpenGL renderer
     * @param glc OpenGL renderer
     * @param s size
     */
    public static void axes(PGraphicsOpenGL glc, float s){
        s *= 0.5;
        glc.pushMatrix();
        glc.noFill();
        glc.beginShape(PGraphicsOpenGL.LINES);
        // X
        glc.stroke(255, 0, 0);
        glc.vertex(-s, 0, 0);
        glc.vertex( s, 0, 0);
        // Y
        glc.stroke(0, 255, 0);
        glc.vertex(0, -s, 0);
        glc.vertex(0,  s, 0);
        // Z
        glc.stroke(0, 0, 255);
        glc.vertex(0, 0, -s);
        glc.vertex(0, 0,  s);
        glc.endShape();
        glc.popMatrix();
    }
    /**
     * Render a 3D box to the specified OpenGL renderer
     * @param glc OpenGL renderer
     * @param s size
     */
    public static void box(PGraphicsOpenGL glc, float s){
        s *= 0.5;
        glc.pushMatrix();
        // Top
        glc.beginShape(PGraphicsOpenGL.QUADS);
        glc.vertex(-s, -s, -s);
        glc.vertex( s, -s, -s);
        glc.vertex( s, -s,  s);
        glc.vertex(-s, -s,  s);
        glc.endShape(PGraphicsOpenGL.CLOSE);
        // Bottom
        glc.beginShape(PGraphicsOpenGL.QUADS);
        glc.vertex(-s,  s, -s);
        glc.vertex( s,  s, -s);
        glc.vertex( s,  s,  s);
        glc.vertex(-s,  s,  s);
        glc.endShape(PGraphicsOpenGL.CLOSE);
        // Struts
        glc.beginShape(PGraphicsOpenGL.LINES);
        glc.vertex(-s, -s, -s);
        glc.vertex(-s,  s, -s);
        glc.vertex( s, -s, -s);
        glc.vertex( s,  s, -s);
        glc.vertex( s, -s,  s);
        glc.vertex( s,  s,  s);
        glc.vertex(-s, -s,  s);
        glc.vertex(-s,  s,  s);
        glc.endShape();
        glc.popMatrix();
    }
}
