package pslib;
import java.io.*;
import java.text.NumberFormat;
/**
 * PSTools Class
 * <p>The PSTools class provides a range of static utility functions.
 * <p>This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * <p>You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see http://www.gnu.org/licenses/.
 * @author Stephen Callear
 */
public class PSTools extends Object{
    /**
     * Converts an input text file to an array of strings
     * <p>If the input file cannot be parsed a null value will be returned.
     * @param f file
     * @return string array
     */
    public static String [] readStrings(String f){
        try {
            return readStrings(new File(f));
        }
        catch (NullPointerException e){
            System.err.println("Error: loadStrings NullPointerException");
            return null;
        }
    }
    /**
     * Converts an input text file to an array of strings
     * <p>If the input file cannot be parsed a null value will be returned.
     * @param f file
     * @return string array
     */
    public static String [] readStrings(File f){
        FileReader fr;
        String [] es = null;
        try {
            fr = new FileReader(f);
        }
        catch (FileNotFoundException e) {
            System.err.println("Error: loadStrings FileNotFoundException - " + f.getName());
            return es;
        }
        BufferedReader br = new BufferedReader(fr);
        int line_count = 0;
        while (true){
            String line;
            try {
                line = br.readLine();
            }
            catch (IOException e){
                System.err.println("Error: loadStrings IOException at readLine()");
                return es;
            }
            if (line != null)
                line_count ++;
            else
                break;
        }
        try {
            fr.close();
            br.close();
            fr = new FileReader(f);
            br = new BufferedReader(fr);
        }
        catch (IOException e){
            return es;
        }
        String [] rs = new String[line_count];
        for (int i = 0; i < rs.length; i ++){
            try {
                rs[i] = br.readLine();
            }
            catch (IOException e){
                return es;
            }
        }
        return rs;
    }
    /**
     * Write an array of strings to the specified file
     * @param sa string array
     * @param f output file
     * @return true if the file was successfully written
     */
    public static boolean writeStrings(String[] sa, String f){
        try {
            return writeStrings(sa, new File(f));
        }
        catch (NullPointerException e){
            System.err.println("Error: writeStrings NullPointerException");
            return false;
        }
    }
    /**
     * Write an array of strings to the specified file
     * @param sa string array
     * @param f output file
     * @return true if the file was successfully written
     */
    public static boolean writeStrings(String[] sa, File f){
        FileWriter fw;
        try {
            fw = new FileWriter(f);
            for(String s : sa)
                if(s != null) fw.write(s + "\n");
            fw.flush();
            fw.close();
            return true;
        }
        catch (IOException e){
            System.err.println("Error : writeStrings IOException.");
            return false;
        }
    }
    /**
     * Combine all the strings within the specified array and return the result
     * @param s array of strings
     * @return a single combined string
     */
    public static String combineStrings(String[] s){
        return combineStrings(s, null);
    }
    /**
     * Combine all the strings within the specified array with the specified spacer
     * string and return the result
     * @param sa array of strings
     * @param ss spacer string
     * @return a single combined string
     */
    public static String combineStrings(String[] sa, String ss){
        StringBuilder str_out = new StringBuilder();
        for(int i = 0; i < sa.length; i ++){
            str_out.append(sa[i]);
            if(ss != null && i < sa.length - 1)
                str_out.append(ss);
        }
        return str_out.toString();
    }
    /**
     * Replaces instances of the character # within a string with a formatted
     * number
     * @param s a string
     * @param n an integer
     * @return the formatted string
     */
    public static String stringNumberFormat(String s, int n){
        return stringNumberFormat(s, '#', n);
    }
    /**
     * Replaces instances of a specified character within a string with a formatted
     * number
     * @param s a string
     * @param c the character to be replaced
     * @param n an integer
     * @return the formatted string
     */
    public static String stringNumberFormat(String s, char c, int n){
        int cnum = 0;
        for (int i = 0; i < s.length(); i ++){
            if (s.charAt(i) == c)
                cnum ++;
        }
        if (cnum > 0){
            NumberFormat nf = NumberFormat.getInstance();
            nf.setMinimumIntegerDigits(cnum);
            nf.setMaximumIntegerDigits(cnum);
            String sn = nf.format(n);
            String sf = s.substring(0, s.indexOf(c));
            String sa = s.substring(s.indexOf(c) + cnum, s.length());
            return sf + sn + sa;
        } else {
            return s;
        }
    }
    /**
     * Split a string using the specified tokens
     * @param s string
     * @param t tokens
     * @return string array
     */
    public static String[] splitTokens(String s, String t){
        char[] s_ch = s.toCharArray();
        char[] t_ch = t.toCharArray();
        if(t_ch.length < 1) return new String[] { s };
        for(int i = 0; i < s_ch.length; i ++)
            for(int j = 1; j < t_ch.length; j ++)
                if(s_ch[i] == t_ch[j]) s_ch[i] = t_ch[0];
        String str_m = String.valueOf(s_ch);
        String[] str_a = str_m.split(String.valueOf(t_ch[0]));
        return cleanStrings(str_a);
    }
    /**
     * Remove any empty entries from the specified string array
     * @param s string array
     * @return the formatted string array
     */
    public static String[] cleanStrings(String[] s){
        int c = 0;
        for(int i = 0; i < s.length; i ++)
            if(!s[i].isEmpty()) c ++;
        String[] s2 = new String[c];
        c = 0;
        for(int i = 0; i < s.length; i ++){
            if(!s[i].isEmpty()){
                s2[c] = s[i];
                c ++;
            }
        }
        return s2;
    }
    /**
     * Remove the specified characters from a string
     * @param s string
     * @param t tokens
     * @return the formatted string
     */
    public static String removeTokens(String s, String t){
        String[] str_arr = splitTokens(s, t);
        return combineStrings(str_arr);      
    }
    /**
     * Removes a file extension from a string assuming the file.extension format
     * @param s a file name as a string
     * @return the formatted string
     */
    public static String removeFileExtension(String s){
        String [] sa = s.split("\\.");
        StringBuilder sb = new StringBuilder();
        for (int i = 0; i < sa.length - 1; i ++){
            sb.append(sa[i]);
            if (i < sa.length - 2)
                sb.append(".");
        }
        return sb.toString();
    }
    /**
     * Remove the file name from the specified string assuming the string adopts
     * the File.separatorChar path character
     * @param s file
     * @return the formatted string
     */
    public static String removeFileName(String s){
        int i = s.lastIndexOf(File.separatorChar);
        char[] sc = new char[i + 1];
        s.getChars(0, i, sc, 0);
        sc[sc.length - 1] = File.separatorChar;
        return new String(sc);
    }
    /**
     * Removes quotations from a string by removing the first and last characters
     * @param str the string to be formatted
     * @return the formatted string
     */
    public static String stripQuotes(String str){
        char[] ca_old = str.toCharArray();
        char[] ca_new = new char[ca_old.length - 2];
        for (int i = 1; i < ca_old.length - 1; i ++)
            ca_new[i - 1] = ca_old[i];
        String str_out = new String(ca_new);
        return str_out;
    }
    /**
     * Removes all spaces within a string
     * @param str the string to be formatted
     * @return the formatted string
     */
    public static String removeWhiteSpace(String str){
        String[] str_arr = str.split("\\s");
        return combineStrings(str_arr);
    }    
    /**
     * Copy an array of strings and return the result
     * @param str string array
     * @return the copied string array
     */
    public static String[] copyStrings(String[] str){
        String[] out = new String[str.length];
        for(int i = 0; i < str.length; i ++)
            out[i] = str[i].toString();
        return out;
    }
    /**
     * Constrain an integer value between the specified minimum and maximum
     * @param value input value
     * @param min lower limit
     * @param max upper limit
     * @return constrained value
     */
    public static int constrainInt(int value, int min, int max){
        if(value < min) return min;
        else if(value > max) return max;
        else return value;
    }
    /**
     * Constrain a float value between the specified minimum and maximum
     * @param value input value
     * @param min lower limit
     * @param max upper limit
     * @return constrained value
     */
    public static float constrainFloat(float value, float min, float max){
        if(value < min) return min;
        else if(value > max) return max;
        else return value;
    }
    /**
     * Return a Colour object with random component values
     * @return a Colour object with random component values
     */
    public static PSColour randomColour(){
        return randomColour(false);
    }
    /**
     * Return a Colour object with random component values
     * @param alpha alpha component toggle (true = random alpha)
     * @return a Colour object with random component values
     */
    public static PSColour randomColour(boolean alpha){
        int[] rgba = new int[4];
        for(int i = 0; i < rgba.length; i ++)
            rgba[i] = (int) (Math.random() * 255d);
        if(alpha)
            return new PSColour(rgba[0], rgba[1], rgba[2], rgba[3]);
        else
            return new PSColour(rgba[0], rgba[1], rgba[2]);
    }   
    /**
     * Return a normalised PSVector object with random component values
     * @return a PSVector object with random component values
     */
    public static PSVector randomVector(){
        float[] xyz = new float[3];
        for(int i = 0; i < xyz.length; i ++)
            xyz[i] = (float) Math.random() * 2f - 1f;
        PSVector v = new PSVector(xyz);
        v.normalise();
        return v;
    }
    /**
     * Return a PSVector object with random component values within the specified
     * radius about the point (0, 0, 0)
     * @param r radius
     * @return a PSVector object with random component values
     */
    public static PSVector randomVector(float r){
        return randomVector(0f, r);
    }
    /**
     * Return a PSVector object with random component values within the specified
     * radii about the point (0, 0, 0)
     * @param rmi minimum radius
     * @param rma maximum radius
     * @return a PSVector object with random component values
     */
    public static PSVector randomVector(float rmi, float rma){
        PSVector v = randomVector();
        float m = ((float) Math.pow(Math.random(), 1d / 3d) * rma - rmi) + rmi;
        v.scale(m);
        return v;
    }
    /**
     * Return a PSVector object with random component values within +/- the specified
     * range about the point (0, 0, 0)
     * @param r range
     * @return a PSVector object with random component values
     */
    public static PSVector randomVector(PSVector r){
        float[] ra = r.toArray();
        float[] xyz = new float[3];
        for(int i = 0; i < xyz.length; i ++)
            xyz[i] = (float) Math.random() * 2f * ra[i] - ra[i];
        return new PSVector(xyz);
    }      
}
