package pslib;
/**
 * PSVector Class
 * <p>The PSVector object specifies a location within xyz coordinate space with
 * floating point precision.
 * <p>This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * <p>You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see http://www.gnu.org/licenses/.
 * @author Stephen Callear
 */
public class PSVector extends Object {
    public float x, y, z;
    /**
     * Create a new object instance with zero values
     */
    public PSVector(){
        set(0f, 0f, 0f);
    }
    /**
     * Create a new object instance with the specified values
     * @param x x
     * @param y y
     */
    public PSVector(float x, float y){
        set(x, y);
    }
    /**
     * Create a new object instance with the specified values
     * @param x x
     * @param y y
     * @param z z
     */
    public PSVector(float x, float y, float z){
        set(x, y, z);
    }
    /**
     * Create a new object instance using values from the specified array
     * @param arr array
     */
    public PSVector(float[] arr){
        set(arr);
    }
    /**
     * Create a new object instance using values from the specified vector
     * @param v vector
     */
    public PSVector(PSVector v){
        set(v);
    }
    /**
     * Set the x and y values
     * @param x x
     * @param y y
     */
    public final void set(float x, float y){
        set(x, y, 0.0f);
    }
    /**
     * Set the component values
     * @param x x
     * @param y y
     * @param z z
     */
    public final void set(float x, float y, float z){
        this.x = x;
        this.y = y;
        this.z = z;
    }
    /**
     * Set the component values using values from the specified array
     * @param arr array
     */
    public final void set(float[] arr){
        float[] xyz = parseArray(arr);
        x = xyz[0];
        y = xyz[1];
        z = xyz[2];
    }
    /**
     * Set the component values using values from the specified vector
     * @param v
     */
    public final void set(PSVector v){
        x = v.x;
        y = v.y;
        z = v.z;
    }
    /**
     * Return a copy of this vector
     * @return a copy of this vector
     */
    public PSVector copy(){
        PSVector v = new PSVector(x, y, z);
        return v;
    }
    /**
     * Return the component values of this vector as an array
     * @return the component values of this vector as an array
     */
    public final float[] toArray(){
        float[] a = { x, y, z };
        return a;
    }
    /**
     * Add the specified values to this vector
     * @param x x
     * @param y y
     */
    public final void add(float x, float y){
        add(x, y, 0f);
    }
    /**
     * Add the specified values to this vector
     * @param x x
     * @param y y
     * @param z z
     */
    public final void add(float x, float y, float z){
        this.x += x;
        this.y += y;
        this.z += z;
    }
    /**
     * Add the specified vector to this vector
     * @param v vector
     */
    public final void add(PSVector v){
        x += v.x;
        y += v.y;
        z += v.z;
    }
    /**
     * Add the two specified vectors and return the result
     * @param v1 vector 1
     * @param v2 vector 2
     * @return the result of v1 + v2
     */
    public final static PSVector add(PSVector v1, PSVector v2){
        PSVector v = new PSVector(v1.x + v2.x, v1.y + v2.y, v1.z + v2.z);
        return v;
    }
    /**
     * Subtract the specified values from this vector
     * @param x x
     * @param y y
     */
    public final void sub(float x, float y){
        sub(x, y, 0.0f);
    }
    /**
     * Subtract the specified values from this vector
     * @param x x
     * @param y y
     * @param z z
     */
    public final void sub(float x, float y, float z){
        this.x -= x;
        this.y -= y;
        this.z -= z;
    }
    /**
     * Subtract the specified vector from this vector
     * @param v vector
     */
    public final void sub(PSVector v){
        x -= v.x;
        y -= v.y;
        z -= v.z;
    }
    /**
     * Subtract one vector from another and return the result
     * @param v1 vector 1
     * @param v2 vector 2
     * @return the result of v1 - v2
     */
    public static PSVector sub(PSVector v1, PSVector v2){
        PSVector v = new PSVector(v1.x - v2.x, v1.y - v2.y, v1.z - v2.z);
        return v;
    }
    /**
     * Multiply this vector by the specified scalar
     * @param s scalar
     */
    public final void scale(float s){
        x *= s;
        y *= s;
        z *= s;
    }
    /**
     * Multiply the component values of this vector by those within the specified
     * vector
     * @param v vector
     */
    public final void scale(PSVector v){
        x *= v.x;
        y *= v.y;
        z *= v.z;
    }
    /**
     * Multiply a vector by the specified scalar and return the result as a new
     * vector
     * @param v vector
     * @param s scalar
     * @return the result of v1 * s as a new vector
     */
    public static PSVector scale(PSVector v, float s){
        PSVector vs = v.copy();
        vs.scale(s);
        return vs;
    }
    /**
     * Multiply two vectors and return the result as a new vector
     * @param v1
     * @param v2
     * @return the result of v1 * v2 as a new vector
     */
    public static PSVector scale(PSVector v1, PSVector v2){
        PSVector vs = v1.copy();
        vs.scale(v2);
        return vs;
    }
    /**
     * Limit this vector to the specified magnitude
     * @param m magnitude
     */
    public final void limit(float m){
        float s = m / magnitude();
        if(s < 1f) scale(s);
    }
    /**
     * Limit a vector to the specified magnitude and return the result as a new vector
     * @param v vector
     * @param m magnitude
     * @return the limited result as a new vector
     */
    public static PSVector limit(PSVector v, float m){
        PSVector vl = v.copy();
        vl.limit(m);
        return vl;
    }
    /**
     * Normalise this vector
     */
    public final void normalise(){
        normaliseTo(1f);
    }
    /**
     * Normalise a vector and return the result as a new vector
     * @param v vector
     * @return the normalised result as a new vector
     */
    public static PSVector normalise(PSVector v){
        PSVector vn = v.copy();
        vn.normalise();
        return vn;
    }
    /**
     * Normalise this vector to the specified magnitude
     * @param m magnitude
     */
    public final void normaliseTo(float m){
        if(magnitude() == 0) return;
        float s = m / magnitude();
        scale(s);
    }
    /**
     * Normalise a vector to the specified magnitude and return the result as a
     * new vector
     * @param v vector
     * @param m magnitude
     * @return the normalised result as a new vector
     */
    public static PSVector normaliseTo(PSVector v, float m){
        PSVector vn = v.copy();
        vn.normaliseTo(m);
        return vn;
    }
    /**
     * Return the magnitude of this vector
     * @return the magnitude of this vector
     */
    public final float magnitude(){
        double m = Math.sqrt(Math.pow(x, 2) + Math.pow(y, 2) + Math.pow(z, 2));
        return (float) m;
    }
    /**
     * Return the distance between this vector and that specified
     * @param v vector
     * @return the distance between this vector and that specified
     */
    public final float distance(PSVector v){
        double d = Math.sqrt(Math.pow(v.x - x, 2) + Math.pow(v.y - y, 2) + Math.pow(v.z - z, 2));
        return (float) d;
    }
    /**
     * Return the dot product of this vector and that specified
     * @param v vector
     * @return the dot product of this vector and that specified
     */
    public final float dot(PSVector v){
        return x * v.x + y * v.y + z * v.z;
    }
    /**
     * Return the cross product of this vector and that specified
     * @param v vector
     * @return the cross product of this vector and that specified
     */
    public final PSVector cross(PSVector v){
        PSVector vc = new PSVector(y * v.z - v.y * z,
                                   z * v.x - v.z * x,
                                   x * v.y - v.x * y);
        return vc;
    }
    /**
     * Return a normalised vector perpendicular to this one
     * @return a normalised vector perpendicular to this one
     */
    public final PSVector getPerpendicular(){
        // Generate an abs value copy of this vector
        float[] abs = toArray();
        for(int i = 0; i < abs.length; i ++)
            abs[i] = Math.abs(abs[i]);
        // Find the axis with the smallest magnitude
        float value = abs[0];
        int index = 0;
        for(int i = 0; i < abs.length; i ++){
            if(abs[i] < value){
                value = abs[i];
                index = i;
            }           
        }
        // Generate axis vector array
        PSVector[] dv = { new PSVector(1f, 0f, 0f),
                          new PSVector(0f, 1f, 0f),
                          new PSVector(0f, 0f, 1f) };
        // Cross this vector with the axis of least magnitude
        PSVector pv = cross(dv[index]);
        pv.normalise();
        return pv;
    }
    /**
     * Return the angle between this vector and that specified
     * @param v vector
     * @return the angle between this vector and that specified
     */
    public final float angleBetween(PSVector v){
        double a = Math.acos((double) dot(v));
        return (float) a;
    }
    /**
     * Rotate this vector about the origin on the specified axis
     * @param a angle
     * @param av axis vector
     */
    public final void rotateAboutOrigin(float a, PSVector av){
        // Credit: http://www.blitzbasic.com/Community/posts.php?topic=57616
        float u, v, w;
        PSVector vn = PSVector.normalise(av);
        u = vn.x;
        v = vn.y;
        w = vn.z;
        float ux, uy, uz, vx, vy, vz, wx, wy, wz, sa, ca;
        ux = u * x;
        uy = u * y;
        uz = u * z;
        vx = v * x;
        vy = v * y;
        vz = v * z;
        wx = w * x;
        wy = w * y;
        wz = w * z;

        sa = (float) Math.sin((double) a);
        ca = (float) Math.cos((double) a);

        x = u * (ux + vy + wz) + (x * (v * v + w * w) - u * (vy + wz)) * ca + (-wy + vz) * sa;
        y = v * (ux + vy + wz) + (y * (u * u + w * w) - v * (ux + wz)) * ca + ( wx - uz) * sa;
        z = w * (ux + vy + wz) + (z * (u * u + v * v) - w * (ux + vy)) * ca + (-vx + uy) * sa;
        
    }
    /**
     * Rotate a vector the origin on the specified axis and return the result as
     * a new vector
     * @param v vector to be rotated
     * @param a angle
     * @param av axis vector
     * @return the result as a new vector
     */
    public static PSVector rotateAboutOrigin(PSVector v, float a, PSVector av){
        PSVector vn = v.copy();
        vn.rotateAboutOrigin(a, av);
        return vn;
    }
    /**
     * Return the pitch of the vector relative to the origin in radians
     * @return the pitch of the vector relative to the origin in radians
     */
    public final float pitch(){
        return (float) Math.atan2((double) z, Math.sqrt(Math.pow((double) x, 2) + Math.pow((double) y, 2)));
    }
    /**
     * Return the yaw of the vector relative to the origin in radians
     * @return the yaw of the vector relative to the origin in radians
     */
    public final float yaw(){
        return (float) Math.atan2((double) x, (double) y);
    }
    /**
     * Return the vector component values as a string
     * @return string
     */
    @Override public String toString(){
        return "[" + x + ", " + y + ", " + z + "]";
    }
    /**
     * Derive an array of length 3 from that specified
     */
    private float[] parseArray(float[] arr){
        float[] xyz = { 0f, 0f, 0f };
        for(int i = 0; i < arr.length && i < 3; i ++)
            xyz[i] = arr[i];
        return xyz;
    
    }
    
}
