package pslib;
import processing.opengl.PGraphicsOpenGL;
import java.util.Arrays;
/**
 * ParentParticle Class
 * <p>The ParentParticle object allows multiple DynamicParticle objects
 * to be attached to a single parent particle. Attached objects inherit the
 * position and velocity properties of the parent.
 * <p>This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * <p>You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see http://www.gnu.org/licenses/.
 * @author Stephen Callear
 */
public class ParentParticle extends DynamicParticle {
    private DynamicParticle[] buffer;
    /**
     * Create a new object instance with the specified buffer size
     * @param b buffer size
     */
    public ParentParticle(int b){
        this(b, 0);
    }
    /**
     * Create a new object instance with the specified buffer size and decay time
     * @param b buffer size
     * @param d decay (ms)
     */
    public ParentParticle(int b, int d){
        super(d);
        buffer = new DynamicParticle[b];
    }
    /**
     * Set the position of this particle and any attached particles
     * @param v position vector
     */
    @Override public void setPosition(PSVector v){
        super.setPosition(v);
        for(DynamicParticle p : buffer)
            if(p != null && !p.isDead()) p.setPosition(v);
    }
    /**
     * Set the velocity of this particle and any attached particles
     * @param v velocity vector
     */
    @Override public void setVelocity(PSVector v){
        super.setVelocity(v);
        for(DynamicParticle p : buffer)
            if(p != null && !p.isDead()) p.setVelocity(v);
    }
    /**
     * Set the colour of all attached particles
     * @param c colour
     */
    @Override public void setColour(PSColour c){
        super.setColour(c);
        for(DynamicParticle p : buffer)
            if(p != null && !p.isDead()) p.setColour(c);
    }
    /**
     * Set the colour of all attached particles with the specified fade time
     * @param c colour
     * @param t fade time (ms)
     */
    @Override public void setColour(PSColour c, int t){
        super.setColour(c, t);
        for(DynamicParticle p : buffer)
            if(p != null && !p.isDead()) p.setColour(c, t);
    }
    /**
     * Set the colour of all attached particles
     * <p>Particle colour will be assigned at random from the specified WeightedList
     * @param c colour list
     */
    public void setColour(WeightedList<PSColour> c){
        super.setColour(c.get());
        for(DynamicParticle p : buffer)
            if(p != null && !p.isDead()) p.setColour(c.get());
    }
    /**
     * Set the colour of all attached particles with the specified fade time
     * <p>Particle colour will be assigned at random from the specified WeightedList
     * @param c colour list
     * @param t fade time (ms)
     */
    public void setColour(WeightedList<PSColour> c, int t){
        super.setColour(c.get(), t);
        for(DynamicParticle p : buffer)
            if(p != null && !p.isDead()) p.setColour(c.get(), t);
    }
    /**
     * Set the alpha value of all attached particles
     * @param a alpha
     */
    @Override public void setAlpha(float a){
        super.setAlpha(a);
        for(DynamicParticle p : buffer)
            if(p != null && !p.isDead()) p.setAlpha(a);
    }
    /**
     * Return the buffer size
     * @return the buffer size
     */
    public final int getBufferSize(){
        return buffer.length;
    }
    /**
     * Return the total attached particles     
     * @return the total attached particles
     */
    public final int attachedParticles(){
        int c = 0;
        for(DynamicParticle p : buffer)
            if(p != null && !p.isDead()) c++;
        return c;
    }
    /**
     * Set the buffer size
     * <p>If the new size is less than the current active particle count, particles
     * will be culled eldest first.
     * @param s buffer size
     */
    public final void setBufferSize(int s){
        if(s == buffer.length) return;
        DynamicParticle[] arr = toArray();
        buffer = new DynamicParticle[s];
        int bs = arr.length;
        if(s < arr.length){
            Arrays.sort(arr);
            bs = s;
        }
        System.arraycopy(arr, 0, buffer, 0, bs);
    }
    /**
     * Return the amount of active particles
     * @return the amount of active particles
     */
    public final int activeParticles(){
        int c = 0;
        for(int i = 0; i < buffer.length; i ++)
            if(buffer[i] != null && !buffer[i].isDead()) c ++;
        return c;
    }
    /**
     * Return all active particles as an array
     * @return all active particles as an array
     */
    public final DynamicParticle[] toArray(){
        DynamicParticle[] arr = new DynamicParticle[activeParticles()];
        int j = 0;
        for(int i = 0; i < buffer.length; i ++){
            if(buffer[i] != null && !buffer[i].isDead()){
                arr[j] = buffer[i];
                j ++;
            }
        }
        return arr;
    }
    /**
     * Return true if this particle and all attached particles are dead
     * @return trues if this particle and all attached particles are dead
     */
    @Override public boolean isDead(){
        if((super.isDead() || !super.isDynamic()) && activeParticles() == 0) return true;
        return false;
    }
    /**
     * Attach a particle to the buffer
     * @param p particle
     * @return true if the particle was successfully attached
     */
    public boolean attach(DynamicParticle p){
        int i = getEmptyIndex();
        if(i < 0 || super.isDead()) return false;
        p.setPosition(getPosition());
        p.setVelocity(getVelocity());
        buffer[i] = p;
        return true;
    }
    /**
     * Attach an array of particles to the buffer
     * @param p particle
     * @return the amount of particles attached
     */
    public int attach(DynamicParticle[] p){
        int c = 0;
        for(int i = 0; i < p.length; i ++)
            if(attach(p[i])) c ++;
        return c;
    }
    /**
     * Return the next available buffer index
     * @return the next available buffer index
     */
    private int getEmptyIndex(){
        for(int i = 0; i < buffer.length; i ++)
            if(buffer[i] == null || buffer[i].isDead()) return i;
        return -1;
    }
    /**
     * Remove a particle from the buffer
     * <p>If multiple references to the particle exist, they will all be removed.
     * @param p DynamicParticle
     * @return true if the particle was found and removed
     */
    public boolean remove(DynamicParticle p){
        boolean b = false;
        for(int i = 0; i < buffer.length; i ++){
            if(buffer[i] == p){
                buffer[i] = null;
                b = true;
            }
        }
        return b;
    }
    /**
     * Clear the buffer
     */
    public final void clear(){
        for(int i = 0; i < buffer.length; i ++)
            buffer[i] = null;
    }
    /**
     * Update this and all attached particles
     */
    @Override public void update(){
        super.update();        
        for(int i = 0; i < buffer.length; i ++)
            if(buffer[i] != null) buffer[i].update();
    }
    /**
     * Render all attached particles to the specified OpenGL renderer
     * @param pgl OpenGL renderer
     */
    @Override public void render(PGraphicsOpenGL pgl){
        render(pgl, null, null, null);
    }
    /**
     * Render all attached particles to the specified OpenGL renderer
     * @param pgl OpenGL renderer
     * @param cpv camera position vector
     * @param ctv camera target vector
     * @param cuv camera up vector;
     */
    @Override public void render(PGraphicsOpenGL pgl, PSVector cpv, PSVector ctv, PSVector cuv){
        for(DynamicParticle p : buffer){
            if(p != null && !p.isDead())
                p.render(pgl, cpv, ctv, cuv);
        }
    }
    /**
     * Render the particle location to the specified OpenGL renderer
     * @param pgl OpenGL renderer
     */
    @Override public void debug(PGraphicsOpenGL pgl){
        debug(pgl, null, null, null);
    }
    /**
     * Render the particle location to the specified OpenGL renderer
     * @param pgl OpenGL renderer
     * @param cpv camera position vector
     * @param ctv camera target vector
     * @param cuv camera up vector
     */
    @Override public void debug(PGraphicsOpenGL pgl, PSVector cpv, PSVector ctv, PSVector cuv){
        super.debug(pgl, cpv, ctv, cuv);
    }
    /**
     * Return a copy of this particle
     * <p>The returned object will contain a copy of all attached particles within
     * this object and will adhere to the rules defined in the <copy>DynamicParticle.map()
     * </code> method.
     * @see DynamicParticle
     * @return a copy of this particle
     */
    @Override public ParentParticle copy(){
        ParentParticle p = new ParentParticle(buffer.length);        
        map(p);
        return p;
    }
    /**
     * Map the properties of this particle to that specified.
     * <p>The returned object takes a copy of all attached particles within this
     * object. DynamicParticle variables are mapped based on the rules defined in
     * <code>DynamicParticle.map()</code>.
     * @see DynamicParticle
     * @param p particle
     */
    public final void map(ParentParticle p){
        super.map(p);
        p.clear();
        if(p.getBufferSize() != buffer.length)
            p.setBufferSize(buffer.length);
        for(DynamicParticle dp : buffer)
            if(dp != null && !dp.isDead()) p.attach(dp.copy());
    }
}
