package pslib;
import processing.opengl.*;
import java.util.ArrayList;
import java.util.Arrays;
/**
 * Ribbon Class
 * <p>The Ribbon object extends the functionality of the buffered particle, by
 * incorporating ribbon rendering functionality.
 * <p>This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * <p>You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see http://www.gnu.org/licenses/.
 * @author Stephen Callear
 */
public class RibbonParticle extends BufferedParticle {
    private float size_taper, alpha_taper, size_curve, alpha_curve, cap_size;
    /**
     * Create a new object instance with the specified buffer size
     * @param b buffer size
     */
    public RibbonParticle(int b) {
        this(b, 0);
    }
    /**
     * Create a new object instance with the specified buffer size and decay
     * @param b buffer size
     * @param d decay (ms)
     */
    public RibbonParticle(int b, int d) {
        super(b, d);
        size_taper = 1.0f;
        size_curve = 1.0f;
        alpha_taper = 1.0f;
        alpha_curve = 1.0f;
        cap_size = 0.25f;
    }
    /**
     * Set the ribbon size taper value
     * <p>
     * Default = 1.0
     * @param t taper
     */
    public final void setSizeTaper(float t){
        size_taper = PSTools.constrainFloat(t, 0.0f, 1.0f);
    }
    /**
     * Set the ribbon size taper curve
     * <p>
     * Default = 1.0
     * @param c curve
     */
    public final void setSizeTaperCurve(float c){
        size_curve = c;
    }
    /**
     * Set the ribbon alpha taper value
     * <p>
     * Default = 1.0
     * @param t taper
     */
    public final void setAlphaTaper(float t){
        alpha_taper = PSTools.constrainFloat(t, 0.0f, 1.0f);
    }
    /**
     * Set the cap length multiplier
     * <p>
     * Default = 0.25
     * @param l length
     */
    public final void setCapSize(float l){
        cap_size = l;
    }
    /**
     * Set the ribbon alpha taper curve
     * <p>
     * Default = 1.0
     * @param c curve
     */
    public final void setAlphaTaperCurve(float c){
        alpha_curve = c;
    }
    /**
     * Render the ribbon to the specified OpenGL renderer
     * @param pgl OpenGL renderer
     */
    @Override public void render(PGraphicsOpenGL pgl){
        render(pgl, null, null, null);
    }
    /**
     * Render the ribbon to the specified OpenGL renderer
     * @param pgl OpenGL renderer
     * @param cpv camera position vector
     * @param ctv camera target vector
     * @param cuv camera up vector
     */
    @Override public void render(PGraphicsOpenGL pgl, PSVector cpv, PSVector ctv, PSVector cuv){
        PSVertex[] vlist = getRibbonVertices(cpv);
        pgl.noStroke();
        pgl.pushMatrix();
        pgl.translate(getPosition().x, getPosition().y, getPosition().z);
        pgl.beginShape(PGraphicsOpenGL.TRIANGLE_STRIP);
        // Render PSVertex list backwards
        for(int i = vlist.length - 1; i >= 0; i --){
            pgl.fill(vlist[i].colour.getRed(), vlist[i].colour.getGreen(), vlist[i].colour.getBlue(), vlist[i].colour.getAlpha());
            pgl.vertex(vlist[i].position.x, vlist[i].position.y, vlist[i].position.z);
        }
        pgl.endShape();
        pgl.popMatrix();
    }
    /**
     * Return a copy of this particle
     * <p>The returned copy will use the size and alpha taper variables of the source
     * particle and will adhere to the rules defined in both <code>BufferedParticle.map()
     * </code> and <code>DynamicParticle.map()</code>.
     * @see BufferedParticle
     * @see DynamicParticle
     * @return a copy of this particle
     */
    @Override public RibbonParticle copy(){
        RibbonParticle p = new RibbonParticle(getBuffer().length);
        map(p);
        return p;
    }
    /**
     * Map the properties of this particle to that specified preserving references
     * <p>The returned object takes a copy of all ribbon particle variables.
     * BufferedParticle and DynamicParticle variables are mapped based on the rules
     * defined in <code>BufferedParticle.map()</code> and <code>DynamicParticle.map()
     * </code> respectively.
     * @see BufferedParticle
     * @see DynamicParticle
     * @param p particle
     */
    public final void map(RibbonParticle p){        
        p.setSizeTaper(size_taper);
        p.setSizeTaperCurve(size_curve);
        p.setAlphaTaper(alpha_taper);
        p.setCapSize(cap_size);
        super.map(p);
    }
    /**
     * Calculate the ribbon vertices and return as an array of PSVertex objects
     * @param cpv camera position vector
     * @return the ribbon vertices as an array of PSVertex objects
     */
    protected final PSVertex[] getRibbonVertices(PSVector cpv){
        // Get buffer data
        StaticParticle[] pb = getBuffer();
        PSVector[] rp = getRelativePositions();
        PSVector[] cp = getCumulativePositions();
        int v_res = 2; // Reserved vertices
        if(cap_size > 0) v_res += 3; // Cap vertices
        ArrayList vlist = new ArrayList(v_res);

        // Calculate ribbon taper
        float[] tl = new float[cp.length];
        float rl = 0;
        for(int i = 0; i < cp.length; i ++){
            rl += cp[i].magnitude();
            tl[i] = rl;
        }
        for(int i = 0; i < tl.length; i ++)
            tl[i] = 1.0f - tl[i] / rl;
        // Calculate buffered element vertices      
        PSVector n, n0, n1, n2, blv;
        PSVector cn = new PSVector();
        PSVector cd = new PSVector();
        PSColour c;
        int vcount = v_res;
        float uvy = 0f;
        for(int i = 0; i < pb.length; i ++){
            // Calculate normal
            int j = i;
            n0 = new PSVector();
            while(n0.magnitude() == 0.0f){
                n0.set(cp[j]);
                j ++;
                if(j >= pb.length) j = 0;
                if(j == i) break;
            }
            // Derive camera vectors
            if(cpv == null) blv = CAM2D_LOOK;
            else blv = PSVector.sub(pb[i].position, cpv);
            blv.normalise();
            // Billboarding borrowed from http://v8media.com/processing/decibel/
            n1 = n0.cross(blv);
            n1.normalise();
            n2 = n0.cross(n1);
            n2.normalise();
            n = n0.cross(n2);
            n.normalise();
            if(i == 0){ // Save cap vectors for later...
                cn = n.copy();
                cd = n0.copy();
                cd.normalise();
            }
            // Colour           
            float at = (float) Math.pow(tl[i], alpha_curve) * alpha_taper + (1.0f - alpha_taper); // Alpha taper
            c = pb[i].colour.copy();
            c.setAlpha(Math.round((float) pb[i].colour.getAlpha() * getRenderAlpha() * at));
            // Vertices
            float st = (float) Math.pow(tl[i], size_curve) * size_taper + (1.0f - size_taper); // Size taper
            n.scale(pb[i].size * st * 0.5f);
            uvy = uvy == 0f ? 1f : 0f;
            vlist.add(new PSVertex(vcount    , PSVector.sub(rp[i], n), new PSVector(0f, uvy), c));
            vlist.add(new PSVertex(vcount + 1, PSVector.add(rp[i], n), new PSVector(1f, uvy), c));
            vcount += 2;
        }
        c = getColour().copy();
        c.setAlpha(Math.round((float)getColour().getAlpha() * getRenderAlpha()));
        // Add unbuffered vertices
        uvy = uvy == 0f ? 1f : 0f;
        vlist.add(new PSVertex(v_res - 1, PSVector.scale(cn, getRenderSize() * -0.5f), new PSVector(0f, uvy), c));
        vlist.add(new PSVertex(v_res - 2, PSVector.scale(cn, getRenderSize() *  0.5f), new PSVector(1f, uvy), c));
        // Add cap vertices
        if(cap_size > 0.0f){
            vlist.add(new PSVertex(0, PSVector.scale(cn, getRenderSize() * -0.5f),     new PSVector(0f, 1f),   c));
            vlist.add(new PSVertex(1, PSVector.scale(cd, getRenderSize() * -cap_size), new PSVector(0.5f, 0f), c));
            vlist.add(new PSVertex(2, PSVector.scale(cn, getRenderSize() *  0.5f),     new PSVector(1f, 1f),   c));
        }
        // Return PSVertex array
        PSVertex[] varray = new PSVertex[vlist.size()];
        vlist.toArray(varray);
        Arrays.sort(varray); // Sort vertices by index value to preserve ordering
        return varray;
    }
}

