package pslib;
/**
 * TargetEmitter Class
 * <p>The TargetEmitter class extends the functionality of the GenericEmitter
 * class by incorporating target seeking; attached particles will seek out the
 * nearest attached target.
 * <p>This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * <p>You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see http://www.gnu.org/licenses/.
 * @author Stephen Callear
 */
public class TargetEmitter extends GenericEmitter {
    private AttractiveForce[] targets;
    /**
     * Create a new object instance with the specified buffer size
     * @param b buffer size
     */
    public TargetEmitter(int b){
        this(b, 0);
    }
    /**
     * Create a new object instance with the specified buffer size and decay time
     * @param b buffer size
     * @param d decay (ms)
     */
    public TargetEmitter(int b, int d){
        super(b, d);        
    }
    /**
     * Set the targets for this object
     * @param t targets
     */
    public final void setTargets(AttractiveForce[] t){
        targets = t;
    }
    /**
     * Return a reference to this objects targets
     * @return a reference to this objects targets
     */
    public final AttractiveForce[] getTargets(){
        return targets;
    }
    
    /**
     * Return a copy of this emitter
     * <p>The returned copy contains references to the target vectors within
     * the source emitter and will adhere to the rules defined in the
     * GenericEmitter map() function.
     * @see GenericEmitter
     * @return a copy of this emitter
     */
    @Override public TargetEmitter copy(){
        TargetEmitter e = new TargetEmitter(getBufferSize());
        e.setTargets(targets);
        super.map(e);
        return e;
    }
    /**
     * Update the specified particle
     * @param p particle
     */
    @Override protected void updateParticle(DynamicParticle p){
        // Targets
        AttractiveForce tf = getClosestTarget(p);
        if(tf != null) tf.update(p);
        // GenericEmitter updates
        super.updateParticle(p);
    }
    private AttractiveForce getClosestTarget(DynamicParticle p){
        if(targets == null || targets.length < 1) return null;
        int index = 0;       
        float dt, d = targets[0].getPosition().distance(p.getPosition());
        for(int i = 1; i < targets.length; i ++){
            dt = targets[i].getPosition().distance(p.getPosition());
            if(dt < d){
                index = i;
                d = dt;
            }
        }
        return targets[index];
    }
}
