package pslib;
import java.util.Random;
import java.util.ArrayList;
/**
 * TemplateEmission Class
 * <p>The TemplateEmission class implements the PSEmission interface allowing the
 * automated emission of random particles derived from the object variables from
 * any emitters to which it is assigned.
 * <p>Unlike the VariantEmission class, individual particle parameters may have
 * independent variance and may be subject to modification after creation.
 * <p>This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * <p>You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see http://www.gnu.org/licenses/.
 * @author Stephen Callear
 * @see VariantEmission
 */
public class TemplateEmission extends Object implements PSEmission {
    private WeightedList<DynamicParticle> template;
    private WeightedList<PSColour> palette;
    private ArrayList<ParticleListener> listeners;
    private float size, alpha, size_v, alpha_v, decay_v;
    private int probability, min_parts, max_parts, decay;
    private boolean active;
    private Random random;
    /**
     * Create a new object instance with the specified variables
     * <p>Default decay, size and alpha values are taken from the specified particle
     * with 0% variance.
     * @param t particle template
     * @param ep emission probability
     * @param mi minimum emitted particles
     * @param ma maximum emitted particles
     */
    public TemplateEmission(DynamicParticle t, int ep, int mi, int ma){
        this(new WeightedList<DynamicParticle>(new DynamicParticle[] { t }), ep, mi, ma);
    }
    /**
     * Create a new object instance with the specified variables
     * <p>Default decay, size and alpha values are taken from a particle selected
     * at random from the specified list with 0% variance.
     * @param t particle template list
     * @param ep emission probability
     * @param mi minimum emitted particles
     * @param ma maximum emitted particles
     */
    public TemplateEmission(WeightedList<DynamicParticle> t, int ep, int mi, int ma){
        template = t;
        palette = null;
        probability = ep;
        min_parts = mi;
        max_parts = ma;
        active = true;
        random = new Random();
        listeners = new ArrayList<ParticleListener>();
        // Set base parameters from randomly selected particle
        DynamicParticle p = template.get();
        decay = p.getInitialDecayTime();
        size = p.getSize();
        alpha = p.getAlpha();
        decay_v = 0f;
        size_v = 0f;
        alpha_v = 0f;
    }
    /**
     * Set the particle template
     * <p>The particle template specifies the list of particles from which emitted
     * particles may be derived. If no palette is specified emitted particles
     * will inherit their colour from the template.
     * @param t particle template
     */
    public final void setTemplate(WeightedList<DynamicParticle> t){
        template = t;
    }
    /**
     * Set the particle colour palette
     * <p>The particle colour palette specifies the list of colours from which
     * emitted particles may select. If no palette is specified emitted particles
     * will inherit their colour from the particle template.
     * @param p colour palette
     */
    public final void setPalette(WeightedList<PSColour> p){
        palette = p;
    }
    /**
     * Set the active state of the behaviour
     * <p>A value of true will result in particle emission on the update routine
     * of any emitters to which this behaviour is attached.
     * <p>Default: true
     * @param a
     */
    public final void setActive(boolean a){
        active = a;
    }
    /**
     * Set the emission probability
     * <p>The emission probability specifies the probability of particle creation
     * in the format 1:probability.
     * @param p emission probability
     */
    public final void setEmissionProbability(int p){
        probability = p;
    }
    /**
     * Set the amount of particles emitted on each emission event
     * @param p particles
     */
    public final void setEmittedParticles(int p){
        min_parts = p;
        max_parts = p;
    }
    /**
     * Set the amount of particles emitted on each emission event
     * <p>The actual amount of emitted particles will be derived at random from
     * the specified range.
     * @param min minimum particles
     * @param max maximum particles
     */
    public final void setEmittedParticles(int min, int max){
        min_parts = min;
        max_parts = max;
    }
    /**
     * Set the emitted particle decay value and variance
     * @param d decay (ms)
     * @param v variance (+/- %)
     */
    public final void setDecay(int d, int v){
        decay = d;
        decay_v = (float) v / 100f;
    }
    /**
     * Set the emitted particle size value and variance
     * @param s size
     * @param v variance (+/- %)
     */
    public final void setSize(float s, int v){
        size = s;
        size_v = (float) v / 100f;
    }
    /**
     * Set the emitted particle alpha value and variance
     * @param a alpha
     * @param v variance (+/- %)
     */
    public final void setAlpha(float a, int v){
        alpha = a;
        alpha_v = (float) v / 100f;
    }
    /**
     * Update the specified emitter
     * <p>The update routine attaches random particles derived from the object
     * variables to the specified emitter. References to created particles are
     * sent to any attached listeners.
     * @param e emitter
     */
    public void update(GenericEmitter e){
        if(!active || template == null || random.nextInt(probability > 0 ? probability : 1) != 0) return;
        int pc = max_parts - min_parts > 0 ? random.nextInt(max_parts - min_parts) + min_parts : 1;
        DynamicParticle p;
        for(int i = 0; i < pc; i ++){
            p = getVariant(template.get());
            if(e.attach(p) && !listeners.isEmpty())
                for(ParticleListener l : listeners)
                    l.particleEvent(p);
        }
    }
    /**
     * Update the specified particle (unused)
     * @param p particle
     */
    public void update(DynamicParticle p) {}
    /**
     * Attach a particle listener
     * <p>Attached listeners will receive a reference to created particles upon
     * emission. If this object is assigned as behaviour to multiple emitters,
     * attached listeners will receive particles from each.
     * @param p particle listener
     */
    public final void addListener(ParticleListener p){
        listeners.add(p);
    }
    /**
     * Remove a particle listener
     * @param p particle listener
     * @return false if the listener does not exist
     */
    public final boolean removeListener(ParticleListener p){
        return listeners.remove(p);
    }
    /**
     * Clear all attached particle listeners
     */
    public final void clearListeners(){
        listeners.clear();
    }
    private int getVariation(int p, float v){
        float f = getVariation((float) p, v);
        return Math.round(f);
    }
    private float getVariation(float p, float v){
        v *= p;
        return p + (random.nextFloat() * v * 2f - v);
    }
    private DynamicParticle getVariant(DynamicParticle t){
        DynamicParticle p = t.copy();
        if(palette != null)
            p.setColour(palette.get().copy());
        p.setDecayTime(getVariation(decay, decay_v));
        p.setSize(getVariation(size, size_v));
        p.setAlpha(getVariation(alpha, alpha_v));
        p.setMass((p.getSize() / t.getSize()) * p.getMass());
        return p;
    }
}