package pslib;
import java.awt.Font;
import java.awt.geom.Rectangle2D;
import com.sun.opengl.util.j2d.*;
import processing.opengl.PGraphicsOpenGL;
/**
 * TextParticle Class
 * <p>The TextParticle class extends the functionality of the DynamicParticle
 * class by allowing text to be rendered at the particle location.
 * <p>This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * <p>You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see http://www.gnu.org/licenses/.
 * @author Stephen Callear
 */
public class TextParticle extends DynamicParticle {
    private TextRenderer renderer;
    private PSVector bounds, offset;
    private String font, text;
    private int font_size, anchor;
    private boolean rvar_aa, rvar_mm, rvar_sm;
    /**
     * Create a new object instance with the specified variables
     * @param f font
     * @param s font size
     */
    public TextParticle(String f, int s){
        this(f, s, 0);
    }
    /**
     * Create a new object instance with the specified variables
     * @param f font
     * @param s font size
     * @param d decay time (ms);
     */
    public TextParticle(String f, int s, int d){
        this(f, s, true, true, d);
    }
    /**
     * Create a new object instance with the specified variables
     * @param f font
     * @param s font size
     * @param aa anti-aliasing
     * @param mm mipmap generation
     * @param d decay time (ms)
     */
    public TextParticle(String f, int s, boolean aa, boolean mm, int d){
        super(d);
        super.setSize(1f);
        renderer = new TextRenderer(new Font(f, Font.TRUETYPE_FONT, s), aa, true, null, mm);
        renderer.setSmoothing(true);
        bounds = new PSVector();
        offset = new PSVector();
        anchor = CORNER;
        text = "";
        // Store renderer values to permit copying
        font = f;
        font_size = s;
        rvar_aa = aa;
        rvar_mm = mm;
        rvar_sm = true;
    }
    /**
     * Set the text string
     * @param t text string
     */
    public final void setText(String t){
        text = t;
        Rectangle2D.Double b = (Rectangle2D.Double) renderer.getBounds(text);
        bounds.set((float) b.width, (float) b.height, 0f);
    }
    /**
     * Return a vector containing the size of the text bounding box
     * <p>The bounding box differs from the rendered size, this should be calculated
     * by scaling the returned vector by the value returned from getSize().
     * @return a vector containing the size of the text bounding box
     */
    public final PSVector getTextBounds(){
        return bounds.copy();
    }
    /**
     * Return a reference to the TextRenderer object employed by this particle
     * @return a reference to the TextRenderer object employed by this particle
     */
    public final TextRenderer getTextRenderer(){
        return renderer;
    }
    /**
     * Set the text offset vector
     * <p>The text offset vector specifies an offset for the render location.
     * By default the value is (0, 0, 0) resulting in the midpoint of the text
     * being aligned to the location vector.
     * @param v
     */
    public final void setOffset(PSVector v){
        offset = v;
    }
    /**
     * Set the text anchor point (CORNER/MIDPOINT)
     * <p>The anchor point defines the point of the text bounding box that is
     * used as the origin prior to translation. The MIDPOINT value uses the
     * midpoint of the bounding box, whilst the CORNER value uses the lower left
     * corner.
     * <p>Default: CORNER
     * @param a anchor point
     */
    public final void setAnchorPoint(int a){
        anchor = a;
    }
    /**
     * Enable/disable smoothing
     * @param s smoothing
     */
    public final void setSmoothing(boolean s){
        renderer.setSmoothing(s);
    }
    /**
     * Render the text to the specified OpenGL renderer
     * @param pgl OpenGL renderer
     */
    @Override public void render(PGraphicsOpenGL pgl){
        render(pgl, null, null, null);
    }
    /**
     * Render the text to the specified OpenGL renderer
     * @param pgl OpenGL renderer
     * @param cpv camera position vector
     * @param ctv camera target vector
     * @param cuv camera up vector
     */
    @Override public void render(PGraphicsOpenGL pgl, PSVector cpv, PSVector ctv, PSVector cuv){
        if(text == null || text.equals("") || bounds.x <= 0f || bounds.y <= 0f) return;
        // Set the text colour
        float[] c = getColour().getComponents(new float[] {});
        renderer.setColor(c[0], c[1], c[2], c[3] * getRenderAlpha());
        // Calculate translations
        pgl.pushMatrix();
        PSVector tv = offset.copy();
        tv.add(getPosition());
        pgl.translate(tv.x, tv.y, tv.z);
        // Apply billboard matrix
        if(cpv != null && ctv != null && cuv != null){
            PSVector tlv = PSVector.sub(ctv, cpv);
            tlv.normalise();
            PSVector trv = cuv.cross(tlv);
            PSVector tuv = tlv.cross(trv);
            pgl.applyMatrix(trv.x, tuv.x, tlv.x, 0.0f,
                            trv.y, tuv.y, tlv.y, 0.0f,
                            trv.z, tuv.z, tlv.z, 0.0f,
                            0f, 0f, 0f, 1f);
        }
        // Correct alignment
        pgl.rotateZ(PGraphicsOpenGL.PI);
        // Centre about anchor
        if(anchor == MIDPOINT){
            tv = bounds.copy();
            tv.scale(getRenderSize() / bounds.y * -0.5f);
            pgl.translate(tv.x, tv.y, tv.z);
        }
        // Render text
        pgl.beginGL();
        renderer.begin3DRendering();
        renderer.draw3D(text, 0f, 0f, 0f, getRenderSize() / bounds.y);
        renderer.end3DRendering();
        renderer.flush();
        pgl.endGL();
        pgl.popMatrix();
    }
    /**
     * Return a copy of this particle
     * @return a copy of this particle
     */
    @Override public DynamicParticle copy(){
        TextParticle p = new TextParticle(font, font_size, rvar_aa, rvar_mm,  getInitialDecayTime());
        p.setText(text);
        p.setOffset(offset.copy());
        p.setSmoothing(rvar_sm);
        super.map(p);
        return p;
    }
}
