package pslib;
import java.util.ArrayList;
/**
 * TimedEmission Class
 * <p>The TimedEmission class implements the PSEmission interface allowing the
 * automated emission of random particles derived from the object variables from
 * any emitters to which it is assigned.
 * <p>Unlike the TemplateEmission class, the particle emission interval and particle
 * amount are user-specified and invariant.
 * @author Stephen Callear
 * <p>This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * <p>You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see http://www.gnu.org/licenses/.
 * @author Stephen Callear
 * @see TemplateEmission
 */
public class TimedEmission extends Object implements PSEmission, PSConstants {
    private WeightedList<DynamicParticle> template;
    private WeightedList<PSColour> palette;
    private ArrayList<ParticleListener> listeners;
    private float size, alpha, size_v, alpha_v, decay_v;
    private int decay, interval, particles, count;
    private boolean active;
    /**
     * Create a new object instance with the specified variables
     * <p>Default decay, size and alpha values are taken from the specified particle
     * with 0% variance.
     * @param t particle template
     * @param i emission interval
     * @param p particle amount
     */
    public TimedEmission(DynamicParticle t, int i, int p){
        this(new WeightedList<DynamicParticle>(new DynamicParticle[] { t }), i, p);
    }
    /**
     * Create a new object instance with the specified variables
     * <p>Default decay, size and alpha values are taken from a particle selected
     * at random from the specified list with 0% variance.
     * @param t particle template list
     * @param i emission interval
     * @param p particle amount
     */
    public TimedEmission(WeightedList<DynamicParticle> t, int i, int p){
        template = t;
        palette = null;
        interval = i > 0 ? i : 1;
        particles = p;  
        count = 0;
        active = true;
        listeners = new ArrayList<ParticleListener>();
        // Set base parameters from randomly selected particle
        DynamicParticle tp = template.get();
        decay = tp.getInitialDecayTime();
        size = tp.getSize();
        alpha = tp.getAlpha();
        decay_v = 0f;
        size_v = 0f;
        alpha_v = 0f;
    }
    /**
     * Set the particle template
     * <p>The particle template specifies the list of particles from which emitted
     * particles may be derived. If no palette is specified emitted particles
     * will inherit their colour from the template.
     * @param t particle template
     */
    public final void setTemplate(WeightedList<DynamicParticle> t){
        template = t;
    }
    /**
     * Set the particle colour palette
     * <p>The particle colour palette specifies the list of colours from which
     * emitted particles may select. If no palette is specified emitted particles
     * will inherit their colour from the particle template.
     * @param p colour palette
     */
    public final void setPalette(WeightedList<PSColour> p){
        palette = p;
    }
    /**
     * Set the active state of the behaviour
     * <p>A value of true will result in particle emission on the update routine
     * of any emitters to which this behaviour is attached.
     * <p>Default: true
     * @param a
     */
    public final void setActive(boolean a){
        active = a;
    }
    /**
     * Set the particle emission interval
     * <p>The particle emission interval specifies how many calls are required
     * to the <code>GenericEmitter.update()</code> between emission events.
     * Typically emitter updates will occur on a per-frame basis, so in many cases
     * this value may be viewed as a frame interval.
     * @param i interval
     */
    public final void setEmissionInterval(int i){
        interval = i;
    }
    /**
     * Set the amount of particles emitted on each particle emission event
     * @param p particles
     */
    public final void setEmittedParticles(int p){
        particles = p;
    }
    /**
     * Set the emitted particle decay value and variance
     * @param d decay time (ms)
     * @param v variance (+/- %)
     */
    public final void setDecay(int d, int v){
        decay = d;
        decay_v = (float) v / 100f;
    }
    /**
     * Set the emitted particle size value and variance
     * @param s size
     * @param v variance (+/- %)
     */
    public final void setSize(float s, int v){
        size = s;
        size_v = (float) v / 100f;
    }
    /**
     * Set the emitted particle alpha value and variance
     * @param a alpha
     * @param v variance (+/- %)
     */
    public final void setAlpha(float a, int v){
        alpha = a;
        alpha_v = (float) v / 100f;
    }
    /**
     * Update the specified emitter
     * <p>The update routine attaches random particles derived from the object
     * variables to the specified emitter. References to created particles are
     * sent to any attached listeners.
     * @param e emitter
     */
    public void update(GenericEmitter e){
        if(!active || template == null) return;
        if(count == interval){
            count = 0; 
            DynamicParticle p;
            for(int i = 0; i < particles; i ++){
                p = getVariant(template.get());
                if(e.attach(p) && !listeners.isEmpty())
                    for(ParticleListener l : listeners)
                        l.particleEvent(p);
            }
        }
        count ++;
    }
    /**
     * Update the specified particle (unused)
     * @param p particle
     */
    public void update(DynamicParticle p) {}
    /**
     * Attach a particle listener
     * <p>Attached listeners will receive a reference to created particles upon
     * emission. If this object is assigned as behaviour to multiple emitters,
     * attached listeners will receive particles from each.
     * @param p particle listener
     */
    public final void addListener(ParticleListener p){
        listeners.add(p);
    }
    /**
     * Remove a particle listener
     * @param p particle listener
     * @return false if the listener does not exist
     */
    public final boolean removeListener(ParticleListener p){
        return listeners.remove(p);
    }
    /**
     * Clear all attached particle listeners
     */
    public final void clearListeners(){
        listeners.clear();
    }
    private int getVariation(int p, float v){
        float f = getVariation((float) p, v);
        return Math.round(f);
    }
    private float getVariation(float p, float v){
        v *= p;
        return p + ((float) Math.random()* v * 2f - v);
    }
    private DynamicParticle getVariant(DynamicParticle t){
        DynamicParticle p = t.copy();
        if(palette != null)
            p.setColour(palette.get().copy());
        p.setDecayTime(getVariation(decay, decay_v));
        p.setSize(getVariation(size, size_v));
        p.setAlpha(getVariation(alpha, alpha_v));
        p.setMass((p.getSize() / t.getSize()) * p.getMass());
        return p;
    }
}
