package pslib;
import java.util.Random;
import java.util.ArrayList;
/**
 * VariantEmission Class
 * <p>The VariantEmission class implements the PSEmission interface allowing
 * the automated emission of random particles with variant properties from the
 * emitter to which it is assigned.
 * <p>Unlike the TemplateEmission class, particle properties are derived entirely
 * from the specified template and percentage variance.
 * <p>This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * <p>You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see http://www.gnu.org/licenses/.
 * @author Stephen Callear
 * @see TemplateEmission
 */
public class VariantEmission extends Object implements PSEmission, PSConstants {
    private WeightedList<DynamicParticle> template;
    private ArrayList<ParticleListener> listeners;
    private float variance;
    private int probability, min_parts, max_parts;
    private boolean active;
    private Random random;
    /**
     * Create a new object instance with the specified variables
     * @param t particle template
     * @param v percentage variance
     * @param ep emission probability
     * @param mi minimum particles
     * @param ma maximum particles
     */
    public VariantEmission(DynamicParticle t, int v, int ep, int mi, int ma){
        this(new WeightedList<DynamicParticle>(new DynamicParticle[] { t }), v, ep, mi, ma);
    }
    /**
     * Create a new object instance with the specified variables
     * @param t particle template array
     * @param v percentage variance
     * @param ep emission probability
     * @param mi minimum particles
     * @param ma maximum particles
     */
    public VariantEmission(WeightedList<DynamicParticle> t, int v, int ep, int mi, int ma){
        template = t;
        variance = (float) v / 100f;
        probability = ep;
        min_parts = mi;
        max_parts = ma;
        active = true;
        listeners = new ArrayList<ParticleListener>();
        random = new Random();
    }
    /**
     * Set the particle template
     * @param t particle template
     */
    public final void setTemplate(DynamicParticle t){
        template = new WeightedList<DynamicParticle>(new DynamicParticle[] { t });
    }
    /**
     * Set the particle template list
     * @param t particle template list
     */
    public final void setTemplate(WeightedList<DynamicParticle> t){
        template = t;
    }
    /**
     * Set the particle parameter variance
     * <p>Particle variations inherit the parameters of the template, with the
     * specified +/- variance being applied to particle decay, alpha and size.
     * Particle mass preserves the ratio between size and mass variables within
     * the template particle.
     * @param v percentage variance
     */
    public final void setVariance(int v){
        variance = v;
    }
    /**
     * Set the emission probability
     * <p>Emission probability defines the likelihood of particle emission and
     * takes the form 1:probability.
     * @param ep emission probability
     */
    public final void setEmissionProbability(int ep){
        probability = ep;
    }
    /**
     * Set the minimum emitted particles
     * @param p minimum particles
     */
    public final void setMinimumParticles(int p){
        min_parts = p;
    }
    /**
     * Set the maximum emitted particles
     * @param p maximum particles
     */
    public final void setMaximumParticles(int p){
        max_parts = p;
    }
    /**
     * Set the active state of the behaviour
     * <p>A value of true will result in particle emission on the update routine
     * of any emitters to which this behaviour is attached.
     * @param a
     */
    public final void setActive(boolean a){
        active = a;
    }
    /**
     * Update the specified GenericEmitter
     * <p>The update routine attaches random particles derived from the particle
     * template to the specified emitter. References to created particles are
     * sent to any attached listeners.
     * <p>Default: true
     * @param e GenericEmitter
     */
    public void update(GenericEmitter e){
        if(!active || template == null || random.nextInt(probability > 0 ? probability : 1) != 0) return;
        int pc = max_parts - min_parts > 0 ? random.nextInt(max_parts - min_parts) + min_parts : 1;
        DynamicParticle p;
        for(int i = 0; i < pc; i ++){
            p = getVariant(template.get());
            if(e.attach(p) && !listeners.isEmpty())
                for(ParticleListener l : listeners)
                    l.particleEvent(p);
        }
    }
    /**
     * Update the specified DynamicPartice (unused)
     * @param p DynamicParticle
     */
    public void update(DynamicParticle p){
        // Unused
    }
    /**
     * Attach a particle listener
     * <p>Attached listeners will receive a reference to created particles upon 
     * emission. If this object is assigned as behaviour to multiple emitters,
     * attached listeners will receive particles from each.
     * @param p particle listener
     */
    public final void addListener(ParticleListener p){
        listeners.add(p);
    }
    /**
     * Remove a particle listener
     * @param p particle listener
     * @return false if the listener does not exist
     */
    public final boolean removeListener(ParticleListener p){
        return listeners.remove(p);
    }
    /**
     * Clear all attached particle listeners
     */
    public final void clearListeners(){
        listeners.clear();
    }
    private int getVariation(int p, float v){
        float f = getVariation((float) p, v);
        return Math.round(f);
    }
    private float getVariation(float p, float v){
        v *= p;
        return p + (random.nextFloat() * v * 2f - v);
    }
    private DynamicParticle getVariant(DynamicParticle t){
        DynamicParticle p = t.copy();
        p.setDecayTime(getVariation(p.getInitialDecayTime(), variance));
        p.setAlpha(getVariation(p.getAlpha(), variance));
        p.setSize(getVariation(p.getSize(), variance));
        p.setMass((p.getSize() / t.getSize()) * p.getMass());
        return p;
    }
}
