package pslib;
import java.util.Random;
import java.util.Iterator;
import java.util.Arrays;
import java.util.List;
import java.util.ArrayList;
/**
 * WeightedList Class
 * <p>The WeightedArray class allows a list of objects to be stored alongside
 * an integer weight. Calls to the <code>get()</code> function will return a random
 * object from the list, the probability of which is determined by its weight value.
 * <p>This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p>This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * <p>You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see http://www.gnu.org/licenses/.
 * @author Stephen Callear
 */
public class WeightedList<T> extends Object implements Iterable<T>{
    private List<T> list;
    private int[] weights;
    private Random random;
    /**
     * Create a new object instance with the specified list elements and equal weighting
     * @param e list elements
     */
    public WeightedList(T[] e){
        this(e, new int[] { 1 });
    }
    /**
     * Create a new object instance with the specified list elements and weighting
     * <p>Negative, zero or absent weight values will set to the default value of 1.
     * @param a
     * @param w
     */
    public WeightedList(T[] a, int[] w){
        list = new ArrayList(a.length);
        list.addAll(Arrays.asList(a));
        // Parse weights
        int[] wtmp = new int[list.size()];
        for(int i = 0; i < wtmp.length; i ++){
            if(i < w.length && w[i] >= 1) wtmp[i] = w[i];
            else wtmp[i] = 1;
        }
        // Generate cumulative weight array
        weights = Arrays.copyOf(wtmp, wtmp.length);
        for(int i = 1; i < weights.length; i ++)
            weights[i] = weights[i - 1] + wtmp[i];      
        random = new Random();
    }
    /**
     * Return a random list element
     * <p>The probability of each list element being returned is defined by its
     * weight value.
     * @return a random list element
     */
    public final T get(){
        int r = random.nextInt(weights[weights.length - 1]) + 1;
        T obj = list.get(0);
        for(int i = 0; i < weights.length - 1; i ++)
            if(r > weights[i]) obj = list.get(i + 1);
        return obj;
    }
    /**
     * Return an iterator for the list
     * @return an iterator for the list
     */
    public final Iterator iterator(){
        return list.iterator();
    }
}


